/*
 * filename: hold_buffer.c
 *
 * Developed by secunet AG
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/config.h>
#include <linux/skbuff.h>
#include <linux/errno.h>
#include <linux/spinlock.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/rtnetlink.h>
#include <linux/netdevice.h>

#include <net/route.h>

#ifdef CONFIG_SMP
#include <linux/interrupt.h>
#endif

#include "ipsec_encap.h"
#include "ipsec_tunnel.h"
#include "ipsec_rcv.h"
#include "ipsec_xform.h"
#include "ipsec_radij.h"
#include "hold_buffer.h"
#include "ipsec_cb.h"

#ifdef CONFIG_IPSEC_PKT_SEQ
#  include "ipsec_threads/seq.h"
#endif

#define UNDEFINED			0
#define TRAP_INSERTED		1
#define TRAP_DELETED		2
#define HOLD_INSERTED		3
#define HOLD_DELETED		4
#define SA_ESTABLISHED		5
#define SA_DELETED			6

#define UNSUPPORTED 0
#define ADD_TRAP	1
#define DEL_TRAP	2
#define ADD_HOLD	3
#define DEL_HOLD	4
#define ADD_SA		5
#define DEL_SA		6
#define ADD_REJECT	7
#define DEL_REJECT	8

#define SEND	1
#define DELETE	2

#define NO_TYPE 0
#define TX_TYPE 1
#define RX_TYPE 2

#define HOLD_BUF_RETRY 3

typedef struct hold_eroute {
	struct hold_eroute *next;
	struct eroute *er;
	struct sk_buff *pkgq;
	unsigned short state;
	unsigned short type;
	spinlock_t lock;
} hold_eroute_t;

#define HB_MAX_BUF_SIZE 200
static unsigned long hb_buf_size = 0;

static hold_eroute_t *her_lst_achr;
static unsigned int in_use = 0;

static spinlock_t inuse_lock = SPIN_LOCK_UNLOCKED;
static rwlock_t her_lst_lock = RW_LOCK_UNLOCKED;

#ifdef CONFIG_SMP
static struct tasklet_struct ipsec_hold_tx_buf_tlet;
static struct tasklet_struct ipsec_hold_rx_buf_tlet;
#endif

static spinlock_t tlet_tx_buf_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t tlet_rx_buf_lock = SPIN_LOCK_UNLOCKED;
static struct sk_buff *tlet_send_buf = NULL;
static struct sk_buff *tlet_rx_buf = NULL;


#define CLEAR_ER(er) {\
	if (er) { \
		memset((er), 0, sizeof(struct eroute)); \
		kfree((er)); \
	}\
}

static void hb_link_skb2queue(struct hold_eroute *her, struct sk_buff *skb)
{
	hb_buf_size++;

	if (her->pkgq) {
		skb->next = her->pkgq;
		skb->prev = her->pkgq->prev;

		her->pkgq->prev->next = skb;
		her->pkgq->prev = skb;

		return;
	}

	skb->next = skb;
	skb->prev = skb;

	her->pkgq = skb;
}

static inline void fillup_matcher(struct sk_buff *skb,
                                  struct sockaddr_encap *matcher)
{
	struct iphdr *iph;

	iph = skb->nh.iph;

	memset(matcher, 0, sizeof(struct sockaddr_encap));
	matcher->sen_len = sizeof (struct sockaddr_encap);
	matcher->sen_family = AF_ENCAP;
	matcher->sen_type = SENT_IP4;

	matcher->sen_ip_src.s_addr = iph->saddr;
	matcher->sen_ip_dst.s_addr = iph->daddr;

	matcher->sen_proto = iph->protocol;
	switch (matcher->sen_proto) {
	case IPPROTO_UDP: {
		struct udphdr *udph;
		
		udph = (struct udphdr *) ((caddr_t)iph + (iph->ihl << 2));

		matcher->sen_ip_sport = udph->source;
		matcher->sen_ip_dport = udph->dest;
		break;
	}
	case IPPROTO_TCP: {
		struct tcphdr *tcph;
		
		tcph = (struct tcphdr *) ((caddr_t)iph + (iph->ihl << 2));

		matcher->sen_ip_sport = tcph->source;
		matcher->sen_ip_dport = tcph->dest;
		break;
	}
	default:
		matcher->sen_ip_sport = 0;
		matcher->sen_ip_dport = 0;
	}

	return;
}


static struct hold_eroute *find_hold_eroute(struct eroute *er, int exact)
{
	struct hold_eroute *ret = NULL;

	if (er) {
		const int size = sizeof(struct sockaddr_encap);
		struct hold_eroute *p = her_lst_achr;

		read_lock_bh(&her_lst_lock);

		for (ret = NULL; p && !ret; p = p->next) {
			spin_lock(&p->lock);
			if (!memcmp((char *) &er->er_eaddr, (char *) &p->er->er_eaddr,
			            2 * size)) {
				ret = p;
				goto next;
			}

			if (!exact) {
				if (p->er->er_eaddr.sen_ip_sport != er->er_eaddr.sen_ip_sport)
					goto next;
				if (p->er->er_eaddr.sen_ip_dport != er->er_eaddr.sen_ip_dport)
					goto next;
				if ((er->er_eaddr.sen_ip_src.s_addr & p->er->er_emask.sen_ip_src.s_addr) !=
				    p->er->er_eaddr.sen_ip_src.s_addr)
					goto next;
				if ((er->er_eaddr.sen_ip_dst.s_addr & p->er->er_emask.sen_ip_dst.s_addr) !=
				    p->er->er_eaddr.sen_ip_dst.s_addr)
					goto next;

				ret = p;
			}

next:
			spin_unlock(&p->lock);
		}

		read_unlock_bh(&her_lst_lock);
	}
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
	else
		printk(KERN_DEBUG "find_hold_eroute: no such eroute.\n");
#endif

	return ret;
}


static struct hold_eroute *find_similar_hold_eroute(struct eroute *er)
{
	struct hold_eroute *ret = NULL;
	struct eroute matcher;
	int exact;

	exact = 1;

again:
	memcpy((char *) &matcher, (char *) er, sizeof(struct eroute));

	matcher.er_eaddr.sen_ip_sport = 0;
	ret = find_hold_eroute(&matcher, exact);
	if (ret)
		goto out;
	matcher.er_eaddr.sen_ip_sport = er->er_eaddr.sen_ip_sport;

	matcher.er_eaddr.sen_ip_dport = 0;
	ret = find_hold_eroute(&matcher, exact);
	if (ret)
		goto out;
	matcher.er_eaddr.sen_ip_dport = er->er_eaddr.sen_ip_dport;

	matcher.er_eaddr.sen_proto = 0;
	ret = find_hold_eroute(&matcher, exact);
	if (ret)
		goto out;
	matcher.er_eaddr.sen_proto = er->er_eaddr.sen_proto;

	matcher.er_eaddr.sen_ip_sport = 0;
	matcher.er_eaddr.sen_ip_dport = 0;
	ret = find_hold_eroute(&matcher, exact);
	if (ret)
		goto out;
	matcher.er_eaddr.sen_ip_sport = er->er_eaddr.sen_ip_sport;
	matcher.er_eaddr.sen_ip_dport = er->er_eaddr.sen_ip_dport;

	matcher.er_eaddr.sen_ip_sport = 0;
	matcher.er_eaddr.sen_proto = 0;
	ret = find_hold_eroute(&matcher, exact);
	if (ret)
		goto out;
	matcher.er_eaddr.sen_ip_sport = er->er_eaddr.sen_ip_sport;
	matcher.er_eaddr.sen_proto = er->er_eaddr.sen_proto;

	matcher.er_eaddr.sen_ip_dport = 0;
	matcher.er_eaddr.sen_proto = 0;
	ret = find_hold_eroute(&matcher, exact);
	if (ret)
		goto out;
	matcher.er_eaddr.sen_ip_dport = er->er_eaddr.sen_ip_dport;
	matcher.er_eaddr.sen_proto = er->er_eaddr.sen_proto;

	matcher.er_eaddr.sen_ip_sport = 0;
	matcher.er_eaddr.sen_ip_dport = 0;
	matcher.er_eaddr.sen_proto = 0;
	ret = find_hold_eroute(&matcher, exact);

	if (exact) {
		exact = 0;
		goto again;
	}

out:
	return ret;
}


static void flush_rx_buf(unsigned long data)
{
	struct sk_buff *rx_buf, *skb;
	int err;

#ifdef CONFIG_SMP
	if (!write_trylock(&ipsec_sync_lock)) {
		local_bh_disable();
		tasklet_schedule(&ipsec_hold_rx_buf_tlet);
		local_bh_enable();

		return;
	}

	write_unlock(&ipsec_sync_lock);
#endif

	spin_lock(&tlet_rx_buf_lock);

	rx_buf = tlet_rx_buf;
	tlet_rx_buf = NULL;

	spin_unlock(&tlet_rx_buf_lock);

	if (rx_buf) {
		rx_buf->prev->next = NULL;

		for (skb = rx_buf; skb; skb = rx_buf) {
			rx_buf = skb->next;

			skb->next = NULL;
			skb->prev = NULL;

			err = ipsec_start_xmit_entry1(skb, IPSEC_SKB_CB(skb)->hb_dev);
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			if (err) {
				printk(KERN_DEBUG "flush_rx_buf: "
				                  "ipsec_start_xmit_entry1 failed %d\n",
			       err);
			}
#endif

			hb_buf_size--;
		}
	}

	if (tlet_rx_buf)
		printk(KERN_NOTICE "flush_rx_buf: Ooops, tlet_rx_buf is not empty!\n");

	return;
}

static void flush_tx_buf(unsigned long data)
{
	struct sk_buff *send_buf = (struct sk_buff *) data;
	struct sk_buff *skb;
	int cnt, err;

	if (send_buf) {
		send_buf->prev->next = NULL;

		spin_lock(&tlet_tx_buf_lock);

		if (tlet_send_buf) {
			struct sk_buff *last;

			for (last = tlet_send_buf; last->next; last = last->next) {
			}
			last->next = send_buf;
		} else
			tlet_send_buf = send_buf;

		spin_unlock(&tlet_tx_buf_lock);
	}

#ifdef CONFIG_SMP
	if (!write_trylock(&ipsec_sync_lock)) {
		local_bh_disable();
		tasklet_schedule(&ipsec_hold_tx_buf_tlet);
		local_bh_enable();

		return;
	}

	write_unlock(&ipsec_sync_lock);
#endif

	spin_lock(&tlet_tx_buf_lock);

	send_buf = tlet_send_buf;
	tlet_send_buf = NULL;

	spin_unlock(&tlet_tx_buf_lock);

	for (skb = send_buf, cnt = 0; skb; skb = send_buf) {
		send_buf = skb->next;

		skb->next = NULL;
		skb->prev = NULL;
		
		err = ipsec_start_xmit_entry1(skb, IPSEC_SKB_CB(skb)->hb_dev);
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
		if (err) {
			printk(KERN_DEBUG "flush_tx_buf: "
			                  "ipsec_start_xmit_entry1 failed with %d\n",
			       err);
		}
#endif
		 cnt++;
	}

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
	if (cnt)
		printk(KERN_DEBUG "flush_tx_buf: flushed %d packets.\n", cnt);
#endif

	return;
}


static int flush_packets(struct hold_eroute *her, int cmd)
{
	struct sk_buff *queue, *send_buf = NULL;


	if (!her)
		return -EFAULT;

	if (cmd < SEND || cmd > DELETE) {
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
		printk(KERN_DEBUG "flush_packets: unknown cmd %d.\n", cmd);
#endif
		return -EINVAL;
	}

	spin_lock(&her->lock);
 	queue = her->pkgq;
	her->pkgq = NULL;
	spin_unlock(&her->lock);

	if (queue) {
		struct sk_buff *skb;
 
		queue->prev->next = NULL;
		while (queue) {
			skb = queue;
			queue = queue->next;

			switch (cmd) {
			case SEND:
				if (!send_buf) {
					skb->next = skb;
					skb->prev = skb;

					send_buf = skb;
					break;
				}

				skb->next = send_buf;
				skb->prev = send_buf->prev;

				send_buf->prev->next = skb;
				send_buf->prev = skb;

				break;
			case DELETE:
				KLIPS_PRINT(debug_tunnel,
				            KERN_DEBUG "flush_packets: delete pkt 0x%08x.\n",
				            (unsigned int) skb);

				kfree_skb(skb);
				break;
			}

			hb_buf_size--;
		}
	}

	if (send_buf)
		flush_tx_buf((unsigned long) send_buf);

	return 0;
}


static int replace_er(struct hold_eroute *her, struct eroute *er, int state,
                      int store)
{
	if (!her || !er)
		return -EFAULT;

	spin_lock(&her->lock);

	if (store) {
		struct eroute *old_er;

		old_er = her->er;
		her->er = er;

		CLEAR_ER(old_er);
	} else
		memcpy((char *) her->er, (char *) er, sizeof(struct eroute));

	her->state = state;

	spin_unlock(&her->lock);

	return 0;
}


static int delete_her(struct hold_eroute *her)
{
	if (her) {
		if (her->pkgq) {
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			printk(KERN_DEBUG "delete_her: deleting used hold_eroute.\n");
#endif
			flush_packets(her, DELETE);
		}

		write_lock_bh(&her_lst_lock);

		if (her == her_lst_achr)
			her_lst_achr = her->next;
		else {
			struct hold_eroute *prev;

			for (prev = her_lst_achr; prev->next != her; prev = prev->next) {
			}

			if (prev)
				prev->next = her->next;
		}

		write_unlock_bh(&her_lst_lock);
		
		CLEAR_ER(her->er);
		kfree(her);

		return 0;
	}

	return -EFAULT;
}


static struct hold_eroute *alloc_her(struct eroute *er, int state, int store)
{
	struct hold_eroute *her;

	if (!er)
		return NULL;

	her = (struct hold_eroute *) kmalloc(sizeof(struct hold_eroute),
	                                     GFP_ATOMIC);
	if (her) {
		memset(her, 0, sizeof(struct hold_eroute));

		if (!store) {
			her->er = (struct eroute *) kmalloc(sizeof(struct eroute),
			                                    GFP_ATOMIC);
			if (!her->er) {
				kfree(her);
				return NULL;
			}

			memcpy((char *) her->er, (char *) er, sizeof(struct eroute));
		} else
			her->er = er;

		her->state = state;
		spin_lock_init(&her->lock);

		write_lock_bh(&her_lst_lock);

		if (!her_lst_achr)
			her_lst_achr = her;
		else {
			struct hold_eroute *last;

			for (last = her_lst_achr; last->next; last = last->next) {
			}
			last->next = her;
		}

		write_unlock_bh(&her_lst_lock);
	}

	return her;
}


static inline int skb_belongs2er(struct eroute *er, struct sk_buff *skb)
{
	struct sockaddr_encap matcher;
	struct eroute *orig_er;

	fillup_matcher(skb, &matcher);

	orig_er = ipsec_findroute(&matcher);
	if (er && !memcmp((char *) er, (char *) orig_er, sizeof(struct eroute)))
		return 1;

	return 0;
}

static int derivation_possible(struct hold_eroute *her, struct eroute *er,
                               int store)
{
	struct hold_eroute *new_her;
	int err;

	err = -ENOMEM;
	new_her = alloc_her(er, HOLD_INSERTED, store);
	if (new_her) {
		struct sk_buff *cur_skb, *skb, *last;
		int pkts_moved = 0;

		new_her->type = TX_TYPE;

		spin_lock(&her->lock);
		spin_lock(&new_her->lock);

		for (cur_skb = her->pkgq, last = NULL; cur_skb; ) {
			skb = cur_skb;
			cur_skb = cur_skb->next;

			if (skb_belongs2er(new_her->er, skb)) {
				pkts_moved++;

				if (skb->prev != skb) {
					skb->prev->next = skb->next;
					skb->next->prev = skb->prev;

					if (skb == her->pkgq)
						her->pkgq = skb->next;
				} else
					her->pkgq = NULL;

				skb->prev = NULL;
				skb->next = NULL;

				hb_link_skb2queue(new_her, skb);

				last = skb;
			}

			if (!her->pkgq || cur_skb == her->pkgq) {
				KLIPS_HB_PRINT("derivation_possible: %d packets moved.\n",
				               pkts_moved);
				break;
			}
		}

		spin_unlock(&new_her->lock);
		spin_unlock(&her->lock);

		err = 0;
	}

	return err;
}

static int run_machine(struct eroute *er, int action, int store)
{
	struct hold_eroute *her;
	int ret;

	KLIPS_HB_PRINT("run_machine: invoked with action %d.\n", action);

	her = find_hold_eroute(er, 1);
	while (!her) {
		ret = -EINVAL;

		switch (action) {
		  case ADD_TRAP:
			her = alloc_her(er, TRAP_INSERTED, store);
			if (her) {
				ret = 0;
				her->type = TX_TYPE;
			}

			break;
		  case ADD_SA:
			her = alloc_her(er, SA_ESTABLISHED, store);
			if (her) {
				ret = 0;
				her->type = RX_TYPE;
			}

			break;
		  case ADD_HOLD:
			her = find_similar_hold_eroute(er);
			if (her)
				continue;

			KLIPS_HB_PRINT("run_machine: "
			               "unable to find similar hold eroute.\n");
			break;
		  default:
			KLIPS_HB_PRINT("run_machine: action %d required a hold eroute.\n",
			               action);
		}

		return ret;
	}

	KLIPS_HB_PRINT("run_machine: found hold eroute with state %d.\n",
	               her->state);

	if (action == UNSUPPORTED) {
		if (her) {
			flush_packets(her, DELETE);
			delete_her(her);
		}

		return -1;
	}

	ret = -1;

	switch (her->state) {
	  case TRAP_INSERTED:
		switch (action) {
		  case DEL_TRAP:
			ret = replace_er(her, er, TRAP_DELETED, store);
			break;
		  case ADD_HOLD:
			ret = derivation_possible(her, er, store);
			if (ret) {
				flush_packets(her, DELETE);
				delete_her(her);

				if (ret == -ENOSYS) {
					her = find_hold_eroute(er, 1);
					if (her) {
						flush_packets(her, DELETE);
						delete_her(her);
					}
				}
			}

			break;
		  default:
			flush_packets(her, DELETE);
			delete_her(her);

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			printk(KERN_DEBUG "run_machine: unsupported action %d in state "
			                  "TRAP_INSERTED.\n",
			       action);
#endif
		}

		break;
	  case TRAP_DELETED:
		switch(action) {
		  case ADD_HOLD:
			ret = replace_er(her, er, HOLD_INSERTED, store);
			break;
		  case DEL_HOLD:
			flush_packets(her, DELETE);
			delete_her(her);

			break;
		  case ADD_SA:
			ret = replace_er(her, er, SA_ESTABLISHED, store);
			break;
		  default:
			flush_packets(her, DELETE);
			delete_her(her);

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			printk(KERN_DEBUG "run_machine: "
			                  "unsupported action %d in state TRAP_DELETED.\n",
			       action);
#endif
		}

		break;
	  case HOLD_INSERTED:
		switch(action) {
		  case DEL_HOLD:
			ret = replace_er(her, er, HOLD_DELETED, store);
			break;
		  default:
			flush_packets(her, DELETE);
			delete_her(her);

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			printk(KERN_DEBUG "run_machine: unsupported action %d in state "
			                  "HOLD_INSERTED.\n",
			       action);
#endif
		}

		break;
	  case HOLD_DELETED:
		switch(action) {
		  case ADD_SA:
			ret = replace_er(her, er, SA_ESTABLISHED, store);
			break;
		  case ADD_TRAP:
			ret = replace_er(her, er, TRAP_INSERTED, store);
			break;
		  default:
			flush_packets(her, DELETE);
			delete_her(her);

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			printk(KERN_DEBUG "run_machine HOLD_DELETED: "
			                  "unsupported action %d.\n",
			       action);
#endif
		}

		break;
	  case SA_ESTABLISHED:
		switch(action) {
		  case ADD_TRAP:
			flush_packets(her, DELETE);
			ret = replace_er(her, er, TRAP_INSERTED, store);
			break;
		  case DEL_SA:
			if (her->pkgq) {
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
				printk(KERN_DEBUG "run_machine: "
				                  "Ooops, some packets ignored.\n");
#endif
				flush_packets(her, DELETE);
			}

			ret = replace_er(her, er, SA_DELETED, store);
			break;
		  default:
			flush_packets(her, DELETE);
			delete_her(her);

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			printk(KERN_DEBUG "run_machine: unsupported action %d in state "
			                  "SA_ESTABLISHED.\n",
			       action);
#endif
		}

		break;
	  case SA_DELETED:
		switch(action) {
		  case ADD_TRAP:
			ret = replace_er(her, er, TRAP_INSERTED, store);
			break;
		  case ADD_SA:
			ret = replace_er(her, er, SA_ESTABLISHED, store);
			break;
		  default:
			flush_packets(her, DELETE);
			delete_her(her);

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			printk(KERN_DEBUG "run_machine: unsupported action %d in state "
			                "SA_DELETED.\n",
			       action);
#endif
		}

		break;
	  default:
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
		printk(KERN_DEBUG "run_machine: unsupported state %d.\n", her->state);
#endif
		flush_packets(her, DELETE);
		delete_her(her);
	}

	return 0;
}


static
struct sk_buff *delete_oldest_pkg(struct sk_buff **queue, spinlock_t *lock)
{
	struct sk_buff *old_skb;
	
	spin_lock(lock);

	old_skb = *queue;

	if ((*queue) && (*queue)->next != *queue) {
		(*queue)->prev->next = (*queue)->next;
		(*queue)->next->prev = (*queue)->prev;

		(*queue) = (*queue)->next;
	} else
		*queue = NULL;

	spin_unlock(lock);

	hb_buf_size--;
	return old_skb;
}


static int check_for_bad_dev(struct net_device *dev)
{
	struct ipsecpriv *prv;

	if (!dev)
		return 1;

	prv = dev->priv;
	if (prv) {
		struct net_device *physdev;

		physdev = prv->dev;
		if (physdev) {
			if (!strncmp(physdev->name, "ppp", 3))
				return 1;

			if (!strncmp(physdev->name, "ippp", 4))
				return 1;
		}
	}

	return 0;
}

int ipsec_store_rx_pkg(struct sk_buff *skb, struct net_device *dev)
{
	if (check_for_bad_dev(dev))
		goto drop_out;

	if (IPSEC_SKB_CB(skb)->hb_retries < HOLD_BUF_RETRY) {
#ifdef CONFIG_IPSEC_PKT_SEQ
		struct sk_buff *skb_c;

		skb_c = skb_clone(skb, GFP_ATOMIC);
		if (skb_c) {
				IPSEC_SKB_CB(skb)->ignore = 1;
				hsips_ip_send(skb);

				memset(IPSEC_SKB_CB(skb_c), 0, sizeof(ipsec_cb_t));
				skb = skb_c;
		}
#endif

		IPSEC_SKB_CB(skb)->hb_retries++;

		if (hb_buf_size > HB_MAX_BUF_SIZE) {
			struct sk_buff *oskb;

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			printk(KERN_DEBUG "ipsec_store_rx_pkg: buffer queue is full!\n");
#endif
			oskb = delete_oldest_pkg(&tlet_rx_buf, &tlet_rx_buf_lock);
			if (!oskb)
				goto drop_out;

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
			printk("ipsec_store_rx_pkg: delete pkt 0x%08x.\n",
			       (unsigned int) oskb);
#endif

			kfree_skb(oskb);
		}
	
		spin_lock(&tlet_rx_buf_lock);

		IPSEC_SKB_CB(skb)->hb_dev = dev;

		if (!tlet_rx_buf) {
			skb->next = skb;
			skb->prev = skb;

			tlet_rx_buf = skb;
		} else {
			skb->next = tlet_rx_buf;
			skb->prev = tlet_rx_buf->prev;
	
			tlet_rx_buf->prev->next = skb;
			tlet_rx_buf->prev = skb;
		}

		hb_buf_size++;

		spin_unlock(&tlet_rx_buf_lock);
	
		return 0;
	}

drop_out:
	return 1;
}


int ipsec_store_tx_pkg(struct eroute *er, struct sk_buff *skb,
                       struct net_device *dev)
{
	int ret;

	ret = -ENOSYS;
	if (check_for_bad_dev(dev))
		goto drop_out;

	ret = -EBUSY;
	if (IPSEC_SKB_CB(skb)->hb_retries < HOLD_BUF_RETRY) {
		struct hold_eroute *her = NULL;

#ifdef CONFIG_IPSEC_PKT_SEQ
        struct sk_buff *skb_c;

        skb_c = skb_clone(skb, GFP_ATOMIC);
        if (skb_c) {
            IPSEC_SKB_CB(skb)->ignore = 1;
            hsips_ip_send(skb);

            memset(IPSEC_SKB_CB(skb_c), 0, sizeof(ipsec_cb_t));
            skb = skb_c;
        }
#endif

		IPSEC_SKB_CB(skb)->hb_retries++;

		ret = -EBADSLT;
		if (er) {
			struct hold_eroute *p = her_lst_achr;

			read_lock_bh(&her_lst_lock);
	
			for (her = NULL; p && !her; p = p->next) {
				spin_lock(&p->lock);
				if (!memcmp((char *) &er->er_eaddr, (char *) &p->er->er_eaddr,
				            2 * sizeof(struct sockaddr_encap)))
					her = p;
				spin_unlock(&p->lock);
			}

			read_unlock_bh(&her_lst_lock);
		}

		if (her) {
			if (hb_buf_size > HB_MAX_BUF_SIZE) {
				struct sk_buff *oskb;

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
				printk(KERN_DEBUG 
				       "ipsec_store_tx_pkg: buffer queue is full!\n");
#endif
				oskb = delete_oldest_pkg(&her->pkgq, &her->lock);
				if (!oskb)
					goto drop_out;

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
				printk(KERN_DEBUG "ipsec_store_tx_pkg: delete pkt 0x%08x.\n",
				       (unsigned int) oskb);
#endif

				kfree_skb(oskb);
			}

			spin_lock(&her->lock);

			IPSEC_SKB_CB(skb)->hb_dev = dev;

			hb_link_skb2queue(her, skb);

			spin_unlock(&her->lock);

			return 0;
		}
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
		else
			printk(KERN_DEBUG "ipsec_store_tx_pkg: no hold eroute found.\n");
#endif
	}
	
drop_out:
	return ret;
}

int hb_process_makeroute(struct eroute *er)
{
	struct eroute *hb_er;
	int err;

	if (!er) {
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
		printk(KERN_DEBUG "hb_process_makeroute: no eroute given.\n");
#endif
		return -EFAULT;
	}

	err = -ENOMEM;
	hb_er = (struct eroute *) kmalloc(sizeof(struct eroute), GFP_ATOMIC);
	if (hb_er) {
		int action;

		memcpy((char *) hb_er, (char *) er, sizeof(struct eroute));

		action = ADD_SA;
		if (hb_er->er_said.proto == IPPROTO_INT) {
			switch (ntohl(hb_er->er_said.spi)) {
			case SPI_TRAP:
				action = ADD_TRAP;
				break;
			case SPI_HOLD:
				action = ADD_HOLD;
				break;
			default:
				action = UNSUPPORTED;
#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
				printk(KERN_DEBUG "alloc_eroute_check: unsupported SPI %d.\n",
				       (unsigned int) ntohl(hb_er->er_said.spi));
#endif
			}
		}

		err = run_machine(hb_er, action, 1);
	}

	return err;
}
	

int hb_process_breakroute(struct sockaddr_encap *eaddr)
{
	struct eroute *er;

	er = ipsec_findroute(eaddr);
	if (er) {
		struct eroute *my_er;
		int action;

		my_er = (struct eroute *) kmalloc(sizeof(struct eroute), GFP_ATOMIC);
		if (my_er)
			memcpy((char *) my_er, (char *) er, sizeof(struct eroute));

		action = DEL_SA;
		if (er->er_said.proto == IPPROTO_INT) {
			switch (ntohl(er->er_said.spi)) {
			case SPI_TRAP:
				action = DEL_TRAP;
				break;
			case SPI_HOLD:
				action = DEL_HOLD;
				break;
			default:
				action = UNSUPPORTED;
			}
		}

		return run_machine((my_er ? my_er : er), action, (my_er ? 1 : 0));
	}

	return -ESRCH;
}


int flush_hold_buffers(struct eroute *er)
{
	hold_eroute_t *her;
	int ret;

	ret = -EINVAL;

	her = find_hold_eroute(er, 1);
	if (her && (her->state == SA_ESTABLISHED)) {
		switch (her->type) {
		  case RX_TYPE:
			flush_rx_buf(0);
			ret = 0;

			break;
		  case TX_TYPE:
			ret = flush_packets(her, SEND);
			break;
		  default:
			printk(KERN_NOTICE "flush_hold_buffers: unknown her type %d.\n",
			       her->type);
		}

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
		printk(KERN_DEBUG "flush_hold_buffers: hold buffer size %lu.\n",
		       hb_buf_size);
#endif
	}

	return ret;
}


void clear_hold_eroute(struct eroute *er)
{
	hold_eroute_t *her;

	her = find_hold_eroute(er, 1);
	if (her)
		delete_her(her);

	return;
}


int cleanup_hold_buffer(void)
{
	while (her_lst_achr)
		delete_her(her_lst_achr);

	spin_lock(&inuse_lock);
	in_use = 0;
	spin_unlock(&inuse_lock);

	return 0;
}


int init_hold_buffer(void)
{
	spin_lock(&inuse_lock);

	if (in_use) {
		spin_unlock(&inuse_lock);

#ifdef CONFIG_IPSEC_HOLD_BUFFER_DEBUG
		printk(KERN_DEBUG "init_hold_buffer: hold buffer already in use.\n");
#endif
		return -EBUSY;
	}

	in_use = 1;

	spin_unlock(&inuse_lock);

#ifdef CONFIG_SMP
	tasklet_init(&ipsec_hold_tx_buf_tlet, flush_tx_buf, 0);
	tasklet_init(&ipsec_hold_rx_buf_tlet, flush_rx_buf, 0);
#endif

	her_lst_achr = NULL;

	return 0;
}
