/*
 * filename: ipsec_frag.c
 *
 * Developed by secunet AG
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/ip.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>

#include <net/ip.h>

#include "ipsec_frag.h"
#include "ipsec_tunnel.h"
#include "ipsec_encap.h"
#include "ipsec_cb.h"

#ifdef CONFIG_IPSEC_PKT_SEQ
#  include "ipsec_threads/seq.h"
#endif

static int frag_on = 0;

static int proc_read_frag(char *page, char **start, off_t off, int count,
                          int *eof, void *data)
{
	return sprintf(page, "%d\n", frag_on);
}


static int proc_write_frag(struct file *file, const char *buffer,
                           unsigned long count, void *data)
{
	int len;

	len = count;

	if (len > 0) {
		if (buffer[0] == '1') {
#ifdef CONFIG_IPSEC_PKT_SEQ
			hsips_frag_on();
#endif
			frag_on = 1;
		}
          
		if (buffer[0] == '0') {
			frag_on = 0;
		}
	}

	return len;
}


static int get_random_mtu(int mtu, int left, int first_frag)
{

	int i;
	i = net_random() % (mtu-100);

	if (first_frag && (i>left))
		i = net_random() % left;

	if (i > left)
		i = left;

	return i;
}



static inline int __ipsec_fragment(struct sk_buff *skb, struct net_device *dev)
{
	struct iphdr *iph;
	int raw = 0;
	int ptr;
	struct sk_buff *skb2;
	unsigned int mtu, hlen, left, len; 
	int offset;
	int not_last_frag;
	int err = 0;
	int first_frag = 1;

	struct eroute *er;

	struct ipsecpriv *prv;
	struct net_device *physdev;

	er = IPSEC_SKB_CB(skb)->er;

	prv = dev->priv;
	if (!prv || (physdev = prv->dev) == NULL)
		return -EFAULT;

	iph = skb->nh.iph;

	hlen = iph->ihl << 2;
	left = skb->len - hlen;
	mtu = physdev->mtu - hlen;
	ptr = raw + hlen;

	offset = (ntohs(iph->frag_off) & IP_OFFSET) << 3;
	not_last_frag = iph->frag_off & htons(IP_MF);

	while (left > 0)	{
		len = get_random_mtu(mtu, left, first_frag);
		first_frag = 0;
		if (len > mtu)
			len = mtu;

		if (len < left)
			len &= ~7;

		skb2 = alloc_skb(len + hlen + dev->hard_header_len + 15, GFP_ATOMIC);
		if (!skb2) {
			printk(KERN_INFO "ipsec_fragment: "
			                 "frag: no memory for new fragment!\n");

			IP_INC_STATS(IpFragFails);
			return -ENOMEM;
		}

		skb2->pkt_type = skb->pkt_type;
		skb2->priority = skb->priority;
		skb_reserve(skb2, (dev->hard_header_len + 15) & ~15);
		skb_put(skb2, len + hlen);
		skb2->nh.raw = skb2->data;
		skb2->h.raw = skb2->data + hlen;
		skb2->protocol = skb->protocol;
		skb2->security = skb->security;

		if (skb->sk)
			skb_set_owner_w(skb2, skb->sk);
		skb2->dst = dst_clone(skb->dst);
		skb2->dev = skb->dev;

		memcpy(skb2->nh.raw, skb->data, hlen);

		if (skb_copy_bits(skb, ptr, skb2->h.raw, len))
			BUG();
		left -= len;

		iph = skb2->nh.iph;
		iph->frag_off = htons((offset >> 3));

		if (offset == 0)
			ip_options_fragment(skb);

		if (left > 0 || not_last_frag)
			iph->frag_off |= htons(IP_MF);
		ptr += len;
		offset += len;

#ifdef CONFIG_NET_SCHED
		skb2->tc_index = skb->tc_index;
#endif
#ifdef CONFIG_NETFILTER
		skb2->nfmark = skb->nfmark;
		skb2->nfct = skb->nfct;
		nf_conntrack_get(skb2->nfct);
#ifdef CONFIG_NETFILTER_DEBUG
		skb2->nf_debug = skb->nf_debug;
#endif
#endif

#if defined CONFIG_IPSEC_NF || defined CONFIG_CLOSEIPSEC
		skb2->ipsec_mark = skb->ipsec_mark;
#endif

		memcpy((char *) IPSEC_SKB_CB(skb2), (char *) IPSEC_SKB_CB(skb),
		       sizeof(ipsec_cb_t));
		er->busy++;

		IP_INC_STATS(IpFragCreates);

		iph->tot_len = htons(len + hlen);

		ip_send_check(iph);

		err = ipsec_encap_pkt(skb2, dev);
		if (err) {
			er->busy--;
			kfree_skb(skb2); 

			IP_INC_STATS(IpFragFails);
			return err;
		}
	}

	er->busy--;
	kfree_skb(skb);
	IP_INC_STATS(IpFragOKs);

	return err;
}


int ipsec_fragment(struct sk_buff *skb, struct net_device *dev)
{
	if (!frag_on)
		return ipsec_encap_pkt(skb, dev);

	return __ipsec_fragment(skb, dev);
}

static int ipsec_check_fragmentation(struct sk_buff *skb,
                                     struct net_device *dev)
{
	if (!frag_on)
		return 0;

	if (dev && strncmp(dev->name, "ipsec", 5))
			return 0;

	return 1;
}


int ipsec_frag_skip_hard_header(struct sk_buff *skb)
{
	struct iphdr *iph = skb->nh.iph;
	int hard_header_len;

	hard_header_len = (unsigned char *) iph - skb->data;
	if (hard_header_len < 0)
		return -EFAULT;

	if (hard_header_len) {
		skb_pull(skb, hard_header_len);
	}

	return 0;
}

extern int (*ipsec_skip_fragment) (struct sk_buff *, struct net_device *);


int ipsec_frag_init(void)
{
	struct proc_dir_entry *ipsec_frag;

	frag_on = 0;

	ipsec_frag = create_proc_entry("ipsec_frag", 0644, proc_net);
	ipsec_frag->nlink      = 1;
	ipsec_frag->read_proc  = proc_read_frag;
	ipsec_frag->write_proc = proc_write_frag;

	ipsec_skip_fragment = ipsec_check_fragmentation;

	return 0;
}


void ipsec_frag_cleanup(void)
{
	remove_proc_entry("ipsec_frag",proc_net);
	ipsec_skip_fragment = NULL;
}
