/*
 * filename: dispatcher.c
 *
 * Developed by secunet AG
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/config.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/list.h>
#include <linux/spinlock.h>

#include <asm/atomic.h>

#include "../ipsec_tunnel.h"
#include "../ipsec.h"
#include "../ipsec_encap.h"
#include "../ipsec_cb.h"

#include "hsips.h"

#ifdef CONFIG_IPSEC_PKT_SEQ
#  include "seq.h"
#endif


static hsips_queue_t ike_queue = { \
	queue: LIST_HEAD_INIT(ike_queue.queue), \
	lock: SPIN_LOCK_UNLOCKED, \
	size: ATOMIC_INIT(0) \
};

static hsips_queue_t data_queue = { \
	queue: LIST_HEAD_INIT(data_queue.queue), \
	lock: SPIN_LOCK_UNLOCKED, \
	size: ATOMIC_INIT(0) \
};

int hsips_get_qsize(void)
{
	return (int) (atomic_read(&ike_queue.size) + atomic_read(&data_queue.size));
}

struct sk_buff *hsips_get_packet(void)
{
	struct sk_buff *skb;
	hsips_queue_t *queue;
	unsigned long flags;

	queue = &ike_queue;

	spin_lock_irqsave(&queue->lock, flags);
	if (list_empty(&queue->queue)) {
		spin_unlock_irqrestore(&queue->lock, flags);

		queue = &data_queue;

		spin_lock_irqsave(&queue->lock, flags);
		if (list_empty(&queue->queue)) {
			spin_unlock_irqrestore(&queue->lock, flags);

			return NULL;
		}
	}

	skb = (struct sk_buff *) queue->queue.next;

	list_del((struct list_head *) skb);
	atomic_dec(&queue->size);

	spin_unlock_irqrestore(&queue->lock, flags);

#ifdef CONFIG_IPSEC_PKT_SEQ
    atomic_inc(&hsips_pks_in_process);
#endif

	return skb;
}

static inline int hsips_check_qsize(atomic_t *drop)
{
	int retval;

	retval = 0;
	if (respawn_thread.task) {
		retval = (hsips_get_qsize() > IPSEC_THREAD_QSIZE) ? 1 : 0;

#ifdef CONFIG_IPSEC_PKT_SEQ
		if (!retval)
			retval = ipsec_threads_check_seq_q();
#endif
	}

	if (retval)
		atomic_inc(drop);

	return retval;
}

int hsips_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	int retval;

	retval = -EINVAL;
	if (!(skb->data && !skb->list && !skb->prev && !skb->next)) {
		printk(KERN_DEBUG "hsips_start_xmit: packet is bogus.\n");
		goto out;
	}

	if (skb_cloned(skb) || atomic_read(&skb->users) != 1) {
		retval = -ENOMEM;
		if (skb_cow(skb, skb_headroom(skb)))
			goto out;
	}

	retval = -EBUSY;	
	if (IPSEC_SKB_CB(skb)->touched || !hsips_check_qsize(&respawn_thread.tx_drop)) {
		hsips_queue_t *queue;
		unsigned long flags;

		skb->list = (struct sk_buff_head *) dev;

#ifdef CONFIG_IPSEC_PKT_SEQ
		if (!IPSEC_SKB_CB(skb)->rearrange)
			hsips_inc_tx_seq(skb);
#endif

		IPSEC_SKB_CB(skb)->direc = HSIPS_TX;
		
		queue = &data_queue;
		if (0 ||
#ifdef CONFIG_NETFILTER
		    skb->nfmark == 1 ||
#endif
		    IPSEC_SKB_CB(skb)->ike_packet) {

			queue = &ike_queue;
		}

		spin_lock_irqsave(&queue->lock, flags);
		list_add_tail(((struct list_head *) skb), &queue->queue);
		atomic_inc(&queue->size);
		spin_unlock_irqrestore(&queue->lock, flags);

		activate_thread();

		retval = 0;
	}

out:
	if (retval)
		atomic_inc(&respawn_thread.tx_err);

	return retval;
}


int hsips_rcv(struct sk_buff *skb)
{
	int retval;

	retval = -EINVAL;
	if (!(skb->data && !skb->list && !skb->prev && !skb->next)) {
		printk(KERN_DEBUG "hsips_rcv: packet is bogus.\n");
		goto out;
	}

	if (skb_cloned(skb) || atomic_read(&skb->users) != 1) {
		retval = -ENOMEM;
		if (skb_cow(skb, skb_headroom(skb)))
			goto out;
	}

	retval = -EBUSY;	
	if (!hsips_check_qsize(&respawn_thread.rx_drop)) {
		hsips_queue_t *queue;
		unsigned long flags;

#ifdef CONFIG_IPSEC_PKT_SEQ
		if (!IPSEC_SKB_CB(skb)->rearrange)
			hsips_inc_rx_seq(skb);
#endif

		IPSEC_SKB_CB(skb)->direc = HSIPS_RX;
		
		queue = &data_queue;
		if (0 ||
#ifdef CONFIG_NETFILTER
		    skb->nfmark == 1 ||
#endif
		    IPSEC_SKB_CB(skb)->ike_packet) {

			queue = &ike_queue;
		}

		spin_lock_irqsave(&queue->lock, flags);
		list_add_tail(((struct list_head *) skb), &queue->queue);
		atomic_inc(&queue->size);
		spin_unlock_irqrestore(&queue->lock, flags);

		activate_thread();

		retval = 0;
	}

out:
	if (retval) {
		atomic_inc(&respawn_thread.rx_err);
		kfree_skb(skb);
	}

	return retval;
}
