/*
 * filename: hsips.c
 *
 * Developed by secunet AG
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/proc_fs.h>

#ifdef CONFIG_IPSECTHREAD_DEV
#  include "thread_dev.h"
#endif

#include "../ipsec_tunnel.h"
#include "../ipsec.h"
#include "hsips.h"

#ifdef CONFIG_IPSEC_PKT_SEQ
#  include "seq.h"
#endif

struct respawn_status respawn_thread;

#define HWPEPP_MODNAME	"hwpepp"

extern struct module *module_list;

static inline struct module *find_hwpepp_mod(void) 
{
	struct module *mod;

	for (mod = module_list; mod; mod = mod->next) {
		if (mod->flags & MOD_DELETED)
			continue;
		if (!strcmp(mod->name, HWPEPP_MODNAME))
			break;
	}
	
	return mod;
}


static inline void hsips_set_thread_cnt(struct respawn_status *rt)
{
	unsigned int cnt;

	cnt = smp_num_cpus;
	if (find_hwpepp_mod()) {
		cnt = 32;
		printk("hsips_set_thread_cnt: perfopt: count = %d\n", cnt);
	}

	if (IPSEC_MAX_THREADS < cnt) {
		printk("hsips_set_thread_cnt: thread count reduced from %d to %d!\n",
		       cnt, IPSEC_MAX_THREADS);
		cnt = IPSEC_MAX_THREADS;
	}

#if CONFIG_IPSECTHREAD_MAX_THREADS
	if (CONFIG_IPSECTHREAD_MAX_THREADS < cnt)
		cnt = CONFIG_IPSECTHREAD_MAX_THREADS;
#endif

	rt->num = cnt;
}

static int ipsec_threads_get_info(char *buffer, char **start, off_t offset,
                                  int length)
{
	struct list_head *le;
	int len = 0;
	off_t begin = 0;

	if (!respawn_thread.task) {
		len += sprintf(buffer + len, "no status info available!\n");
		goto out;
	}
	
	list_for_each(le, &hsips_thread_list) {
		struct ipst_status *thread;

		thread = list_entry(le, struct ipst_status, list);
		if (!thread->task)
			continue;

		spin_lock(&thread->lock);

		len += sprintf(buffer + len, 
		               "#%d(%d): pid=%d cnt=%d up=%d dn=%d sig=%d pkts=%lu\n",
		               thread->num, thread->task->processor, thread->pid,
		               atomic_read(&thread->count), atomic_read(&thread->up),
		               atomic_read(&thread->dn), atomic_read(&thread->sig),
					   pkts_per_run_array[thread->num]);
							       
		spin_unlock(&thread->lock);
	}

	len += sprintf(buffer + len, "rx_err=%d rx_drop=%d\n",
	               atomic_read(&respawn_thread.rx_err),
	               atomic_read(&respawn_thread.rx_drop));
	len += sprintf(buffer + len, "tx_err=%d tx_drop=%d\n",
	               atomic_read(&respawn_thread.tx_err),
	               atomic_read(&respawn_thread.tx_drop));
	len += sprintf(buffer + len, "qsize=%d\n", hsips_get_qsize());

out:	
	*start = buffer + (offset - begin);
	len -= (offset - begin);
	if (len > length)
		len = length;
	return len;
}

int hsips_get_thread_cnt(void)
{
	return respawn_thread.num;
}

int hsips_init(void)
{
	int retval;

	memset((char *) pkts_per_run_array, 0, 32 * sizeof(long));

	retval = 0;

#ifdef CONFIG_IPSECTHREAD_DEV
	retval = register_hsipsthreaddev();
#else
	if (!retval) {
		memset((char *) &respawn_thread, 0, sizeof(struct respawn_status));
		hsips_set_thread_cnt(&respawn_thread);

		retval = kernel_thread(hsips_respawn, (void *) &respawn_thread, 0);
		if (!(retval < 0)) {
			proc_net_create("ipsec_threads", 0, ipsec_threads_get_info);
			retval = 0;
		} else
			respawn_thread.num = 0;
	}
#endif

#ifdef CONFIG_IPSEC_PKT_SEQ
	if (!retval) {
		hsips_seq_init();
	}
#endif

	printk("HSIPS: started with %d threads\n", respawn_thread.num);

	return retval;
}


int hsips_cleanup(void)
{
	int retval;

	respawn_thread.num = 0;

#ifdef CONFIG_IPSEC_PKT_SEQ
	hsips_seq_cleanup();
#endif

#ifndef CONFIG_IPSECTHREAD_DEV
	proc_net_remove("ipsec_threads");
#endif
	
	retval = hsips_destroy(&respawn_thread);

#ifdef CONFIG_IPSECTHREAD_DEV
	retval |= unregister_hsipsthreaddev();
#endif

	return retval;
}

