/*
 * filename: respawn.c
 *
 * Developed by secunet AG
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#define __KERNEL_SYSCALLS__

#include <asm/atomic.h>

#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/smp.h>
#include <linux/unistd.h>
#include <linux/module.h>

#include "hsips.h"
#include "../ipsec_encap.h"
#include "../ipsec_cb.h"
#include "../ipsec_tunnel.h"
#include "../ipsec.h"

#ifdef CONFIG_IPSEC_PKT_SEQ
#  include "seq.h"
#endif

static struct semaphore sync_sem;
static atomic_t del_threads;

DECLARE_WAIT_QUEUE_HEAD(respawn_wait);
struct list_head hsips_thread_list = LIST_HEAD_INIT(hsips_thread_list);

extern int ipsec_rcv(struct sk_buff *);
extern void free_eroute(struct eroute *er);

#define THREAD_NUM 2

static hsips_queue_t thread_idle_list = { \
	queue: LIST_HEAD_INIT(thread_idle_list.queue), \
	lock: SPIN_LOCK_UNLOCKED, \
	size: ATOMIC_INIT(0) \
};

static void remove_from_idle_list(struct ipst_status *status)
{
	struct list_head *le;
	unsigned long flags;

	spin_lock_irqsave(&thread_idle_list.lock, flags);

	list_for_each(le, &thread_idle_list.queue) {
		struct ipst_status *thread;

		thread = list_entry(le, struct ipst_status, idle_list);
		if (thread->pid == status->pid) {
			list_del(&thread->idle_list);
			break;
		}
	}

	spin_unlock_irqrestore(&thread_idle_list.lock, flags);
}

void activate_thread(void)
{
	struct ipst_status *thread;
	unsigned long flags;

	spin_lock_irqsave(&thread_idle_list.lock, flags);

	thread = NULL;
	if (!list_empty(&thread_idle_list.queue)) {
		thread = list_entry(thread_idle_list.queue.next, struct ipst_status,
		                    idle_list);
		list_del(&thread->idle_list);
	}

	spin_unlock_irqrestore(&thread_idle_list.lock, flags);

	if (thread) {
		up(&thread->sem);
		atomic_inc(&thread->up);

		atomic_dec(&thread_idle_list.size);
	}
}

static int deactivate_thread(struct ipst_status *thread)
{
	unsigned long flags;
	int err;

	spin_lock_irqsave(&thread_idle_list.lock, flags);
	list_add_tail(&thread->idle_list, &thread_idle_list.queue);
	spin_unlock_irqrestore(&thread_idle_list.lock, flags);

	atomic_inc(&thread_idle_list.size);
	
	err = down_interruptible(&thread->sem);
	if (!err)
		atomic_inc(&thread->dn);

	return err;
}

static int ipst_destroy(struct ipst_status *status)
{
	if (!status)
		return -EFAULT;

	remove_from_idle_list(status);

	spin_lock(&status->lock);

	if (status->task && status->task->pid == status->pid)
		send_sig(SIGTERM, status->task, 1);

	atomic_set(&status->state, IPST_UNDEF);

	spin_unlock(&status->lock);

	remove_from_idle_list(status);

	return 0;
}

static int process_packet(struct sk_buff *skb, int thread_num)
{
	skb->next = NULL;
	skb->prev = NULL;

	IPSEC_SKB_CB(skb)->hsips_hid = thread_num;

	if (IPSEC_SKB_CB(skb)->direc == HSIPS_TX) {
		struct net_device *dev;
		
		dev = (struct net_device *) skb->list;
		skb->list = NULL;

		return ipsec_start_xmit_entry2(skb, dev);
	}

	skb->list = NULL;

	return ipsec_rcv(skb);
}

static int initialize_signals(void)
{
	spin_lock_irq(&current->sigmask_lock);
	
	sigfillset(&current->blocked);
	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGTERM) | 
			sigmask(SIGSTOP) | sigmask(SIGCONT)); 
	recalc_sigpending(current);

	spin_unlock_irq(&current->sigmask_lock);

	return 0;
}

static int handle_thread_signals(void)
{
	int term, stopped;

	term = stopped = 0;

	KLIPS_PRINT(debug_netlink,
	            KERN_NOTICE "HSIPS(%d): Ring Ring - signal received.\n",
	            current->pid);
	printk("HSIPS(%d): Ring Ring - signal received.\n", current->pid);

	spin_lock_irq(&current->sigmask_lock);

	if (sigismember(&current->pending.signal, SIGKILL)) {
		sigdelset(&current->pending.signal, SIGKILL);
		term = 1;
	}
	if (sigismember(&current->pending.signal, SIGTERM)) {
		sigdelset(&current->pending.signal, SIGTERM);
		term = 1;
	}
	if (sigismember(&current->pending.signal, SIGSTOP)) {
		sigdelset(&current->pending.signal, SIGSTOP);
		stopped = 1;
	} 

	recalc_sigpending(current);

	spin_unlock_irq(&current->sigmask_lock);

	if (term) {
		KLIPS_PRINT(debug_netlink,
		            KERN_NOTICE "HSIPS(%d): thread has been terminated.\n",
		            current->pid);
		printk("HSIPS(%d): thread has been terminated.\n", current->pid);

		return 1;
	}

	if (stopped) {
		KLIPS_PRINT(debug_netlink,
		            KERN_NOTICE "HSIPS(%d): thread has been stopped.\n",
		            current->pid);
		printk("HSIPS(%d): thread has been stopped.\n", current->pid);

		current->state = TASK_STOPPED;

		schedule();

		KLIPS_PRINT(debug_netlink,
		            KERN_NOTICE "HSIPS(%d): thread has been restarted.\n",
		            current->pid);
		if (sigismember(&current->pending.signal, SIGCONT)) {
			printk("HSIPS(%d): thread has been restarted: sigdelset(SIGCONT).\n", current->pid);
			sigdelset(&current->pending.signal, SIGCONT);
		} 
		printk("HSIPS(%d): thread has been restarted: final.\n", current->pid);
	}

	return 0;
}

unsigned long pkts_per_run_array[IPSEC_MAX_THREADS];

static int ipst_init(void *status_p)
{
	struct ipst_status *status;
	int err;

	status = (struct ipst_status *) status_p;
	if (!status)
		return -EFAULT;
	
	daemonize();

	printk(KERN_DEBUG "ipst_init(%d): starting on cpu(%d) ...\n",
	       current->pid, current->processor);

	sprintf(current->comm, "hsips_thread");

	spin_lock(&status->lock);

	status->pid = current->pid;
	status->task = current;
	status->task->cpus_allowed = 1 << (status->num % smp_num_cpus);

	status->destroy = ipst_destroy;
	status->ctl = NULL;

	atomic_set(&status->state, IPST_RUN);

	spin_unlock(&status->lock);

	initialize_signals();

	for (;;) {
		struct sk_buff *skb;
		unsigned long pkts_per_run;

		pkts_per_run = 0;
		do {
			skb = hsips_get_packet();
			if (!skb) {
				if (pkts_per_run_array[status->num] < pkts_per_run)
					pkts_per_run_array[status->num] = pkts_per_run;

				pkts_per_run = 0;

				err = deactivate_thread(status);
				if (err) {
					KLIPS_PRINT(debug_netlink,
					            "HSIPS(%d): operation interrupted by signal. err=%d\n",
					            status->num, err);
					atomic_inc(&status->sig);

					break;
				}

				continue;
			}

			pkts_per_run++;

			atomic_inc(&status->count);

			err = process_packet(skb, status->num);
			if (err) {
				KLIPS_PRINT(debug_netlink,
				            "HSIPS(%d): error while processing skb.\n",
							current->pid);

				err = 0;
			}

			if (current->need_resched) {
				if (pkts_per_run_array[status->num] < pkts_per_run)
					pkts_per_run_array[status->num] = pkts_per_run;

				pkts_per_run = 0;
				schedule();
			}
		} while (!signal_pending(current));

		if (signal_pending(current)) {
			if (handle_thread_signals())
				break;
		}
	}
			
	atomic_set(&status->state, IPST_UNDEF);
	KLIPS_PRINT(debug_netlink,
	            KERN_DEBUG "HSIPS(%d): finished on cpu(%d) !\n",
	            current->pid, current->processor);

	if (atomic_read(&del_threads))
		up(&sync_sem);

	return 0;
}

static unsigned long create_work_threads(int num)
{
	struct ipst_status *thread;
	unsigned long mask;
	int i;

	mask = 0;

	if (num > 32) {
		printk(KERN_WARNING "HSIPS: can not handle more than 32 threads.\n");
		return 0;
	}

	for (i = 0; i < num; i++) {
		thread = kmalloc(sizeof(struct ipst_status), GFP_KERNEL);
		if (!thread)
			return -EINVAL;

		memset((char *) thread, 0, sizeof(struct ipst_status));

		mask |= 1 << i;

		atomic_set(&thread->state, IPST_UNDEF);
		thread->init = ipst_init;
		thread->num = i;

		list_add(&thread->list, &hsips_thread_list);
	}	

	return mask;
}

int hsips_destroy(struct respawn_status *status)
{
	if (!status)
		return EFAULT;

	atomic_set(&del_threads, 1);

	printk(KERN_DEBUG "hsips_destroy: killing ipsec threads .. \n");

	while (!list_empty(&hsips_thread_list)) {
		struct ipst_status *thread;
		int num;

		thread = (struct ipst_status *) hsips_thread_list.next;
		num = thread->num;
					
		if ((atomic_read(&thread->state) == IPST_RUN) && thread->destroy) {
			thread->destroy(thread);

			printk(KERN_DEBUG
			       "HSIPS: waiting for child #%d to finish termination\n", num);

			down(&sync_sem);
		}

		list_del(&thread->list);
		kfree(thread);
	}

	if (status->task) {
		printk(KERN_DEBUG "HSIPS: killing respawn thread .. \n");

		if (status->task->pid == status->pid) {
			DECLARE_WAIT_QUEUE_HEAD(wq);

			if (!send_sig(SIGTERM, status->task, 1))
				while (status->task) {
					interruptible_sleep_on_timeout(&wq, 10);
				}
		} else
			printk(KERN_NOTICE "HSIPS: lost respawn task. \n");
	}
	
	return 0;
}

int hsips_respawn(void *status_p)
{
	struct respawn_status *status = status_p;

	unsigned long reinit_mask;
	int retval;

	if (!status || status->task)
		return -EFAULT;

	if (!status->num)
		return -EINVAL;

	daemonize();

	sprintf(current->comm, "hsips_respawn");

	printk(KERN_DEBUG "hsips_respawn(%d): starting on cpu(%d) ...\n",
	       current->pid, current->processor);
	
	status->pid = current->pid;
	status->task = current;

	reinit_mask = create_work_threads(status->num);

	retval = -ENOMEM;
	if (!reinit_mask)
		goto cleanup_out;

	initialize_signals();

	sema_init(&sync_sem, 0);
	atomic_set(&del_threads, 0);

	for (;;) {
		struct list_head *le;

		list_for_each(le, &hsips_thread_list) {
			struct ipst_status *thread;

			thread = list_entry(le, struct ipst_status, list);
			if (thread->pid != 0) {
				if (waitpid(thread->pid, NULL, WNOHANG) == thread->pid) {
					printk(KERN_DEBUG "hsips_respawn: found dead child %d\n",
					       thread->pid);
					reinit_mask |= 1 << thread->num;
				}
			}

			if (!(reinit_mask & (1 << thread->num)))
				continue;

			if (!thread->task) {
				printk(KERN_DEBUG "hsips_respawn: "
				                  "initial setup for child no. %d (0x%08x)\n",
				       thread->num, (unsigned int) thread);

				atomic_set(&thread->state, IPST_INIT);
				spin_lock_init(&thread->lock);
				sema_init(&thread->sem, 0);
			}

			thread->pid = kernel_thread(thread->init, (void *) thread, SIGCHLD);
			if (thread->pid < 0) {
				printk(KERN_NOTICE "hsips_respawn: thread create failed. \n");

				atomic_set(&thread->state, IPST_UNDEF);
				thread->pid = 0;

				continue;
			}

			reinit_mask &= ~(1 << thread->num);
		}

		interruptible_sleep_on_timeout(&respawn_wait, 1000);

		if (signal_pending(current)) {
			printk(KERN_DEBUG "hsips_respawn: Ring Ring - signal received\n");
			break;        
		}
	}

cleanup_out:
	status->pid = 0;
	status->task = NULL;

	hsips_destroy(status);

	return retval;
}
