/*
 * filename: seq.c
 *
 * Developed by secunet AG
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/skbuff.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>

#include <net/ip.h>

#include "seq.h"
#include "../ipsec_cb.h"

static int seq_rx_active = 1;
static int seq_tx_active = 1;

atomic_t hsips_pks_in_process;

#define SEQ_INIT 0x0000
#define TLET_MAX_TRY 5
#define TLET_RX 0
#define TLET_TX 1

static char version[] = "SINA IPSEC hsips seq.c: add MW: add cnt dropped packets BR: add STOP: add deadlock BR: add reset BR";

static unsigned long hsips_tlet_flush_buf_lock_max = 0; 
static unsigned long hsips_tlet_flush_buf_lock_count = 0;
static unsigned long hsips_flush_all_packets_lock_max = 0;
static unsigned long hsips_flush_all_packets_lock_count = 0;
static unsigned long hsips_seq_timer_expiry_lock_max = 0;
static unsigned long hsips_seq_timer_expiry_lock_count = 0;

typedef struct hsips_seq_op_t {
	char name[5];
	unsigned int type;
	struct tasklet_struct tlet;

	spinlock_t lock;
	struct sk_buff *tlet_queue;
	unsigned long max_tlet_qsize;
	unsigned long tlet_qsize;
	unsigned long tlet_try_cnt;

	int qsize;
	struct sk_buff *rearrange_queue;

	unsigned long out_of_order_evt;
	unsigned long in_order_evt;
	unsigned long in_order_pkt_array[32];

	unsigned short last;
	unsigned short curr;

	struct timer_list timer;
	unsigned long timeout;
	unsigned long max_timeout;
	unsigned long timer_evt;
	unsigned long timer_evt_max_qsize;

	unsigned long pkt_dropped;
	unsigned long max_dropped;
	unsigned long pkt_delayed;
	unsigned long pkt_flushed;

	unsigned long flush_cnt;

	unsigned long out_of_reset;
	unsigned long reset;

	int (*send)(struct sk_buff *);
} hsips_seq_op_s;

static hsips_seq_op_s hsips_seq_rx = {
	name: "rx",
	type: TLET_RX, \
	lock: SPIN_LOCK_UNLOCKED, \
	tlet_queue: NULL, \
	max_tlet_qsize: 0, \
	tlet_qsize: 0, \
	qsize: 0, \
	rearrange_queue: NULL, \
	out_of_order_evt: 0, \
	in_order_evt: 0, \
	last: 0x0000, \
	curr: SEQ_INIT, \
	timeout: 0, \
	max_timeout: 0, \
	timer_evt: 0, \
	timer_evt_max_qsize: 0, \
	pkt_dropped: 0, \
	max_dropped: 0, \
	pkt_delayed: 0, \
	pkt_flushed: 0, \
	flush_cnt: 0, \
	out_of_reset: 0, \
	reset: 0, \
	send: netif_rx \
};

static hsips_seq_op_s hsips_seq_tx = {
	name: "tx",
	type: TLET_TX, \
	lock: SPIN_LOCK_UNLOCKED, \
	tlet_queue: NULL, \
	max_tlet_qsize: 0, \
	tlet_qsize: 0, \
	qsize: 0, \
	rearrange_queue: NULL, \
	out_of_order_evt: 0, \
	in_order_evt: 0, \
	last: 0x0000, \
	curr: SEQ_INIT, \
	timeout: 0, \
	max_timeout: 0, \
	timer_evt: 0, \
	timer_evt_max_qsize: 0, \
	pkt_dropped: 0, \
	max_dropped: 0, \
	pkt_delayed: 0, \
	pkt_flushed: 0, \
	flush_cnt: 0, \
	out_of_reset: 0, \
	reset: 0, \
	send: ip_send \
};


static spinlock_t hsips_seq_lock = SPIN_LOCK_UNLOCKED;

static atomic_t ipsec_thread_seq_q;

static void inc_seq(hsips_seq_op_s *seq_op, struct sk_buff *skb)
{
	IPSEC_SKB_CB(skb)->stamp = jiffies;
	IPSEC_SKB_CB(skb)->rearrange = 1;

	spin_lock(&hsips_seq_lock);
	
	if (!IPSEC_SKB_CB(skb)->processed) {
		IPSEC_SKB_CB(skb)->hsips_seq = (unsigned long) ++seq_op->curr;
		IPSEC_SKB_CB(skb)->processed = 1;
	}

	spin_unlock(&hsips_seq_lock);

	return;
}


#define LOW16(a) ((unsigned short) (a))

static inline int compare_seq(unsigned short new, unsigned short cur,
                              unsigned short buf_size)
{
	if (new < cur) {
		if ((new + buf_size) < cur)
			return 0;
		
		return 1;
	}
	
	if ((new - buf_size) > cur)
		return 1;

	return 0;
}


static inline void queue_pkt(struct sk_buff **queue, struct sk_buff *skb)
{
	struct sk_buff *prev, *cur = *queue;

	prev = NULL;

	while (!compare_seq(LOW16(IPSEC_SKB_CB(skb)->hsips_seq),
	                    LOW16(IPSEC_SKB_CB(cur)->hsips_seq),
	                    IPSEC_THREAD_MAX_RANGE)) {

		if (!cur->next) {
			skb->next = NULL;
			cur->next = skb;
			return;
		}

		prev = cur;
		cur = cur->next;
	}

	if (prev) {
		skb->next = prev->next;
		prev->next = skb;

		return;
	}

	skb->next = *queue;
	*queue = skb;

	return;
}


static void hsips_tlet_flush_buf(unsigned long data)
{
	hsips_seq_op_s *seq_op = (hsips_seq_op_s *) data;
	struct sk_buff *queue;
	int ret = 0;
	int dropped = 0;
	int dropped_cnt = 0;
	long qsize = 0;
	unsigned int direc = 0;

	if (!spin_trylock(&seq_op->lock)) {

		local_bh_disable();
		tasklet_schedule(&seq_op->tlet);
		local_bh_enable();

		hsips_tlet_flush_buf_lock_count++;

		return;
	}

	if (hsips_tlet_flush_buf_lock_max < hsips_tlet_flush_buf_lock_count)
		hsips_tlet_flush_buf_lock_max = hsips_tlet_flush_buf_lock_count;
	hsips_tlet_flush_buf_lock_count = 0;

	queue = seq_op->tlet_queue;
	seq_op->tlet_queue = NULL;

	qsize = seq_op->tlet_qsize;
	seq_op->tlet_qsize = 0;
	direc = seq_op->type;

	spin_unlock(&seq_op->lock);

	if (queue) {
		struct sk_buff *skb;
		unsigned long pkts_per_run = 0;

		queue->prev->next = NULL;

		for (skb = queue; skb; skb = queue) {
			pkts_per_run++;
			
			queue = queue->next;

			skb->next = NULL;
			skb->prev = NULL;

			qsize--;

			memset(IPSEC_SKB_CB(skb), 0, sizeof(ipsec_cb_t));
			ret = seq_op->send(skb);

			if (TLET_RX == direc) {
				if (NET_RX_DROP == ret) {
					dropped = 1;
					dropped_cnt++;
				}
			} else {
				if (ret < 0) {
					dropped = 1;
					dropped_cnt++;
				}
			}
		}

		if (seq_op->max_tlet_qsize < pkts_per_run)
			seq_op->max_tlet_qsize = pkts_per_run;

		if (dropped) {
			seq_op->pkt_dropped = seq_op->pkt_dropped + dropped_cnt;

			if (seq_op->max_dropped < dropped_cnt) {
				seq_op->max_dropped = dropped_cnt;
			}
		} 
	}

	return;
}


static int check_seq(struct sk_buff *skb, hsips_seq_op_s *seq_op)
{
	int ret = 0;
	int out_of_reset = 0;
	unsigned long pkts_per_run = 0;

	atomic_dec(&hsips_pks_in_process);

	if (IPSEC_SKB_CB(skb)->bypass || IPSEC_SKB_CB(skb)->ike_packet) {
		struct sk_buff *skb_c;

		skb_c = skb_clone(skb, GFP_ATOMIC);
		if (skb_c) {
			memset(IPSEC_SKB_CB(skb_c), 0, sizeof(ipsec_cb_t));
			seq_op->send(skb_c);
		}

		IPSEC_SKB_CB(skb)->bypass = 0;
		IPSEC_SKB_CB(skb)->ignore = 1;
	}

	spin_lock(&seq_op->lock);

	if (LOW16(IPSEC_SKB_CB(skb)->hsips_seq) != LOW16(seq_op->last + 1)) {
		unsigned long process_time;

		if (compare_seq(LOW16(IPSEC_SKB_CB(skb)->hsips_seq),
			LOW16(seq_op->last + 1),
			IPSEC_THREAD_MAX_RANGE)) {
			out_of_reset = 1;
			goto tlet_processing;
		};

		seq_op->out_of_order_evt++;

		if (!seq_op->rearrange_queue) {
			skb->next = NULL;
			skb->prev = NULL;

			seq_op->rearrange_queue = skb;
		} else
			queue_pkt(&seq_op->rearrange_queue, skb);

		seq_op->qsize++;

		process_time = jiffies - IPSEC_SKB_CB(skb)->stamp;
		if (!process_time)
			process_time = 100;

		if (seq_op->rearrange_queue) {
			unsigned long expires;
				
			expires = IPSEC_SKB_CB(seq_op->rearrange_queue)->stamp;

			process_time = process_time << 1;
			if (!seq_op->timeout) {
				seq_op->timeout = process_time;

				seq_op->timer.expires = expires + seq_op->timeout;

				add_timer(&seq_op->timer);
			} else {
				seq_op->timeout = (7 * seq_op->timeout + process_time) >> 3;

				expires += seq_op->timeout;

				mod_timer(&seq_op->timer, expires);
			}
			if (seq_op->max_timeout < seq_op->timeout)
				seq_op->max_timeout = seq_op->timeout;
		}

		goto unlock_out;
	}

tlet_processing:

	skb->next = NULL;
	skb->prev = NULL;

	if (out_of_reset)
		seq_op->out_of_reset++;
	else
		seq_op->in_order_evt++;

	do {
		atomic_inc(&ipsec_thread_seq_q);

		pkts_per_run++;

		ret = 0;
		if (!IPSEC_SKB_CB(skb)->ignore) {
			if (seq_op->tlet_queue) {
				skb->next = seq_op->tlet_queue;
				skb->prev = seq_op->tlet_queue->prev;

				seq_op->tlet_queue->prev->next = skb;
				seq_op->tlet_queue->prev = skb;
			} else {
				skb->next = skb;
				skb->prev = skb;

				seq_op->tlet_queue = skb;
			}
			seq_op->tlet_qsize++;
		} else {
			kfree_skb(skb);
		}

		if (out_of_reset)
			break;

		seq_op->last++;

		skb = seq_op->rearrange_queue;
		if (skb) {
			if (LOW16(IPSEC_SKB_CB(skb)->hsips_seq) != LOW16(seq_op->last + 1))
				break;

			seq_op->rearrange_queue = skb->next;
			seq_op->qsize--;
	
			skb->next = NULL;
			skb->prev = NULL;
		}
	} while (skb);

	if (pkts_per_run > 31)
		seq_op->in_order_pkt_array[0]++;
	else 
		seq_op->in_order_pkt_array[pkts_per_run]++;

unlock_out:
	if (!seq_op->rearrange_queue && seq_op->timeout)
		del_timer(&seq_op->timer);

	spin_unlock(&seq_op->lock);

	tasklet_schedule(&seq_op->tlet);

	return ret;
}


static int flush_all_packets(hsips_seq_op_s *seq_op)
{
	struct sk_buff *skb;

	if (!spin_trylock(&seq_op->lock)) {

		hsips_flush_all_packets_lock_count++;

		return 0;
	}

	if (hsips_flush_all_packets_lock_max < hsips_flush_all_packets_lock_count)
		hsips_flush_all_packets_lock_max = hsips_flush_all_packets_lock_count;
	hsips_flush_all_packets_lock_count = 0;

	seq_op->timer_evt++;
	
	if (seq_op->timer_evt_max_qsize < seq_op->qsize)
		seq_op->timer_evt_max_qsize = seq_op->qsize;

	while (seq_op->rearrange_queue) {
		skb = seq_op->rearrange_queue;
		seq_op->rearrange_queue = skb->next;

		seq_op->qsize--;

		seq_op->last = LOW16(IPSEC_SKB_CB(skb)->hsips_seq);
		atomic_inc(&ipsec_thread_seq_q);

		if (IPSEC_SKB_CB(skb)->ignore) {
			kfree_skb(skb);
			continue;
		}

		seq_op->tlet_qsize++;
		if (seq_op->tlet_queue) {
			skb->next = seq_op->tlet_queue;
			skb->prev = seq_op->tlet_queue->prev;

			seq_op->tlet_queue->prev->next = skb;
			seq_op->tlet_queue->prev = skb;

			continue;
		}

		skb->next = skb;
		skb->prev = skb;

		seq_op->tlet_queue = skb;
	}

	if (seq_op->tlet_queue)
		tasklet_schedule(&seq_op->tlet);

	seq_op->reset = jiffies;

	seq_op->timeout = 0;
	del_timer(&seq_op->timer);

	spin_unlock(&seq_op->lock);

	return 1;
}


static void seq_reset(void)
{
	spin_lock(&hsips_seq_lock);

	flush_all_packets(&hsips_seq_rx);
	flush_all_packets(&hsips_seq_tx);

	atomic_set(&ipsec_thread_seq_q, IPSEC_THREAD_MAX_BUF);

	spin_unlock(&hsips_seq_lock);

	return;
}


static void seq_timer_expiry(unsigned long data)
{
	hsips_seq_op_s *seq_op = (hsips_seq_op_s *) data;
	int ret = 0;

	if (!spin_trylock(&hsips_seq_lock)) {

		hsips_seq_timer_expiry_lock_count++;

		goto seq_timer_expiry_err_out;
	}
	if (hsips_seq_timer_expiry_lock_max < hsips_seq_timer_expiry_lock_count)
		hsips_seq_timer_expiry_lock_max = hsips_seq_timer_expiry_lock_count;
	hsips_seq_timer_expiry_lock_count = 0;

	ret = flush_all_packets(seq_op);

	spin_unlock(&hsips_seq_lock);

	if (0 != ret) {
		return;
	};

seq_timer_expiry_err_out:
	mod_timer(&seq_op->timer, jiffies + 2);

	return;
}

#ifdef CONFIG_SYSCTL

#include "../ipsec_sysctl.h"

int hsips_seq_var;

int hsips_seq_sysctl_strategy(ctl_table *ctl, int write, struct file *filp,
                              void *buffer, size_t *lenp)
{
	int ret, reset = 0;

	if (!ctl)
		return -EFAULT;

	if ((strlen(ctl->procname) == 9) && !strncmp(ctl->procname, "seq_reset", 9))
		reset = 1;

	if (!write) {
		if (reset) 
			return -ENOSYS;

		hsips_seq_var = 0;

		if (!strncmp(ctl->procname, "tx_qsize", 6))
			hsips_seq_var = hsips_seq_tx.qsize;
		if (!strncmp(ctl->procname, "rx_qsize", 6))
			hsips_seq_var = hsips_seq_rx.qsize;
		if (!strncmp(ctl->procname, "seq_tx_enable", 6))
			hsips_seq_var = seq_rx_active;
		if (!strncmp(ctl->procname, "seq_rx_enable", 6))
			hsips_seq_var = seq_tx_active;

		ret = proc_dointvec(ctl, write, filp, buffer, lenp);
		return ret;
	}

	ret = proc_dointvec(ctl, write, filp, buffer, lenp);
	if (!ret) {
		int set;

		if (!strncmp(ctl->procname, "tx_qsize", 6) || !strncmp(ctl->procname, "rx_qsize", 6))
			return -EPERM;

		if (reset) {
			hsips_seq_var = 0;
			seq_reset();

			return 0;
		}

		ret = 0;

		set = hsips_seq_var ? 1 : 0;

		if (!strncmp(ctl->procname, "seq_rx_enable", 6)) {
			if (seq_rx_active)
				seq_rx_active = set;
		} else {
			if (seq_tx_active)
				seq_tx_active = set;
		}

		hsips_seq_var = 0;
	}
			
	return ret;
}


#ifdef CONFIG_IPSEC_PKT_SEQ
#  undef  IPSEC_PROTO_DIR_ENTRIES
#  define IPSEC_PROTO_DIR_ENTRIES 3
#endif

static struct hsips_seq_sysctl_table
{
	struct ctl_table_header *sysctl_header;
	ctl_table hsips_conf_dir[6];
	ctl_table ipsec_proto_dir[2];
	ctl_table root_dir[2];
} hsips_seq_sysctl = {
	NULL,
	{
		{NET_IPSEC_SEQ_ENABLE, "seq_rx_enable",
			&hsips_seq_var, sizeof(int), 0644, NULL,
			&hsips_seq_sysctl_strategy},
		{NET_IPSEC_SEQ_ENABLE, "seq_tx_enable",
			&hsips_seq_var, sizeof(int), 0644, NULL,
			&hsips_seq_sysctl_strategy},
		{NET_IPSEC_SEQ_RESET, "rx_qsize",
			&hsips_seq_var, sizeof(int), 0444, NULL,
			&hsips_seq_sysctl_strategy},
		{NET_IPSEC_SEQ_RESET, "tx_qsize",
			&hsips_seq_var, sizeof(int), 0444, NULL,
			&hsips_seq_sysctl_strategy},
		{NET_IPSEC_SEQ_RESET, "seq_reset",
			&hsips_seq_var, sizeof(int), 0200, NULL,
			&hsips_seq_sysctl_strategy},
		{0}
	},
	{
		{NET_IPSEC, "ipsec", NULL, 0, 0555,
			hsips_seq_sysctl.hsips_conf_dir},
		{0}
	},
	{
		{CTL_NET, "net", NULL, 0, 0555,
			hsips_seq_sysctl.ipsec_proto_dir},
		{0}
	}
};

#endif /* CONFIG_SYSCTL */


void hsips_inc_tx_seq(struct sk_buff *skb)
{
#ifdef CONFIG_SYSCTL
	if (!seq_tx_active)
		return;
#endif

	inc_seq(&hsips_seq_tx, skb);
}


void hsips_inc_rx_seq(struct sk_buff *skb)
{
#ifdef CONFIG_SYSCTL
	if (!seq_rx_active)
		return;
#endif

	inc_seq(&hsips_seq_rx, skb);
}


int hsips_ip_send(struct sk_buff *skb)
{
#ifdef CONFIG_SYSCTL
	if (!seq_tx_active) {
		if (IPSEC_SKB_CB(skb)->ignore) {
			kfree_skb(skb);
			return 0;
		}

		return ip_send(skb);
	}
#endif

	return check_seq(skb, &hsips_seq_tx);
}


int hsips_netif_rx(struct sk_buff *skb)
{
#ifdef CONFIG_SYSCTL
	if (!seq_rx_active) {
		if (IPSEC_SKB_CB(skb)->ignore) {
			kfree_skb(skb);
			return 0;
		}

		return netif_rx(skb);
	}
#endif

	return check_seq(skb, &hsips_seq_rx);
}


int ipsec_threads_check_seq_q(void)
{
	if (!seq_rx_active && !seq_tx_active)
		return 0;

	if (atomic_read(&ipsec_thread_seq_q) <= 0)
		return 1;

	atomic_dec(&ipsec_thread_seq_q);

	return 0;
}

static int ipsec_seq_stats_get_info(char *buffer, char **start, off_t offset,
                                  int length)
{
        struct list_head *le;
        int len = 0, i = 0;
        off_t begin = 0;
	
	len += sprintf(buffer + len, "hsips_tlet_flush_buf_lock_max   : %lu\n",
			hsips_tlet_flush_buf_lock_max);
	len += sprintf(buffer + len, "hsips_tlet_flush_buf_lock_count : %lu\n",
			hsips_tlet_flush_buf_lock_count);

	len += sprintf(buffer + len, "hsips_flush_all_packets_lock_max : %lu\n",
			hsips_flush_all_packets_lock_max);
	len += sprintf(buffer + len, "hsips_flush_all_packets_lock_count : %lu\n",
			hsips_flush_all_packets_lock_count);
	len += sprintf(buffer + len, "hsips_seq_timer_expiry_lock_max : %lu\n",
			hsips_seq_timer_expiry_lock_max);
	len += sprintf(buffer + len, "hsips_seq_timer_expiry_lock_count : %lu\n",
			hsips_seq_timer_expiry_lock_count);

	len += sprintf(buffer + len, "hsips_seq_rx stats\n");
	len += sprintf(buffer + len, "  max_tlet_qsize     : %lu\n"
			, hsips_seq_rx.max_tlet_qsize);


	len += sprintf(buffer + len, "  type               : %u\n"
			, hsips_seq_rx.type);
	len += sprintf(buffer + len, "  tlet_qsize         : %lu\n"
			, hsips_seq_rx.tlet_qsize);
	len += sprintf(buffer + len, "  max_dropped        : %lu\n"
			, hsips_seq_rx.max_dropped);
	len += sprintf(buffer + len, "  pkt_dropped        : %lu\n"
			, hsips_seq_rx.pkt_dropped);
	len += sprintf(buffer + len, "  pkt_delayed        : %lu\n"
			, hsips_seq_rx.pkt_delayed);
	len += sprintf(buffer + len, "  pkt_flushed        : %lu\n"
			, hsips_seq_rx.pkt_flushed);
	len += sprintf(buffer + len, "  flush_cnt          : %lu\n"
			, hsips_seq_rx.flush_cnt);


	len += sprintf(buffer + len, "  out_of_reset       : %lu\n"
			, hsips_seq_rx.out_of_reset);
	len += sprintf(buffer + len, "  max_timeout        : %lu\n"
			, hsips_seq_rx.max_timeout);
	len += sprintf(buffer + len, "  timer_evt          : %lu\n"
			, hsips_seq_rx.timer_evt);
	len += sprintf(buffer + len, "  timer_evt_max_qsize: %lu\n"
			, hsips_seq_rx.timer_evt_max_qsize);
	len += sprintf(buffer + len, "  out_of_order_evt   : %lu\n"
			, hsips_seq_rx.out_of_order_evt);
	len += sprintf(buffer + len, "  in_order_evt       : %lu\n"
			, hsips_seq_rx.in_order_evt);
	len += sprintf(buffer + len, "  in_order_pkt_array : ");
	for (i=0; i<32; i++)
		len += sprintf(buffer + len, "[%d]=%lu "
				, i
				, hsips_seq_rx.in_order_pkt_array[i]);
	len += sprintf(buffer + len, "\n\n");
	
	len += sprintf(buffer + len, "hsips_seq_tx stats\n");
	len += sprintf(buffer + len, "  max_tlet_qsize     : %lu\n"
			, hsips_seq_tx.max_tlet_qsize);

	len += sprintf(buffer + len, "  type               : %u\n"
			, hsips_seq_tx.type);
	len += sprintf(buffer + len, "  tlet_qsize         : %lu\n"
			, hsips_seq_tx.tlet_qsize);
	len += sprintf(buffer + len, "  max_dropped        : %lu\n"
			, hsips_seq_tx.max_dropped);
	len += sprintf(buffer + len, "  pkt_dropped        : %lu\n"
			, hsips_seq_tx.pkt_dropped);
	len += sprintf(buffer + len, "  pkt_delayed        : %lu\n"
			, hsips_seq_tx.pkt_delayed);
	len += sprintf(buffer + len, "  pkt_flushed        : %lu\n"
			, hsips_seq_tx.pkt_flushed);
	len += sprintf(buffer + len, "  flush_cnt          : %lu\n"
			, hsips_seq_tx.flush_cnt);

	len += sprintf(buffer + len, "  out_of_reset       : %lu\n"
			, hsips_seq_tx.out_of_reset);
	len += sprintf(buffer + len, "  max_timeout        : %lu\n"
			, hsips_seq_tx.max_timeout);

	len += sprintf(buffer + len, "  timer_evt          : %lu\n"
			, hsips_seq_tx.timer_evt);
	len += sprintf(buffer + len, "  timer_evt_max_qsize: %lu\n"
			, hsips_seq_tx.timer_evt_max_qsize);
	len += sprintf(buffer + len, "  out_of_order_evt   : %lu\n"
			, hsips_seq_tx.out_of_order_evt);
	len += sprintf(buffer + len, "  in_order_evt       : %lu\n"
			, hsips_seq_tx.in_order_evt);
	len += sprintf(buffer + len, "  in_order_pkt_array : ");
	for (i=0; i<32; i++)
		len += sprintf(buffer + len, "[%d]=%lu "
				, i
				, hsips_seq_tx.in_order_pkt_array[i]);
	len += sprintf(buffer + len, "\n\n");

out:
        *start = buffer + (offset - begin);
        len -= (offset - begin);
        if (len > length)
                len = length;
        return len;
}


static void *hsips_seq_sysctl_header = NULL;

int hsips_seq_init(void)
{
	seq_rx_active = seq_tx_active = 1;

	tasklet_init(&hsips_seq_tx.tlet, hsips_tlet_flush_buf,
	             (unsigned long) &hsips_seq_tx);
	tasklet_init(&hsips_seq_rx.tlet, hsips_tlet_flush_buf,
	             (unsigned long) &hsips_seq_rx);

	init_timer(&hsips_seq_tx.timer);
	hsips_seq_tx.timer.function = &seq_timer_expiry;
	hsips_seq_tx.timer.data = (unsigned long) &hsips_seq_tx;

	init_timer(&hsips_seq_rx.timer);
	hsips_seq_rx.timer.function = &seq_timer_expiry;
	hsips_seq_rx.timer.data = (unsigned long) &hsips_seq_rx;

	atomic_set(&ipsec_thread_seq_q, IPSEC_THREAD_MAX_BUF);

	memset((char *) hsips_seq_tx.in_order_pkt_array, 0, 32 * sizeof(long));
	memset((char *) hsips_seq_rx.in_order_pkt_array, 0, 32 * sizeof(long));

#ifdef CONFIG_SYSCTL
	hsips_seq_sysctl_header = 
     	register_sysctl_table(hsips_seq_sysctl.root_dir, 0);
#endif

	proc_net_create("ipsec_seq_stats", 0, ipsec_seq_stats_get_info);

	return 0;
}

void hsips_seq_cleanup(void)
{
#ifdef CONFIG_SYSCTL
	unregister_sysctl_table(hsips_seq_sysctl_header);
	hsips_seq_sysctl_header = NULL;
#endif

	atomic_set(&ipsec_thread_seq_q, 0);

	seq_reset();

	tasklet_kill(&hsips_seq_tx.tlet);
	tasklet_kill(&hsips_seq_rx.tlet);

	proc_net_remove("ipsec_seq_stats");

	return;
}

#ifdef CONFIG_IPSEC_FRAG
void hsips_frag_on(void)
{
    seq_rx_active = seq_tx_active = 0;

    return;
}
#endif
