/*
 * filename: thread_dev.c
 *
 * Developed by secunet AG
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/devfs_fs_kernel.h>

#include "thread_dev.h"
#include "hsips.h"

static int dev_busy = 0;

static int ipst_dev_open(struct inode *inode, struct file *file)
{
	if (dev_busy)
		return -EBUSY;
	
	dev_busy = 1;

	printk(KERN_DEBUG "ipst_dev: pid %d open thread_dev \n", current->pid);

	return 0;
}

static int ipst_dev_release(struct inode *inode, struct file *file)
{
	if (!dev_busy)
		return -ESRCH;

	dev_busy = 0;

	return 0;
}

static int ipst_dev_ioctl(struct inode *inode, struct file *file,
                          unsigned int cmd, unsigned long arg)
{
	struct list_head *le;
	pid_t pid = 0;
	int retval;

	if (!dev_busy)
		return -ESRCH;
	
	retval = -EINVAL;
	switch (cmd) {
	  case IPST_START:
		if (respawn_thread.task) {
			retval = -EBUSY;
			break;
		}

		printk(KERN_DEBUG "hsips_init: switching to respawn threads ...\n");

		respawn_thread.num = 1;
		if (arg > 0) 
			respawn_thread.num = (int) arg;
			  
		retval = -ENOMEM;
		pid = kernel_thread(hsips_respawn, (void *) &respawn_thread, 0);
		if (pid < 0) {
			printk(KERN_NOTICE "hsips_init: failed.\n");
			retval = pid;
		}

		break;
	  case IPST_STOP:
		retval = hsips_destroy(&respawn_thread);
		memset((char *) &respawn_thread, 0, sizeof(struct respawn_status));

		retval = 0;

		break;
	  case IPST_STATUS:
		retval = -ENOSYS;
		if (!respawn_thread.task) {
			printk(KERN_NOTICE "HSIPS: no status info available!\n");
			break;
		}

		printk(KERN_NOTICE "ipst_dev: status info for all running threads:\n");
		printk(KERN_NOTICE "HSIPS: number of threads: %d\n", 
		       respawn_thread.num);

		list_for_each(le, &hsips_thread_list) {
			struct ipst_status *thread;

			thread = list_entry(le, struct ipst_status, list);
			if (!thread->task)
				continue;

			spin_lock(&thread->lock);

			printk(KERN_NOTICE "HSIPS: #%d: num=%d pid=%d cnt=%d up=%d dn=%d sig=%d cpu(%d)\n",
			       i, thread->num, thread->pid,
			       atomic_read(&thread->count), atomic_read(&thread->up),
			       atomic_read(&thread->dn), atomic_read(&thread->sig),
			       thread->task->processor);

			spin_unlock(&thread->lock);
		}

		printk(KERN_NOTICE "HSIPS: qsize=%d\n", hsips_get_qsize());
		retval = 0;

		break;
	}

	return retval;
}


static struct file_operations ipst_fops = {
	open:		ipst_dev_open,
	release:	ipst_dev_release,
	ioctl:		ipst_dev_ioctl,
};


int register_hsipsthreaddev(void)
{
	if (devfs_register_chrdev(HSIPS_THREADDEV_NO, "ipsec_thread", &ipst_fops)) {
		printk(KERN_NOTICE "ipsec_thread: unable to get major %d\n",
		       HSIPS_THREADDEV_NO);
		return -EIO;
	}

	return 0;
}


int unregister_hsipsthreaddev(void)
{
	return devfs_unregister_chrdev(HSIPS_THREADDEV_NO, "ipsec_thread");
}


