/*
 * Common routines for IPSEC transformations.
 * Copyright (C) 1996, 1997  John Ioannidis.
 * Copyright (C) 1998, 1999  Richard Guy Briggs.
 *
 * Modified by secunet AG.
 *  
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/config.h>
#include <linux/version.h>

#include <linux/kernel.h> /* printk() */
#include <linux/slab.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/skbuff.h>
#include <linux/random.h>	/* get_random_bytes() */
#include <freeswan.h>
#include <linux/spinlock.h> /* *lock* */
#include <asm/uaccess.h>
#include <linux/in6.h>
#include <asm/checksum.h>
#include <net/ip.h>
#include <linux/cryptdev.h>

#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_ipe4.h"
#include "ipsec_ah.h"
#include "ipsec_esp.h"

#include <pfkeyv2.h>
#include <pfkey.h>

#ifdef CONFIG_IPSECTHREAD
#  include "ipsec_threads/hsips.h"
#endif

#ifdef CONFIG_IPSEC_DEBUG
int debug_xform = 0;
#endif /* CONFIG_IPSEC_DEBUG */

#define SENDERR(_x) do { len = -(_x); goto errlab; } while (0)

extern int plain_key_count;

struct xformsw xformsw[] = {
{ XF_IP4,		0,		"IPv4_Encapsulation"},
{ XF_AHHMACMD5,		XFT_AUTH,	"HMAC_MD5_Authentication"},
{ XF_AHHMACSHA1,	XFT_AUTH,	"HMAC_SHA-1_Authentication"},
{ XF_ESP3DES,		XFT_CONF,	"3DES_Encryption"},
{ XF_ESP3DESMD596,	XFT_CONF,	"3DES-MD5-96_Encryption"},
{ XF_ESP3DESSHA196,	XFT_CONF,	"3DES-SHA1-96_Encryption"},
{ XF_ESPNULLMD596,	XFT_CONF,	"NULL-MD5-96_ESP_*Plaintext*"},
{ XF_ESPNULLSHA196,	XFT_CONF,	"NULL-SHA1-96_ESP_*Plaintext*"},
};

struct tdb *tdbh[TDB_HASHMOD];

DECLARE_WAIT_QUEUE_HEAD(ipsec_sync_lock_wq);
rwlock_t ipsec_sync_lock = RW_LOCK_UNLOCKED;
unsigned int ipsec_sync_lock_hd;

struct xformsw *xformswNXFORMSW = &xformsw[sizeof(xformsw)/sizeof(xformsw[0])];


int
ipsec_tdbinit(void)
{
	int i;

	for(i = 1; i < TDB_HASHMOD; i++) {
		tdbh[i] = NULL;
	}
	return 0;
}

struct tdb *
get_tdb(struct sa_id *said)
{
	struct tdb *tdbp, *tdbq;
        char sa[SATOA_BUF];

	if(!said) {
		KLIPS_PRINT(debug_xform,
			    "klips_error:get_tdb: null pointer passed in!\n");
		return NULL;
	}
	
	satoa(*said, 0, sa, SATOA_BUF);

	tdbp = __find_tdb(said);
	
	if(!tdbp) {
		KLIPS_PRINT(debug_xform,
			    "klips_error:get_tdb: no tdb found in "
			    "hashtable for SA=%s!\n"
			    , sa);
		return NULL;
	}
	
	if ( !atomic_read(&tdbp->del_flag) ) {
		tdbq = tdbp;
		
		while(tdbq->tdb_onext) {
			tdbq = tdbq->tdb_onext;
		}

		while(tdbq) {
			atomic_inc(&tdbq->use_count);
			
			satoa(tdbq->tdb_said, 0, sa, SATOA_BUF);
			KLIPS_PRINT(debug_netlink, 
				"get_tdb: SA=%s del_flag=%d use_count=%d\n"
				, sa
				, atomic_read(&tdbq->del_flag)
				, atomic_read(&tdbq->use_count));
			tdbq = tdbq->tdb_inext;
		}
		
		return tdbp;
	} else {
		KLIPS_PRINT(debug_xform, 
			"klips_debug:get_tdb: "
			"tdb of SA:%s marked for deletion!.\n"
			, sa);
		return NULL;
	}
}

int
put_tdb(struct tdb *tdbp)
{
	struct tdb *tdbq;
        char sa[SATOA_BUF];

	if(!tdbp) {
		KLIPS_PRINT(debug_xform,
			    "klips_error:put_tdb: null pointer passed in!\n");
		return -ENODATA;
	}

	tdbq = tdbp;
	
	while(tdbq->tdb_onext) {
		tdbq = tdbq->tdb_onext;
	}

	while(tdbq) {
		atomic_dec(&tdbq->use_count);
		
		satoa(tdbq->tdb_said, 0, sa, SATOA_BUF);
		KLIPS_PRINT(debug_netlink, 
			"put_tdb: SA=%s del_flag=%d use_count=%d\n"
			, sa
			, atomic_read(&tdbq->del_flag)
			, atomic_read(&tdbq->use_count));

		tdbq = tdbq->tdb_inext;
	}
	return 0;
}
	
struct tdb *
__find_tdb(struct sa_id *said)
{
	int hashval;
	struct tdb *tdbp;
        char sa[SATOA_BUF];

	if(!said) {
		KLIPS_PRINT(debug_xform,
			    "klips_error:find_tdb: null pointer passed in!\n");
		return NULL;
	}

	satoa(*said, 0, sa, SATOA_BUF);

	hashval = (said->spi+said->dst.s_addr+said->proto) % TDB_HASHMOD;
	
	KLIPS_PRINT(debug_xform,
		    "klips_debug:find_tdb: linked entry in tdb table "
		    "for hash=%d of SA:%s requested.\n"
		    , hashval, sa);

	if(!(tdbp = tdbh[hashval])) {
		KLIPS_PRINT(debug_xform,
			"klips_debug:find_tdb: no entries in tdb table "
			"for hash=%d of SA:%s.\n"
			, hashval, sa);
		return NULL;
	}

	for (; tdbp; tdbp = tdbp->tdb_hnext) {
		if ((tdbp->tdb_said.spi == said->spi) &&
		    (tdbp->tdb_said.dst.s_addr == said->dst.s_addr) &&
		    (tdbp->tdb_said.proto == said->proto)) {
				KLIPS_PRINT(debug_netlink, 
					"find_tdb: hash=%d SA=%s "
					"del_flag=%d use_count=%d\n"
					, hashval
					, sa
					, atomic_read(&tdbp->del_flag)
					, atomic_read(&tdbp->use_count));
			return tdbp;
		}
	}
	
	KLIPS_PRINT(debug_xform,
		    "klips_debug:find_tdb: no entry in linked list "
		    "for hash=%d of SA:%s.\n"
		    , hashval, sa);
	return NULL;
}

/* void */
int
insert_tdb(struct tdb *tdbp)
{
	int error = 0;
	unsigned int hashval;

	if(!tdbp) {
		KLIPS_PRINT(debug_xform,
			    "klips_error:puttdb: null pointer passed in!\n");
		return -ENODATA;
	}
	hashval = ((tdbp->tdb_said.spi + tdbp->tdb_said.dst.s_addr + tdbp->tdb_said.proto) % TDB_HASHMOD);

	
	tdbp->tdb_hnext = tdbh[hashval];
	tdbh[hashval] = tdbp;
	

	return error;
}

/* This tdb better be locked before it is handed in, or races might
 * happen */

/* void */
int
__deltdb(struct tdb *tdbp)
{
	unsigned int hashval;
	struct tdb *tdbtp;
        char sa[SATOA_BUF];

	if(!tdbp) {
		KLIPS_PRINT(debug_xform,
			    "klips_error:deltdb: null pointer passed in!\n");
		return -ENODATA;
	}
	
	satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
	if(tdbp->tdb_inext || tdbp->tdb_onext) {
		KLIPS_PRINT(debug_xform,
			    "klips_error:deltdb: SA:%s still linked!\n",
			    sa);
		return -EMLINK;
	}
	
	hashval = ((tdbp->tdb_said.spi + tdbp->tdb_said.dst.s_addr + tdbp->tdb_said.proto) % TDB_HASHMOD);
	
	KLIPS_PRINT(debug_xform,
		    "klips_debug:deltdb: deleting SA:%s, hashval=%d.\n",
		    sa, hashval);
	if(!tdbh[hashval]) {
		KLIPS_PRINT(debug_xform,
			    "klips_debug:deltdb: no entries in tdb table "
			    "for hash=%d of SA:%s.\n"
			    , hashval, sa);
		return -ENOENT;
	}
	
	if (tdbp == tdbh[hashval]) {
		tdbh[hashval] = tdbh[hashval]->tdb_hnext;
		tdbp->tdb_hnext = NULL;
		KLIPS_PRINT(debug_xform,
			    "klips_debug:deltdb: successfully deleted "
			    "first tdb in chain.\n");
		return 0;
	} else {
		for (tdbtp = tdbh[hashval]; tdbtp; tdbtp = tdbtp->tdb_hnext) {
			if (tdbtp->tdb_hnext == tdbp) {
				tdbtp->tdb_hnext = tdbp->tdb_hnext;
				tdbp->tdb_hnext = NULL;
				KLIPS_PRINT(debug_xform,
					    "klips_debug:deltdb: successfully "
					    "deleted link in tdb chain.\n");
				return 0;
			}
		}
	}
	
	KLIPS_PRINT(debug_xform,
		    "klips_debug:deltdb: no entries in linked list "
		    "for hash=%d of SA:%s.\n"
		    , hashval, sa);
	return -ENOENT;
}

int
markdel_tdbchain(struct tdb *tdbp)
{
	int error = 0;
        char sa[SATOA_BUF];

	if(!tdbp) {
		KLIPS_PRINT(debug_xform,
		    "klips_error:markdel_tdbchain: null pointer passed in!\n");
		return -ENODATA;
	}

	while(tdbp->tdb_onext) {
		tdbp = tdbp->tdb_onext;
	}

	while(tdbp) {
		atomic_set(&tdbp->del_flag, 1);

		satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
		KLIPS_PRINT(debug_netlink, 
			"klips_debug:markdel_tdbchain: "
			"passed SA:%s use_count=%d del_flag=%d\n"
			, sa
			, atomic_read(&tdbp->use_count)
			, atomic_read(&tdbp->del_flag));

		tdbp = tdbp->tdb_inext;
	}
	
	return error;
}

/* This tdb better be locked before it is handed in, or races might
 * happen */

int
__deltdbchain(struct tdb *tdbp)
{
	struct tdb *tdbdel = NULL;
	int error = 0;
        char sa[SATOA_BUF];

	if(!tdbp) {
		KLIPS_PRINT(debug_xform,
		    "klips_error:deltdbchain: null pointer passed in!\n");
		return -ENODATA;
	}

	satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
	KLIPS_PRINT(debug_xform,
		    "klips_debug:deltdbchain: passed SA:%s\n", sa);
	while(tdbp->tdb_onext) {
		tdbp = tdbp->tdb_onext;
	}

	while(tdbp) {
		/* XXX send a pfkey message up to advise of deleted TDB */
		satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
		KLIPS_PRINT(debug_xform,
			    "klips_debug:deltdbchain: "
			    "unlinking and deleting SA:%s", sa);
		tdbdel = tdbp;
		tdbp = tdbp->tdb_inext;
		if(tdbp) {
			satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
			KLIPS_PRINT(debug_xform,
				    ", inext=%s", sa);
			tdbdel->tdb_inext = NULL;
			tdbp->tdb_onext = NULL;
		}
		KLIPS_PRINT(debug_xform,
			    ".\n");
		if((error = __deltdb(tdbdel))) {
			KLIPS_PRINT(debug_xform,
				    "klips_debug:deltdbchain: "
				    "deltdb returned error %d.\n", -error);
			return error;
		}
		if((error = ipsec_tdbwipe(tdbdel))) {
			KLIPS_PRINT(debug_xform,
				    "klips_debug:deltdbchain: "
				    "ipsec_tdbwipe returned error %d.\n"
				    , -error);
			return error;
		}
	}
	return error;
}

int 
ipsec_tdbcleanup(__u8 proto)
{
	int i;
	int error = 0;
	struct tdb *tdbp, **tdbprev, *tdbdel;
        char sa[SATOA_BUF];

	KLIPS_PRINT(debug_xform,
		    "klips_debug:ipsec_tdbcleanup: cleaning up proto=%d.\n"
		    , proto);

	for (i = 0; i < TDB_HASHMOD; i++) {
		tdbprev = &(tdbh[i]);
		tdbp = tdbh[i];
		for(; tdbp;) {
			satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
			KLIPS_PRINT(debug_xform,
				    "klips_debug:ipsec_tdbcleanup: "
				    "checking SA:%s, hash=%d"
				    , sa, i);
			tdbdel = tdbp;
			tdbp = tdbdel->tdb_hnext;
			if(tdbp) {
				satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
				KLIPS_PRINT(debug_xform,
					    ", hnext=%s", sa);
			}
			if(*tdbprev) {
				satoa((*tdbprev)->tdb_said, 0, sa, SATOA_BUF);
				KLIPS_PRINT(debug_xform,
					    ", *tdbprev=%s", sa);
				if((*tdbprev)->tdb_hnext) {
					satoa((*tdbprev)->tdb_hnext->tdb_said
						, 0, sa, SATOA_BUF);
					KLIPS_PRINT(debug_xform,
						    ", *tdbprev->tdb_hnext=%s"
						    , sa);
				}
			}
			KLIPS_PRINT(debug_xform,
				    ".\n");
			if(!proto || (proto == tdbdel->tdb_said.proto)) {
				satoa(tdbdel->tdb_said, 0, sa, SATOA_BUF);
				KLIPS_PRINT(debug_xform,
					    "klips_debug:ipsec_tdbcleanup: "
					    "deleting SA chain:%s.\n"
					    , sa);
				/* *tdbprev = tdbdel->tdb_hnext; */
				if((error = __deltdbchain(tdbdel))) {
					goto errlab;
				}
				tdbprev = &(tdbh[i]);
				tdbp = tdbh[i];

				KLIPS_PRINT(debug_xform,
					    "klips_debug:ipsec_tdbcleanup: "
					    "deleted SA chain:%s"
					    , sa);
				if(tdbp) {
					satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
					KLIPS_PRINT(debug_xform,
						    ", tdbh[%d]=%s", i, sa);
				}
				if(*tdbprev) {
					satoa((*tdbprev)->tdb_said, 0, sa
							, SATOA_BUF);
					KLIPS_PRINT(debug_xform,
						    ", *tdbprev=%s", sa);
					if((*tdbprev)->tdb_hnext) {
						satoa((*tdbprev)->tdb_hnext->tdb_said, 0, sa, SATOA_BUF);
						KLIPS_PRINT(debug_xform,
							", *tdbprev->tdb_hnext=%s", sa);
					}
				}
				KLIPS_PRINT(debug_xform,
					    ".\n");
			} else {
				tdbprev = &tdbdel;
			}
		}
	}
 errlab:


	return(-error);
}

int
__deloldtdbs(void)
{
	struct tdb *tdbp;
	int i;
        char sa[SATOA_BUF];

retry:
	for (i = 0; i < TDB_HASHMOD; i++) {
		for (tdbp = tdbh[i]; tdbp; tdbp = tdbp->tdb_hnext) {
					
			if ( atomic_read(&tdbp->del_flag) &&
			     !atomic_read(&tdbp->use_count) ) {
				satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
				KLIPS_PRINT(debug_netlink, 
					"klips_debug:deloldtdbs: calling "
					"deltdbchain for SA:%s, "
		                        "use_count=%d, del_flag=%d\n"
					, sa
					, atomic_read(&tdbp->use_count)
					, atomic_read(&tdbp->del_flag));
				
				__deltdbchain(tdbp);

				goto retry;
			}
		}
	}

	return 0;
}


int
ipsec_tdbwipe(struct tdb *tdbp)
{
	int hno;
	
	if(!tdbp) {
		return -1;
	}

	if(tdbp->tdb_addr_s) {
		memset((caddr_t)(tdbp->tdb_addr_s), 0, tdbp->tdb_addr_s_size);
		kfree(tdbp->tdb_addr_s);
	}
	tdbp->tdb_addr_s = NULL;

	if(tdbp->tdb_addr_d) {
		memset((caddr_t)(tdbp->tdb_addr_d), 0, tdbp->tdb_addr_d_size);
		kfree(tdbp->tdb_addr_d);
	}
	tdbp->tdb_addr_d = NULL;

	if(tdbp->tdb_addr_p) {
		memset((caddr_t)(tdbp->tdb_addr_p), 0, tdbp->tdb_addr_p_size);
		kfree(tdbp->tdb_addr_p);
	}
	tdbp->tdb_addr_p = NULL;

	if(tdbp->tdb_key_a) {
		memset((caddr_t)(tdbp->tdb_key_a), 0, tdbp->tdb_key_a_size);
		kfree(tdbp->tdb_key_a);
	}
	tdbp->tdb_key_a = NULL;

	if(tdbp->tdb_key_e) {
		memset((caddr_t)(tdbp->tdb_key_e), 0, tdbp->tdb_key_e_size);
		kfree(tdbp->tdb_key_e);
	}
	tdbp->tdb_key_e = NULL;

	if(tdbp->tdb_iv) {
		memset((caddr_t)(tdbp->tdb_iv), 0, tdbp->tdb_iv_size);
		kfree(tdbp->tdb_iv);
	}
	tdbp->tdb_iv = NULL;

	hno = 1;
#ifdef CONFIG_SMP
	hno = smp_num_cpus;
#endif
#ifdef CONFIG_IPSECTHREAD
	hno = hsips_get_thread_cnt();
#endif
	
	if (hno > 0) {
		int i;
		
		for (i = 0; i < hno; i++) {
			if (tdbp->tdb_i_combined_handle[i])
				crypto_close(tdbp->tdb_i_combined_handle[i]);
			if (tdbp->tdb_o_combined_handle[i])
				crypto_close(tdbp->tdb_o_combined_handle[i]);
			if (tdbp->tdb_authalg_g_handle[i]) {
				if (tdbp->tdb_key_bits_a / 8 == tdbp->tdb_key_a_size)
					plain_key_count--;
				crypto_close(tdbp->tdb_authalg_g_handle[i]);
			}
			if (tdbp->tdb_authalg_v_handle[i]) {
				if (tdbp->tdb_key_bits_a / 8 == tdbp->tdb_key_a_size)
					plain_key_count--;
				crypto_close(tdbp->tdb_authalg_v_handle[i]);
			}
			if (tdbp->tdb_encalg_e_handle[i]) {
				if (tdbp->tdb_key_bits_e / 8 == tdbp->tdb_key_e_size)
					plain_key_count--;
				crypto_close(tdbp->tdb_encalg_e_handle[i]);
			}
			if (tdbp->tdb_encalg_d_handle[i]) {
				if (tdbp->tdb_key_bits_e / 8 == tdbp->tdb_key_e_size)
					plain_key_count--;
				crypto_close(tdbp->tdb_encalg_d_handle[i]);
			}
		}
	} else
		printk(KERN_NOTICE "ipsec_tdbwipe: no handles available\n");

#if defined CONFIG_SMP || defined CONFIG_IPSECTHREAD
	kfree(tdbp->tdb_i_combined_handle);
	kfree(tdbp->tdb_o_combined_handle);
	kfree(tdbp->tdb_authalg_g_handle);
	kfree(tdbp->tdb_authalg_v_handle);
	kfree(tdbp->tdb_encalg_e_handle);
	kfree(tdbp->tdb_encalg_d_handle);
#endif

	memset((caddr_t)tdbp, 0, sizeof(*tdbp));
	
	kfree(tdbp);
	tdbp = NULL;

	return 0;
}

/*
 * Revision 1.45  2000/09/08 19:16:51  rgb
 * Change references from DEBUG_IPSEC to CONFIG_IPSEC_DEBUG.
 * Removed all references to CONFIG_IPSEC_PFKEYv2.
 *
 * Revision 1.44  2000/08/30 05:29:04  rgb
 * Compiler-define out no longer used tdb_init() in ipsec_xform.c.
 *
 * Revision 1.43  2000/08/18 21:30:41  rgb
 * Purged all tdb_spi, tdb_proto and tdb_dst macros.  They are unclear.
 *
 * Revision 1.42  2000/08/01 14:51:51  rgb
 * Removed _all_ remaining traces of DES.
 *
 * Revision 1.41  2000/07/28 14:58:31  rgb
 * Changed kfree_s to kfree, eliminating extra arg to fix 2.4.0-test5.
 *
 * Revision 1.40  2000/06/28 05:50:11  rgb
 * Actually set iv_bits.
 *
 * Revision 1.39  2000/05/10 23:11:09  rgb
 * Added netlink debugging output.
 * Added a cast to quiet down the ntohl bug.
 *
 * Revision 1.38  2000/05/10 19:18:42  rgb
 * Cast output of ntohl so that the broken prototype doesn't make our
 * compile noisy.
 *
 * Revision 1.37  2000/03/16 14:04:59  rgb
 * Hardwired CONFIG_IPSEC_PFKEYv2 on.
 *
 * Revision 1.36  2000/01/26 10:11:28  rgb
 * Fixed spacing in error text causing run-in words.
 *
 * Revision 1.35  2000/01/21 06:17:16  rgb
 * Tidied up compiler directive indentation for readability.
 * Added ictx,octx vars for simplification.(kravietz)
 * Added macros for HMAC padding magic numbers.(kravietz)
 * Fixed missing key length reporting bug.
 * Fixed bug in tdbwipe to return immediately on NULL tdbp passed in.
 *
 * Revision 1.34  1999/12/08 00:04:19  rgb
 * Fixed SA direction overwriting bug for netlink users.
 *
 * Revision 1.33  1999/12/01 22:16:44  rgb
 * Minor formatting changes in ESP MD5 initialisation.
 *
 * Revision 1.32  1999/11/25 09:06:36  rgb
 * Fixed error return messages, should be returning negative numbers.
 * Implemented SENDERR macro for propagating error codes.
 * Added debug message and separate error code for algorithms not compiled
 * in.
 *
 * Revision 1.31  1999/11/23 23:06:26  rgb
 * Sort out pfkey and freeswan headers, putting them in a library path.
 *
 * Revision 1.30  1999/11/18 04:09:20  rgb
 * Replaced all kernel version macros to shorter, readable form.
 *
 * Revision 1.29  1999/11/17 15:53:40  rgb
 * Changed all occurrences of #include "../../../lib/freeswan.h"
 * to #include <freeswan.h> which works due to -Ilibfreeswan in the
 * klips/net/ipsec/Makefile.
 *
 * Revision 1.28  1999/10/18 20:04:01  rgb
 * Clean-out unused cruft.
 *
 * Revision 1.27  1999/10/03 19:01:03  rgb
 * Spinlock support for 2.3.xx and 2.0.xx kernels.
 *
 * Revision 1.26  1999/10/01 16:22:24  rgb
 * Switch from assignment init. to functional init. of spinlocks.
 *
 * Revision 1.25  1999/10/01 15:44:54  rgb
 * Move spinlock header include to 2.1> scope.
 *
 * Revision 1.24  1999/10/01 00:03:46  rgb
 * Added tdb structure locking.
 * Minor formatting changes.
 * Add function to initialize tdb hash table.
 *
 * Revision 1.23  1999/05/25 22:42:12  rgb
 * Add deltdbchain() debugging.
 *
 * Revision 1.22  1999/05/25 21:24:31  rgb
 * Add debugging statements to deltdbchain().
 *
 * Revision 1.21  1999/05/25 03:51:48  rgb
 * Refix error return code.
 *
 * Revision 1.20  1999/05/25 03:34:07  rgb
 * Fix error return for flush.
 *
 * Revision 1.19  1999/05/09 03:25:37  rgb
 * Fix bug introduced by 2.2 quick-and-dirty patch.
 *
 * Revision 1.18  1999/05/05 22:02:32  rgb
 * Add a quick and dirty port to 2.2 kernels by Marc Boucher <marc@mbsi.ca>.
 *
 * Revision 1.17  1999/04/29 15:20:16  rgb
 * Change gettdb parameter to a pointer to reduce stack loading and
 * facilitate parameter sanity checking.
 * Add sanity checking for null pointer arguments.
 * Add debugging instrumentation.
 * Add function deltdbchain() which will take care of unlinking,
 * zeroing and deleting a chain of tdbs.
 * Add a parameter to tdbcleanup to be able to delete a class of SAs.
 * tdbwipe now actually zeroes the tdb as well as any of its pointed
 * structures.
 *
 * Revision 1.16  1999/04/16 15:36:29  rgb
 * Fix cut-and-paste error causing a memory leak in IPIP TDB freeing.
 *
 * Revision 1.15  1999/04/11 00:29:01  henry
 * GPL boilerplate
 *
 * Revision 1.14  1999/04/06 04:54:28  rgb
 * Fix/Add RCSID Id: and Log: bits to make PHMDs happy.  This includes
 * patch shell fixes.
 *
 * Revision 1.13  1999/02/19 18:23:01  rgb
 * Nix debug off compile warning.
 *
 * Revision 1.12  1999/02/17 16:52:16  rgb
 * Consolidate satoa()s for space and speed efficiency.
 * Convert DEBUG_IPSEC to KLIPS_PRINT
 * Clean out unused cruft.
 * Ditch NET_IPIP dependancy.
 * Loop for 3des key setting.
 *
 * Revision 1.11  1999/01/26 02:09:05  rgb
 * Remove ah/esp/IPIP switching on include files.
 * Removed CONFIG_IPSEC_ALGO_SWITCH macro.
 * Removed dead code.
 * Clean up debug code when switched off.
 * Remove references to INET_GET_PROTOCOL.
 * Added code exclusion macros to reduce code from unused algorithms.
 *
 * Revision 1.10  1999/01/22 06:28:55  rgb
 * Cruft clean-out.
 * Put random IV generation in kernel.
 * Added algorithm switch code.
 * Enhanced debugging.
 * 64-bit clean-up.
 *
 * Revision 1.9  1998/11/30 13:22:55  rgb
 * Rationalised all the klips kernel file headers.  They are much shorter
 * now and won't conflict under RH5.2.
 *
 * Revision 1.8  1998/11/25 04:59:06  rgb
 * Add conditionals for no IPIP tunnel code.
 * Delete commented out code.
 *
 * Revision 1.7  1998/10/31 06:50:41  rgb
 * Convert xform ASCII names to no spaces.
 * Fixed up comments in #endif directives.
 *
 * Revision 1.6  1998/10/19 14:44:28  rgb
 * Added inclusion of freeswan.h.
 * sa_id structure implemented and used: now includes protocol.
 *
 * Revision 1.5  1998/10/09 04:32:19  rgb
 * Added 'klips_debug' prefix to all klips printk debug statements.
 *
 * Revision 1.4  1998/08/12 00:11:31  rgb
 * Added new xform functions to the xform table.
 * Fixed minor debug output spelling error.
 *
 * Revision 1.3  1998/07/09 17:45:31  rgb
 * Clarify algorithm not available message.
 *
 * Revision 1.2  1998/06/23 03:00:51  rgb
 * Check for presence of IPIP protocol if it is setup one way (we don't
 * know what has been set up the other way and can only assume it will be
 * symmetrical with the exception of keys).
 *
 * Revision 1.1  1998/06/18 21:27:51  henry
 * move sources from klips/src to klips/net/ipsec, to keep stupid
 * kernel-build scripts happier in the presence of symlinks
 *
 * Revision 1.3  1998/06/11 05:54:59  rgb
 * Added transform version string pointer to xformsw initialisations.
 *
 * Revision 1.2  1998/04/21 21:28:57  rgb
 * Rearrange debug switches to change on the fly debug output from user
 * space.  Only kernel changes checked in at this time.  radij.c was also
 * changed to temporarily remove buggy debugging code in rj_delete causing
 * an OOPS and hence, netlink device open errors.
 *
 * Revision 1.1  1998/04/09 03:06:13  henry
 * sources moved up from linux/net/ipsec
 *
 * Revision 1.1.1.1  1998/04/08 05:35:02  henry
 * RGB's ipsec-0.8pre2.tar.gz ipsec-0.8
 *
 * Revision 0.5  1997/06/03 04:24:48  ji
 * Added ESP-3DES-MD5-96
 *
 * Revision 0.4  1997/01/15 01:28:15  ji
 * Added new transforms.
 *
 * Revision 0.3  1996/11/20 14:39:04  ji
 * Minor cleanups.
 * Rationalized debugging code.
 *
 * Revision 0.2  1996/11/02 00:18:33  ji
 * First limited release.
 *
 *
 */
