#ifdef __KERNEL__

#include <asm/uaccess.h>
#include <asm/atomic.h>

#include <linux/sched.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/limits.h>
#include <linux/smp_lock.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/ipv6.h>
#include <linux/tcp.h>
#include <linux/udp.h>

#else

#include "structs.h"

#endif 


#include <linux/audit.h>
#include <linux/audit_event.h>
#include <linux/acl.h>

#ifdef CONFIG_AUDIT

char res_msk[] = AU_INIT_RES_EVENTMASK;

#ifndef SEEK_CUR
#  define SEEK_CUR 1
#endif

#ifdef CONFIG_AUDIT_PROC
extern unsigned long au_rec_filtered;
#endif

#ifdef __KERNEL__
static
#endif
struct audit_ip_acl acl_ipv4_anchor;

static spinlock_t acl_ipv4_spinlock = SPIN_LOCK_UNLOCKED;
static __u32 netmask[33];

#define DATA_SIZE	sizeof(struct in_addr)

#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)

#undef DATA_SIZE
#define DATA_SIZE	sizeof(struct in6_addr)

#ifdef __KERNEL__
static
#endif
struct audit_ip_acl acl_ipv6_anchor;
static spinlock_t acl_ipv6_spinlock = SPIN_LOCK_UNLOCKED;


#endif 

#endif 


extern unsigned int au_dummy[];

static struct acl_entry acl_host_anchor = ACL_ANCHOR;
static spinlock_t acl_host_spinlock = SPIN_LOCK_UNLOCKED;


static unsigned long acl_count = 1;

#define LARGE_PRIME_NO 2147483647

static struct audit_ip_acl *get_ip_entry(struct audit_ip_acl *anchor, int fam,
                                         void *low, void *high, __u8 cidr,
                                         __u8 direc);
static struct audit_ip_acl *acl_get_ip_entry(struct audit_ip_acl *anchor,
                                             int fam, void *low, void *high,
                                             __u8 cidr, __u8 direc,
                                             unsigned long args,
                                             unsigned int size);
static int umount_entry(struct audit_ip_acl *entry);

int init_netacl(void)
{
	__u32 msk[AU_MSK_SIZE32];
	int err, i;

	netmask[0] = 0x0;
	for (i = 1; i < 33; i++)
		netmask[i] = netmask[i - 1] | (1 << (32 - i));
	
#ifdef CONFIG_AUDIT_NETINIT
	memset((char *) msk, 0xff, AU_MSK_SIZE8);
#else
	memset((char *) msk, 0, AU_MSK_SIZE8);
#endif
		
	memset(&acl_ipv4_anchor, 0, sizeof(struct audit_ip_acl));
	acl_ipv4_anchor.direc = 1;
	err = audit_set_evt_msk(&acl_ipv4_anchor.msk, msk, GFP_KERNEL);

#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	if (!err) {
		memset(&acl_ipv6_anchor, 0, sizeof(struct audit_ip_acl));
		acl_ipv4_anchor.direc = 1;

		err = -ENOMEM;
		acl_ipv6_anchor.tip.ip_rng = kmalloc(sizeof(struct audit_ip_rng),
		                                     GFP_KERNEL);
		if (acl_ipv6_anchor.tip.ip_rng) {
			memset(&acl_ipv6_anchor.tip.ip_rng, 0, sizeof(struct audit_ip_rng));
			err = audit_set_evt_msk(&acl_ipv6_anchor.msk, msk, GFP_KERNEL);
		}
	}
#endif

	return err;
}

inline unsigned long acl_get_hash(char *res)
{
	unsigned long hash;

	for (hash = 0; *res != '\0'; res++)
		hash = (64 * hash + *res) % LARGE_PRIME_NO;
	
	return hash;
}

static struct acl_rel_entry *get_rel(struct acl_entry *p, int id, int type)
{
	struct acl_rel_entry *q, *r = NULL, *s = NULL;

	if (!p)
		return NULL;
	
	q = p->rel;

	while (q && q->type <= type) {
		if (q->type == type)
			if (type == ACL_TYPE_ALL || q->id == id) {
				return q;
			}

		s = q;
		q = q->next;
	}

	r = (struct acl_rel_entry *) kmalloc(sizeof(struct acl_rel_entry),
	                                     GFP_KERNEL);

	if (r) {
		r->type = type;
		r->id = id;

#ifdef CONFIG_AUDIT
		r->v.msk = NULL;
#endif
		r->op_used = 0;
		r->op_mask = 0;

		r->ace = p;

		r->prev = s;

		if (q) {
			r->next = q;
			q->prev = r;
		} else
			r->next = NULL;

		if (s) 
			s->next = r;
		else
			p->rel = r;
	}

	return r;
}

static void *acl_get_entry_file(unsigned long *param, unsigned int size)
{
	int type, id;
	char *name;
	unsigned long hash;
	char cur_dir[PATH_MAX];
	int len, new_sib;
	struct acl_entry *p = &acl_host_anchor, *q;

	if (size < 3)
		return NULL;
	
	name = (char *) param[0];
	id = (int) param[1];
	type = (int) param[2];

	if (type < ACL_TYPE_PID || type > ACL_TYPE_ALL)
		return NULL;
	
	if (!name || *name != '/')
		return NULL;

	while (name && *name == '/')
		name++;

	while (name && *name != '\0') {
		if (!name)
			break;

		q = p;
		p = p->cld;

		for (len = 0;
		     !get_user(cur_dir[len], name) &&
			   cur_dir[len] != '/' && cur_dir[len] != '\0' &&
			   len < (PATH_MAX - 1) && ++name;
			 len++) {
		}

		cur_dir[len] = '\0';

		hash = acl_get_hash(cur_dir);

		new_sib = 0;
		while (p) {
			new_sib = 1;

			if (p->hash == hash && !strcmp(p->name, cur_dir)) {
				while (name && *name == '/')
					name++;

				if (!name || *name == '\0')
					return (void *) get_rel(p, id, type);

				break;
			}

			q = p;
			p = p->sib_r;
		}

		if (!p) {
			p = (struct acl_entry *) kmalloc(sizeof(struct acl_entry),
			                                 GFP_KERNEL);
			if (!p)
				return NULL;

			p->name = (char *) kmalloc(len + 1, GFP_KERNEL);

			if (!p->name)
				return NULL;

			acl_count++;

			strncpy(p->name, cur_dir, len);
			p->name[len] = '\0';
			p->hash = hash;

			p->rel = NULL;
			p->sib_r = NULL;
			p->sib_l = NULL;
			p->cld = NULL;

			if (new_sib) {
				q->sib_r = p;
				p->sib_l = q;
				p->parent = q->parent;
			} else {
				q->cld = p;
				p->parent = q;
			}
		}

		while (name && *name == '/')
			name++;
	}

	return (void *) get_rel(p, id, type);
}



static int acl_check_perm_file(unsigned long *args, unsigned int size)
{
	char *filename;
	int retval;
	
	if (size < 1) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "acl_check_perm_file: wrong size %d\n", size);
#endif
		return -EINVAL;
	}
	
	filename = (char *) args[0];

	retval = -EINVAL;
	if (filename) {
		struct nameidata nd;

		retval = user_path_walk(filename, &nd);
		if (!retval) {
			struct inode *inode = nd.dentry->d_inode;

			retval = -EROFS;
			if (!IS_RDONLY(inode)) {
				retval = -EPERM;
				if (inode->i_uid == current->uid)
					retval = 0;
			}

			path_release(&nd);
		}
	}
#ifdef CONFIG_AUDIT_DEBUG
	else
		printk(KERN_DEBUG "acl_check_perm_file: no filename given\n");
#endif

	return retval;
}


static int acl_set_exec(int cmd, void *entry, unsigned long *param,
                        unsigned int size)
{
	int err = -EINVAL;

	if (size == 4) {
		int val = (int) param[3];

		if (entry && (val == 0 || val == 1)) {
			unsigned long invert = 0, mask = 0;

			err = 0;

			invert |= 1 << ACL_EXECVE;
			invert = ~invert;

			mask |= val << ACL_EXECVE;

			((struct acl_rel_entry *) entry)->op_mask &= invert;
			((struct acl_rel_entry *) entry)->op_mask |= mask;

			mask |= 1 << ACL_EXECVE;

			((struct acl_rel_entry *) entry)->op_used &= invert;
			((struct acl_rel_entry *) entry)->op_used |= mask;
		}
	}
#ifdef CONFIG_AUDIT_DEBUG
	else
		printk(KERN_DEBUG "acl_set_exec: wrong size %d\n", size);
#endif

	return err;
}

#ifdef CONFIG_AUDIT


static int acl_check_perm_losa(unsigned long *args, unsigned int size)
{
	if (!audit_check_capability(AU_CAP_ADM) && !audit_check_capability(AU_CAP_SETID))
		return -EPERM;

	return 0;
}


static int acl_set_losa(int cmd, void *entry_p, unsigned long *param,
                        unsigned int size)
{
	struct acl_rel_entry *entry;
	int err;

	entry = (struct acl_rel_entry *) entry_p;
	if (entry) {
		switch (cmd) {
			case ACL_SET_AUEVT:
				if (size != 5) {
#ifdef CONFIG_AUDIT_DEBUG
					printk(KERN_DEBUG "acl_set_losa: wrong size(1) %d\n", size);
#endif
					err = -EINVAL;

					break;
				}

				err = audit_update_evt_msk(&entry->v.msk, (int) param[3],
				                           (int) param[4], 2, GFP_KERNEL);
				break;
			case ACL_SET_AUMSK: {
				__u32 msk[AU_MSK_SIZE32];

				if (size != 4) {
#ifdef CONFIG_AUDIT_DEBUG
					printk(KERN_DEBUG "acl_set_losa: wrong size(2) %d\n", size);
#endif
					err = -EINVAL;

					break;
				}

				if (copy_from_user((char *) msk, (void *) param[3],
				                   AU_MSK_SIZE8)) {
#ifdef CONFIG_AUDIT_DEBUG
					printk(KERN_DEBUG "acl_set_losa: copy error\n");
#endif
					err = -EFAULT;

					break;
				}

				err = audit_set_evt_msk(&entry->v.msk, (__u32 *) msk,
				                        GFP_KERNEL);
				break;
			}
			default:
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "acl_set_losa: unknown cmd %d\n", cmd);
#endif
				err = -EINVAL;
		}
	} else {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "acl_set_losa: no entry given\n");
#endif
		err = -EFAULT;
	}

	return err;
}


static int ip_acl_update_msk(int fam, struct audit_ip_acl *entry, 
                             unsigned long param, unsigned int size)
{
	struct audit_acl_msk *msk;
	__u16 val_msk[AU_MSK_SIZE16];
	int err, evt, val;
	unsigned long *args;

	args = (unsigned long *) param;
	if (!args) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "ip_acl_update_msk: no args given\n");
#endif
		return -EFAULT;
	}

	val = 0;
	evt = -1;

	switch (size) {
		case 11: {
			int subevt;

			evt = (int) args[8];
			subevt = (int) args[9];
			val = (int) args[10];

			evt -= AU_MIN_NET_EVT;
			if ((evt & 0xfffffff0) || (subevt & 0xfffffff0)) {
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "ip_acl_update_msk: "
				                  "getting evt %d and subevt %d\n",
				       evt, subevt);
#endif
				return -EINVAL;
			}
			if (val != 0 && val != 1) {
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "ip_acl_update_msk: "
				                  "getting val %d\n", val);
#endif
				return -EINVAL;
			}

			evt = (evt << 4) + subevt;

			break;
		}
		case 9: {
			if (copy_from_user((char *) val_msk, (char *) args[8],
			                   AU_MSK_SIZE8)) {
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "ip_acl_update_msk: copy error\n");
#endif
				return -EFAULT;
			}

			break;
		}
		default:
#ifdef CONFIG_AUDIT_DEBUG
			printk(KERN_DEBUG "ip_acl_update_msk: getting %d arguments.\n",
			       size);
#endif
			return -EINVAL;
	}

	msk = NULL;
	err = 0;

	if (res_msk[evt]) {
		struct audit_port_acl *cur_port, *cur_port_prev;
		__u16 l_sport, l_dport, h_sport, h_dport;
		__u8 p_direc;

		l_sport = (__u16) args[3];
		h_sport = (__u16) args[4];
		l_dport = (__u16) args[5];
		h_dport = (__u16) args[6];

		p_direc = (__u8) args[7];
	
		if (!l_sport && !l_dport && h_sport == 0xffff && h_dport == 0xffff)
			goto no_ports;

		for (cur_port = entry->ports, cur_port_prev = NULL; cur_port;
		     cur_port_prev = cur_port, cur_port = cur_port->next) {

			if (l_sport > cur_port->low_sport)
				continue;
			if (l_sport < cur_port->low_sport) {
				cur_port = NULL;
				break;
			}
				
			if (h_sport > cur_port->high_sport)
				continue;
			if (h_sport < cur_port->high_sport) {
				cur_port = NULL;
				break;
			}

			if (l_dport > cur_port->low_dport)
				continue;
			if (l_dport < cur_port->low_dport) {
				cur_port = NULL;
				break;
			}

			if (h_dport > cur_port->high_dport)
				continue;
			if (h_dport < cur_port->high_dport) {
				cur_port = NULL;
				break;
			}

			if (p_direc != cur_port->direc) {
				if (!p_direc)
					continue;
					
				cur_port = NULL;
			}

			break;
		}

		if (!cur_port) {
			cur_port = (struct audit_port_acl *) kmalloc(sizeof(*cur_port),
			                                             GFP_KERNEL);
			if (!cur_port)
				return -ENOMEM;

			memset(cur_port, 0, sizeof(*cur_port));

			cur_port->low_sport = l_sport;
			cur_port->high_sport = h_sport;
			cur_port->low_dport = l_dport;
			cur_port->high_dport = h_dport;

			cur_port->direc = p_direc;

			if (!cur_port_prev) {
				if (entry->ports)
					cur_port->next = entry->ports;

				entry->ports = cur_port;
			} else {
				cur_port->next = cur_port_prev->next;
				cur_port_prev->next = cur_port;
			}
		}

		if (size == 9) {
			err = audit_set_evt_msk(&cur_port->msk, (__u32 *) val_msk,
			                        GFP_KERNEL);
		} else
			err = audit_update_evt_msk(&cur_port->msk, evt, val, 1, GFP_KERNEL);

		return err;
	}

no_ports:
	if (size == 9)
		err = audit_set_evt_msk(&entry->msk, (__u32 *) val_msk, GFP_KERNEL);
	else
		err = audit_update_evt_msk(&entry->msk, evt, val, 1, GFP_KERNEL);

	return err;
}

#ifdef __KERNEL__
static
#endif
int cmp_ip_addr(int fam, void *addr1, void *addr2, __u8 cidr)
{
	if (!addr1 || !addr2) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "cmp_ip_addr: no addr given\n");
#endif
		return -EFAULT;
	}
	
	switch (fam) {
		case AF_INET: {
			struct in_addr *s_addr1, *s_addr2;
			__u32 n_addr1, n_addr2;
	
			s_addr1 = addr1;
			s_addr2 = addr2;

			n_addr1 = ntohl(s_addr1->s_addr) & netmask[cidr];
			n_addr2 = ntohl(s_addr2->s_addr) & netmask[cidr];

			if (n_addr1 < n_addr2)
				return -1;
			if (n_addr1 > n_addr2)
				return 1;

			return 0;
		}
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
		case AF_INET6: {
			struct in6_addr *s_addr1, *s_addr2;
			__u32 n_addr1, n_addr2;
			int i, max_i = 0;
	
			s_addr1 = addr1;
			s_addr2 = addr2;

			if (!cidr)
				return 0;

			if (cidr > 96)
				max_i = 3;
			if (!max_i && (cidr > 64))
				max_i = 2;
			if (!max_i && (cidr > 32))
				max_i = 1;

			for (i = 0; i < max_i; i++) {
				if (ntohl(s_addr1->s6_addr32[i]) < ntohl(s_addr2->s6_addr32[i]))
					return -1;
				if (ntohl(s_addr1->s6_addr32[i]) > ntohl(s_addr2->s6_addr32[i]))
					return 1;
			}

			n_addr1 = ntohl(s_addr1->s6_addr32[3]) & netmask[cidr - (32 * max_i)];
			n_addr2 = ntohl(s_addr2->s6_addr32[3]) & netmask[cidr - (32 * max_i)];

			if (n_addr1 < n_addr2)
				return -1;
			if (n_addr1 > n_addr2)
				return 1;

			return 0;
		}
#endif
	}

#ifdef CONFIG_AUDIT_DEBUG
	printk(KERN_DEBUG "cmp_ip_addr: wrong argument\n");
#endif
	return -EINVAL;
}


static void free_ip_acl(struct audit_ip_acl *p)
{
	if (!p->direc)
		kfree(p->tip.ip_rng);
	
	kfree(p);
}

static struct audit_ip_acl *alloc_ip_acl(char *low, char *high, int fam,
                                         __u8 cidr, __u8 direc)
{
	struct audit_ip_acl *new;

	if (!low) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "alloc_ip_acl: no low address given\n");
#endif
		return ERR_PTR(-EFAULT);
	}

	new = (struct audit_ip_acl *) kmalloc(sizeof(struct audit_ip_acl),
	                                      GFP_KERNEL);
	if (!new) {
		new = ERR_PTR(-ENOMEM);
		goto out;
	}

	memset((char *) new, 0, sizeof(struct audit_ip_acl));

	if (!cmp_ip_addr(fam, low, high, cidr)) {
		high = NULL;
		direc = 1;
	}

	new->cidr = cidr;
	if (direc && (fam == AF_INET)) {
		struct in_addr *tmp;

		tmp = (struct in_addr *) low;
		new->tip.ipv4_addr.s_addr = tmp->s_addr & htonl(netmask[cidr]);
		new->direc = 1;

		goto out;
	}

	new->tip.ip_rng = (struct audit_ip_rng *) kmalloc(sizeof(*new->tip.ip_rng),
		                                              GFP_KERNEL);

	if (!new->tip.ip_rng) {
		kfree(new);
		new = ERR_PTR(-ENOMEM);

		goto out;
	}

	memset((char *) new->tip.ip_rng, 0, sizeof(struct audit_ip_rng));
	new->direc = direc;

	switch (fam) {
		case AF_INET: {
			struct in_addr tmp;

			memcpy((char *) &tmp, (char *) low, sizeof(struct in_addr));
			tmp.s_addr = tmp.s_addr & htonl(netmask[cidr]);

			memcpy((char *) &new->tip.ip_rng->low.ip, (char *) &tmp,
			       sizeof(struct in_addr));

			if (!high)
				goto clean_up;
			memcpy((char *) &tmp, (char *) high, sizeof(tmp));
			tmp.s_addr = tmp.s_addr & htonl(netmask[cidr]);

			memcpy((char *) &new->tip.ip_rng->high.ip, (char *) &tmp,
			       sizeof(struct in_addr));

			break;
		}
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
		case AF_INET6: {
			struct in6_addr tmp;

			memcpy((char *) &tmp, (char *) low, sizeof(tmp));
			if (cidr < 33) {
				tmp.s6_addr32[0] &= htonl(netmask[cidr]);
				tmp.s6_addr32[1] = 0;
				tmp.s6_addr32[2] = 0;
				tmp.s6_addr32[3] = 0;
			} else if (cidr < 65) {
				tmp.s6_addr32[1] &= htonl(netmask[cidr - 32]);
				tmp.s6_addr32[2] = 0;
				tmp.s6_addr32[3] = 0;
			} else if (cidr < 97) {
				tmp.s6_addr32[2] &= htonl(netmask[cidr - 64]);
				tmp.s6_addr32[3] = 0;
			} else
				tmp.s6_addr32[3] &= htonl(netmask[cidr - 96]);

			memcpy((char *) &new->tip.ip_rng->low.ip6, (char *) &tmp,
			       sizeof(struct in6_addr));

			if (!direc) {
				if (!high)
					goto clean_up;

				memcpy((char *) &tmp, (char *) high, sizeof(tmp));
				if (cidr < 33) {
					tmp.s6_addr32[0] &= htonl(netmask[cidr]);
					tmp.s6_addr32[1] = 0;
					tmp.s6_addr32[2] = 0;
					tmp.s6_addr32[3] = 0;
				} else if (cidr < 65) {
					tmp.s6_addr32[1] &= htonl(netmask[cidr - 32]);
					tmp.s6_addr32[2] = 0;
					tmp.s6_addr32[3] = 0;
				} else if (cidr < 97) {
					tmp.s6_addr32[2] &= htonl(netmask[cidr - 64]);
					tmp.s6_addr32[3] = 0;
				} else
					tmp.s6_addr32[3] &= htonl(netmask[cidr - 96]);

				memcpy((char *) &new->tip.ip_rng->high.ip6, (char *) &tmp,
				       sizeof(struct in6_addr));
			}

			break;
		}
#endif
		default:
clean_up:
#ifdef CONFIG_AUDIT_DEBUG
			printk(KERN_DEBUG "alloc_ip_acl: unsupported family %d\n", fam);
#endif
			free_ip_acl(new);
			new = ERR_PTR(-EINVAL);
	}
	
out:
	return new;
}

#ifdef __KERNEL__
static
#endif
void *get_low_addr(int fam, struct audit_ip_acl *cur)
{
	void *retval;

	switch (fam) {
		case AF_INET:
			if (cur->direc || !cur->tip.ip_rng) {
				retval = (void *) &cur->tip.ipv4_addr;
				break;
			}
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
		case AF_INET6:
			if (cur->tip.ip_rng) {
				retval = (void *) &cur->tip.ip_rng->low.ip6;
				break;
			}
#endif
		default:
			retval = NULL;
	}

	return retval;
}


#ifdef __KERNEL__
static
#endif
void *get_high_addr(int fam, struct audit_ip_acl *cur)
{
	void *retval;

	if (cur->direc || !cur->tip.ip_rng)
		return get_low_addr(fam, cur);

	switch (fam) {
	case AF_INET:
		retval = (void *) &cur->tip.ip_rng->high.ip;
		break;
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	case AF_INET6:
		retval = (void *) &cur->tip.ip_rng->high.ip6;
		break;
#endif
	default:
		retval = NULL;
	}

	return retval;
}


static struct audit_ip_acl *create_new_ip_entry(struct audit_ip_acl *cur,
                                                int fam, void *low, void *high,
                                                __u8 cidr, __u8 direc)
{
	struct audit_ip_acl *new;

	new = alloc_ip_acl((char *) low, (char *) high, fam, cidr, direc);
	if (!IS_ERR(new)) {
		new->net = cur->net;

		if (cmp_ip_addr(fam, get_low_addr(fam, cur), low, cidr) == 1) {
			if (cur->net->subnet == cur) {
				cur->net->subnet = new;
			} else {
				new->prev = cur->prev;
				new->prev->next = new;
			}

			new->next = cur;
			cur->prev = new;
		} else  {
			if (cur->next) {
				new->next = cur->next;
				cur->next->prev = new;
			}

			cur->next = new;
			new->prev = cur;
		}
	}

	return new;
}

#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
static int check_cidr_range(__u32 addr, __u8 cidr, __u8 max_cidr, int offset)
{
	if (!cidr)
		return 0;

	if ((offset < 0) || (cidr == max_cidr)) {
		if (((offset >= 0 ) && (addr > ULONG_MAX - offset)) ||
		    ((offset < 0) && (addr < (-1 * offset)))) {

#ifdef CONFIG_AUDIT_DEBUG
			printk(KERN_DEBUG "check_cidr_range: wrong offset %d or addr %d\n",
			       offset, addr);
#endif
			return -EINVAL;
		} else {
			__u32 tmp;

			tmp = 1 << (cidr - 1);
			if ((offset >= 0) && ((ntohl(addr) + (__u32) offset) > (tmp - 1))) {
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "check_cidr_range: "
				                  "wrong offset %d or addr %d\n",
				       offset, addr);
#endif
				return -EINVAL;
			}
		}
	}

	return 0;
}
#endif

static int check_cidr_range32(__u32 addr, __u8 cidr, int offset)
{
	if (!cidr || !offset)
		return 0;

	if (offset > 0) {
		if (cidr == 32) {
			if ((ULONG_MAX - offset) < ntohl(addr)) {
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "check_cidr_range32: "
				                  "wrong offset %d or addr %d\n",
				       offset, addr);
#endif
				return -EINVAL;
			}
		} else {
			if ((ntohl(addr) + (__u32) offset) > ((1 << cidr) - 1)) {
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "check_cidr_range32: "
				       "sum of offset %d and addr %d greater than cidr %d\n",
				       offset, ntohl(addr), ((1 << cidr) - 1));
#endif
				return -EINVAL;
			}
		}

		return 0;
	}

	if ((ntohl(addr) + (__u32) offset) < 0) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "check_cidr_range32: "
		                  "sum of offset %d and addr %d less than zero\n",
		       offset, addr);
#endif
		return -EINVAL;
	}

	return 0;
}


#ifdef __KERNEL__
static
#endif
int seek_ip_addr(int fam, __u8 cidr, void *dest, void *src, int offset,
                 int whence)
{
	int err;

	if (!dest) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "seek_ip_addr: no destination given\n");
#endif
		return -EFAULT;
	}

	err = 0;
	if (cidr && (whence == SEEK_CUR)) {
		char data[DATA_SIZE];

		memset(data, 0, DATA_SIZE);

		switch (fam) {
			size_t size;

			case AF_INET: {
				struct in_addr *tmp, *d_tmp;
				__u32 new_addr;

				d_tmp = (struct in_addr *) data;

				tmp = (struct in_addr *) src;
				new_addr = (ntohl(tmp->s_addr) >> (32 - cidr));

				err = check_cidr_range32(htonl(new_addr), cidr, offset);
				if (err)
					break;

				new_addr += offset;

				d_tmp->s_addr = htonl(new_addr << (32 - cidr));
				size = sizeof(struct in_addr);

				memcpy(dest, data, size);
				break;
			}
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
			case AF_INET6: {
				struct in6_addr *tmp, *d_tmp;
				__u32 new_addr;

				d_tmp = (struct in6_addr *) data;

				tmp = (struct in6_addr *) src;
				if (cidr > 96) {
					new_addr = (ntohl(tmp->s6_addr32[3]) >> (128 - cidr));

					err = check_cidr_range(new_addr, cidr, 128, offset);
					if (err)
						break;

					new_addr += offset;

					d_tmp->s6_addr32[3] = htonl(new_addr << (128 - cidr));
					d_tmp->s6_addr32[2] = tmp->s6_addr32[2];
					d_tmp->s6_addr32[1] = tmp->s6_addr32[1];
					d_tmp->s6_addr32[0] = tmp->s6_addr32[0];
					size = sizeof(struct in6_addr);
				} else if (cidr > 64) {
					new_addr = (ntohl(tmp->s6_addr32[2]) >> (96 - cidr));
					err = check_cidr_range(new_addr, cidr, 96, offset);
					if (err)
						break;

					new_addr += offset;

					d_tmp->s6_addr32[2] = htonl(new_addr << (96 - cidr));
					d_tmp->s6_addr32[1] = tmp->s6_addr32[1];
					d_tmp->s6_addr32[0] = tmp->s6_addr32[0];
					size = sizeof(struct in6_addr);
				} else if (cidr > 32) {
					new_addr = (ntohl(tmp->s6_addr32[1]) >> (6 - cidr));
					
					err = check_cidr_range(new_addr, cidr, 64, offset);
					if (err)
						break;

					new_addr += offset;

					d_tmp->s6_addr32[1] = htonl(new_addr << (64 - cidr));
					d_tmp->s6_addr32[0] = tmp->s6_addr32[0];
					size = sizeof(struct in6_addr);
				} else {
					new_addr = (ntohl(tmp->s6_addr32[0]) >> (32 - cidr));

					err = check_cidr_range(new_addr, cidr, 32, offset);
					if (err)
						break;

					new_addr += offset;

					d_tmp->s6_addr32[0] = htonl(new_addr << (32 - cidr));
					size = sizeof(struct in6_addr);
				}

				memcpy(dest, data, size);
				break;
			}
#endif
			default:
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "seek_ip_addr: unsupported family %d\n", fam);
#endif
				err = -EINVAL;
		}
	}

	return err;
}


static int ip_split_entry(int fam, struct audit_ip_acl **cur_p,
                          struct audit_ip_acl **head_p, void *low)
{
	struct audit_ip_acl *head, *cur;
	int err;

	if (!cur_p || !head_p || !low) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "ip_split_entry: zero pointer given\n");
#endif
		return -EFAULT;
	}
	
	cur = *cur_p;
	if (!cur) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "ip_split_entry: no cur entry given\n");
#endif
		return -EFAULT;
	}
	
	if (cur->direc) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "ip_split_entry: cur has no range.\n");
#endif
		return -EINVAL;
	}

	head = cur;
	cur = alloc_ip_acl((char *) low, get_high_addr(fam, head), fam, cur->cidr,
	                   (!cmp_ip_addr(fam, low, get_high_addr(fam, head),
					                cur->cidr) ? 1 : 0));
	if (IS_ERR(cur))
		return PTR_ERR(cur);

	cur->msk = head->msk;
	if (head->msk)
		atomic_inc(&head->msk->count);

	if (head->ports) {
		struct audit_port_acl *new, *orig, *prev;

		for (orig = head->ports, prev = NULL; orig; orig = orig->next) {
			new = (struct audit_port_acl *) kmalloc(sizeof(*new), GFP_KERNEL);
			if (!new) {
				free_ip_acl(cur);
				return -ENOMEM;
			}

			memcpy((char *) new, (char *) orig, sizeof(*new));
			atomic_inc(&orig->msk->count);
			new->next = NULL;

			if (prev)
				prev->next = new;
			else
				cur->ports = new;

			prev = new;
		}
	}

	err = seek_ip_addr(fam, head->cidr, (void *) head->tip.ip_rng->high.raw,
					   low, -1, SEEK_CUR);
	if (err) {
		free_ip_acl(cur);
		return err;
	}

	if (!cmp_ip_addr(fam, get_low_addr(fam, head), get_high_addr(fam, head),
	                 head->cidr)) {

		if (fam == AF_INET) {
			struct in_addr tmp;
			memcpy((void *) &tmp, (void *) &head->tip.ip_rng->low.ip,
			       sizeof(tmp));
			kfree(head->tip.ip_rng);

			memcpy((void *) &head->tip.ipv4_addr, (void *) &tmp, sizeof(tmp));
		}

		head->direc = 1;
	}

	if (head->subnet) {
		struct audit_ip_acl *tmp;

		for (tmp = head->subnet; tmp; tmp = tmp->next) {
			if (cmp_ip_addr(fam, get_high_addr(fam, tmp),
			                  get_low_addr(fam, cur), tmp->cidr) == -1)
				continue;

			if (cmp_ip_addr(fam, get_low_addr(fam, tmp),
			                     get_low_addr(fam, cur), tmp->cidr) != -1)
				break;

			{
				struct audit_ip_acl *tmp2;

				tmp2 = NULL;
				err = ip_split_entry(fam, &tmp, &tmp2, get_low_addr(fam, cur));
				if (err)
					return err;

				if (!tmp2->net)
					tmp2->net = head;

				break;
			}
		}

		if (tmp) {
			if (head->subnet == tmp)
				head->subnet = NULL;
			else
				tmp->prev->next = NULL;

			cur->subnet = tmp;
			tmp->prev = NULL;
			for ( ; tmp; tmp = tmp->next)
				tmp->net = cur;
		}
	}
	
	cur->net = head->net;
	cur->prev = head;
	if (head->next) {
		head->next->prev = cur;
		cur->next = head->next;
	}
	head->next = cur;
	
	*head_p = head;
	*cur_p = cur;

	return 0;
}


static struct audit_ip_acl *ip_match_first(struct audit_ip_acl *cur, int fam,
                                           void *low, void *high, __u8 direc,
                                           unsigned long param,
                                           unsigned int size)
{
	struct audit_ip_acl *head = NULL;
	char tmp[DATA_SIZE];
	int err;

	if (direc ||
	    cmp_ip_addr(fam, get_high_addr(fam, cur), high, cur->cidr) == 1) {
		if (!cur->direc) {
			if (!high)
				high = low;

			err = seek_ip_addr(fam, cur->cidr, tmp, high, 1, SEEK_CUR);
			if (err)
				return ERR_PTR(err);

			err = ip_split_entry(fam, &cur, &head, tmp);
			if (err)
				return ERR_PTR(err);
		} else
			head = cur;

		return head;
	}

	err = seek_ip_addr(fam, cur->cidr, tmp, get_high_addr(fam, cur), 1,
	                   SEEK_CUR);
	if (err)
		return ERR_PTR(err);

	err = ip_acl_update_msk(fam, cur, param, size);
	if (err)
		return ERR_PTR(err);
	
	if (cur->next && cmp_ip_addr(fam, get_low_addr(fam, cur->next), high,
	                             cur->cidr) != 1) {
		err = seek_ip_addr(fam, cur->cidr, low, tmp, 0, SEEK_CUR);
		if (err)
			return ERR_PTR(err);

		return cur->next;
	}

	return create_new_ip_entry(cur, fam, tmp, high, cur->cidr, direc);
}


static int create_new_ip_before(struct audit_ip_acl *cur, int fam, void *low,
				                __u8 cidr, __u8 direc, unsigned long param,
                                unsigned int size)
{
	int err;
	struct in6_addr tmp;

	err = seek_ip_addr(fam, cidr, (void *) &tmp,
	                   get_low_addr(fam, cur), -1, SEEK_CUR);
	if (!err) {
		struct audit_ip_acl *new;

		new = create_new_ip_entry(cur, fam, low, (void *) &tmp, cidr, direc);
		if (!IS_ERR(new)) {
			err = seek_ip_addr(fam, cidr, low, get_low_addr(fam, cur), 0,
			                   SEEK_CUR);
			if (!err)
				err = ip_acl_update_msk(fam, new, param, size);
		} else
			err = PTR_ERR(new);
	}

	return err;
}


#define ACL_IPADDR_MATCH			1001
#define ACL_IPADDR_OVERLAY_BEGIN	1002
#define ACL_IPADDR_MATCH_FIRST		1003
#define ACL_IPADDR_INBOUND			1004
#define ACL_IPADDR_MATCH_LAST		1005
#define ACL_IPADDR_OVERLAY_END		1006
#define ACL_IPADDR_BEFORE			1007
#define ACL_IPADDR_AFTER			1008

static int check_rng_ip_addr(int fam, struct audit_ip_acl *cur, void *low,
                             void *high)
{
	void *cur_low, *cur_high;
	int cmp_low;

	if (!cur) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "check_rng_ip_addr: no cur entry given\n");
#endif
		return -EFAULT;
	}

	cur_low = get_low_addr(fam, cur);
	if (!cur->direc)
		cur_high = get_high_addr(fam, cur);
	else
		cur_high = cur_low;

	if (!cur_low || !cur_high) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_NOTICE "check_rng_ip_addr: "
		                   "no addr for cur_low (%x08) or cur_high (%x08) \n",
		       (unsigned int) cur_low, (unsigned int) cur_high);
#endif
		return -EFAULT;
	}

	if (IS_ERR(cur_low)) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_NOTICE "check_rng_ip_addr: error in cur_low.\n");

#endif
		return PTR_ERR(cur_low);
	}

	if (IS_ERR(cur_high)) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_NOTICE "check_rng_ip_addr: error in cur_high.\n");
#endif
		return PTR_ERR(cur_high);
	}

	if (!high)
		high = low;

	cmp_low = cmp_ip_addr(fam, cur_low, low, cur->cidr);
	switch (cmp_low) {
		int tmp;

		case 0:
			tmp = cmp_ip_addr(fam, cur_high, high, cur->cidr);

			if (!tmp)
				return ACL_IPADDR_MATCH;

			return ACL_IPADDR_MATCH_FIRST;
		case 1:
			tmp = cmp_ip_addr(fam, high, cur_low, cur->cidr);
			switch (tmp) {
				case -1:
					return ACL_IPADDR_BEFORE;

				case 0:
					if (cmp_ip_addr(fam, high, cur_high, cur->cidr) == 0)
						return ACL_IPADDR_MATCH_LAST;

					return ACL_IPADDR_OVERLAY_END;
				case 1:
					tmp = cmp_ip_addr(fam, high, cur_high, cur->cidr);
					switch (tmp) {
						case -1:
							return ACL_IPADDR_OVERLAY_END;
	
						case 0:
							return ACL_IPADDR_MATCH_LAST;

						case 1:
							return ACL_IPADDR_INBOUND;
					}
			}
		case -1:
			tmp = cmp_ip_addr(fam, cur_high, low, cur->cidr);
			switch (tmp) {
				case -1:
					return ACL_IPADDR_AFTER;

				case 0:
					tmp = cmp_ip_addr(fam, cur_high, high, cur->cidr);
					switch (tmp) {
						case 0:
							return ACL_IPADDR_MATCH_LAST;

						case -1:
							return ACL_IPADDR_OVERLAY_BEGIN;
						default:
#ifdef CONFIG_AUDIT_DEBUG
							printk(KERN_DEBUG "check_rng_ip_addr: "
							                  "unsupported switch-branch (2) %d\n",
							       tmp);
#endif
							return -EINVAL;
					}
				case 1:
					tmp = cmp_ip_addr(fam, cur_high, high, cur->cidr);
					switch (tmp) {
						case -1:
							return ACL_IPADDR_OVERLAY_BEGIN;
	
						case 0:
							return ACL_IPADDR_MATCH_LAST;

						case 1:
							return ACL_IPADDR_INBOUND;
					}
			}
	}

#ifdef CONFIG_AUDIT_DEBUG
	printk(KERN_DEBUG "check_rng_ip_addr: unsupported switch-branch (1) %d\n",
	       cmp_low);
#endif
	return -EINVAL;
}


static struct audit_ip_acl *create_ip_entry(struct audit_ip_acl *anchor,
                                            int fam, void *low, void *high,
                                            __u8 cidr, __u8 direc,
                                            unsigned long param,
                                            unsigned int size)
{
	struct audit_ip_acl *cur, *prev, *new;
	int pos, err;

	if (!low)
		return ERR_PTR(-EFAULT);

	if (!high)
		high = low;

	for (err = 0, cur = anchor, pos = ACL_IPADDR_AFTER, prev = NULL;
	     cur && !err;
	     cur = cur->next) {

		prev = cur;

		pos = check_rng_ip_addr(fam, cur, low, high);
		switch (pos) {
			case ACL_IPADDR_MATCH:
				return cur;
			case ACL_IPADDR_OVERLAY_BEGIN:
				new = NULL;

				err = ip_split_entry(fam, &cur, &new, low);
				if (err)
					return ERR_PTR(err);

				err = ip_acl_update_msk(fam, cur, param, size);
				if (err)
					return ERR_PTR(err);
				
				err = seek_ip_addr(fam, cidr, low, get_high_addr(fam, cur), 1,
				                   SEEK_CUR);
				prev = cur;

				break;
			case ACL_IPADDR_MATCH_FIRST:
				cur = ip_match_first(cur, fam, low, high, direc, param, size);
				if (IS_ERR(cur))
					return cur;
				if (!cmp_ip_addr(fam, low, get_low_addr(fam, cur), cidr))
					return cur;

				break;
			case ACL_IPADDR_INBOUND:
				new = NULL;

				if (cmp_ip_addr(fam, get_high_addr(fam, cur), high,
				                cidr) == 1) {
					err = ip_split_entry(fam, &cur, &new, low);
					if (err)
						return ERR_PTR(err);
				} else {
					err = create_new_ip_before(cur, fam, low, cidr, direc,
					                           param, size);
					if (err)
						return ERR_PTR(err);
				}

				return ip_match_first(cur, fam, low, high, direc, param, size);
			case ACL_IPADDR_MATCH_LAST:
				if (cmp_ip_addr(fam, get_low_addr(fam, cur), low, cidr) == 1) {
					err = create_new_ip_before(cur, fam, low, cidr, direc,
					                           param, size);
				} else
					err = ip_split_entry(fam, &cur, &new, low);

				if (err)
					return ERR_PTR(err);

				return cur;
			case ACL_IPADDR_OVERLAY_END: {
				char tmp[DATA_SIZE];
				new = NULL;

				err = create_new_ip_before(cur, fam, low, cidr, direc, param,
				                           size);
				if (err)
					return ERR_PTR(err);

				err = seek_ip_addr(fam, cidr, tmp, high, 1, SEEK_CUR);
				if (err)
					return ERR_PTR(err);

				err = ip_split_entry(fam, &cur, &new, tmp);
				if (err)
					return ERR_PTR(err);

				return cur;
			}
			case ACL_IPADDR_BEFORE:
				return create_new_ip_entry(cur, fam, low, high, cidr, direc);
			case ACL_IPADDR_AFTER:
				break;
			default:
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_NOTICE "create_ip_entry: "
				                   "error while executing check_rng_ip_addr\n");
#endif
				return ERR_PTR(pos);
		}
	}

	if (err)
		return ERR_PTR(err);

	if (!cur)
		return create_new_ip_entry(prev, fam, low, high, cidr, direc);

	return create_new_ip_entry(cur, fam, low, high, cidr, direc);
}


static struct audit_ip_acl *get_ip_layer(struct audit_ip_acl *anchor, int fam, 
                                         void *low, void *high, __u8 cidr, 
                                         __u8 direc, unsigned long param,
                                         unsigned int size);

static struct audit_ip_acl *layer_match_first(struct audit_ip_acl *cur, int fam,
                                              void *low, void *high, __u8 direc,
                                              unsigned long param,
                                              unsigned int size)
{
	struct audit_ip_acl *head = NULL;
	char tmp[DATA_SIZE];
	int err;

	if (direc ||
	    cmp_ip_addr(fam, get_high_addr(fam, cur), high, cur->cidr) == 1) {

		if (direc)
			err = seek_ip_addr(fam, cur->cidr, tmp, (struct in_addr *) low, 1,
			                   SEEK_CUR);
		else
			err = seek_ip_addr(fam, cur->cidr, tmp, (struct in_addr *) high, 1,
			                   SEEK_CUR);

		if (err)
			goto err_out;

		err = ip_split_entry(fam, &cur, &head, tmp);
		if (err)
			goto err_out;

		return head;
	}

	err = seek_ip_addr(fam, cur->cidr, tmp, get_high_addr(fam, cur), 1,
	                   SEEK_CUR);
	if (!err) {
		unsigned long args[11];
		struct audit_ip_acl *new, *foo;
		__u8 cidr;

		head = NULL;

		if (!param || copy_from_user((void *) args, (void *) param,
		                             size * sizeof(long))) {
			err = -EFAULT;
			goto err_out;
		}
			
		cidr = (__u8) args[2];

		foo = cur;
		if (cur->next && cmp_ip_addr(fam, get_low_addr(fam, cur->next),
		                             high, cidr) != 1) {
			char tmp_high[DATA_SIZE];
			__u8 direc;

			direc = !cmp_ip_addr(fam, tmp, high, cidr) ? 1 : 0;
			head = foo;
			for (;;) {
				foo = foo->next;

				if (cmp_ip_addr(fam, high, get_high_addr(fam, foo),
				                foo->cidr) != 1) {
					err = seek_ip_addr(fam, cidr, tmp_high, high, 0, SEEK_CUR);
					if (err)
						goto err_out;
				} else {
					err = seek_ip_addr(fam, cidr, tmp_high,
					                   get_high_addr(fam, foo), 0, SEEK_CUR);
					if (err)
						goto err_out;

					switch (fam) {
						case AF_INET: {
							__u32 msk, addr;

							msk = ~netmask[foo->cidr];
							msk &= netmask[cidr];

							addr = ((struct in_addr *) tmp_high)->s_addr;
							addr |= htonl(msk);

							((struct in_addr *) tmp_high)->s_addr = addr;

							break;
						}
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
						case AF_INET6: {
							struct in6_addr msk, addr;
							int i, inx;

							memset(&msk, 0, sizeof(struct in6_addr));
							memset(&addr, 0, sizeof(struct in6_addr));

							for (i = 0; i <= (foo->cidr - 1) / 32; i++) {
								if (i == ((foo->cidr - 1) / 32)) {
									inx = foo->cidr - (i * 32);
									msk.s6_addr32[i] = ~netmask[inx];
								} else
									msk.s6_addr32[i] = 0xffffffff;

								msk.s6_addr32[i] &= netmask[cidr - (i * 32)];
							}

							memcpy(&addr, tmp_high, sizeof(struct in6_addr));

							for (i = 0; i < 4; i++)
								addr.s6_addr32[i] |= htonl(msk.s6_addr32[i]);

							memcpy(tmp_high, &addr, sizeof(struct in6_addr));

							break;
						}
#endif
						default:
#ifdef CONFIG_AUDIT_DEBUG
							printk(KERN_DEBUG "layer_match_first: "
							                  "unsupported family %d\n", fam);
#endif
							err = -EINVAL;
							goto err_out;
					}
				}

				if (cmp_ip_addr(fam, tmp, get_low_addr(fam, foo),
				                foo->cidr) == -1) {
					char tmp_high2[DATA_SIZE];

					err = seek_ip_addr(fam, cidr, tmp_high2,
					                   get_low_addr(fam, foo), -1, SEEK_CUR);
					if (err)
						goto err_out;

					if (!foo->prev) {
#ifdef CONFIG_AUDIT_DEBUG
						printk(KERN_NOTICE "layer_match_first: "
						                    "internal error\n");
#endif
						return ERR_PTR(-EINVAL);
					}

					head = create_new_ip_entry(foo->prev, fam, tmp, tmp_high2,
					                           foo->cidr, 0);
					if (IS_ERR(head))
						return head;

					new = alloc_ip_acl((char *) tmp, (char *) tmp_high2, fam,
					                   cidr, 0);
					if (IS_ERR(new))
						return new;

					new->net = head;
					head->subnet = new;
					err = ip_acl_update_msk(fam, new, param, size);
					if (err)
						goto err_out;

					err = seek_ip_addr(fam, cidr, tmp, get_low_addr(fam, foo),
					                   0, SEEK_CUR);
					if (err)
						goto err_out;
				}

				new = acl_get_ip_entry(foo, fam, tmp, tmp_high, cidr, direc,
				                       param, size);
				if (IS_ERR(new))
					return new;

				err = ip_acl_update_msk(fam, new, param, size);
				if (err)
					goto err_out;

				if (!foo->next || cmp_ip_addr(fam, get_low_addr(fam, foo->next),
			                                  high, cidr) == 1) {
					break;
				}
				
				err = seek_ip_addr(fam, foo->cidr, tmp, get_high_addr(fam, foo),
				                   1, SEEK_CUR);
				if (err)
					goto err_out;
			}

			if (cmp_ip_addr(fam, tmp_high, high, cidr) == 0)
				memset(tmp, 0, DATA_SIZE);
		}

		if (cmp_ip_addr(fam, tmp, get_high_addr(fam, foo), cur->cidr) == 1) {
			head = create_new_ip_entry(cur, fam, tmp, high, cur->cidr, 0);
			if (IS_ERR(head))
				return head;

			new = alloc_ip_acl((char *) tmp, (char *) high, fam, cidr, 0);
			if (IS_ERR(new))
				return new;

			new->net = head;
			head->subnet = new;
			err = ip_acl_update_msk(fam, new, param, size);
			if (err)
				goto err_out;
		}

		if (!cur->next) {
#ifdef CONFIG_AUDIT_DEBUG
			printk(KERN_DEBUG
			       "layer_match_first: debug, this should never happen \n");
#endif
			err = -EINVAL;
			goto err_out;
		}

		err = seek_ip_addr(fam, cidr, high, get_low_addr(fam, cur->next), -1,
		                   SEEK_CUR);
		if (err)
			goto err_out;
	
		return cur;
	}

err_out:
	return ERR_PTR(err);
}


static int create_new_layer_before(struct audit_ip_acl *cur, int fam, void *low,
				                   __u8 cidr, __u8 direc, unsigned long param,
                                   unsigned int size)
{
	int err;
	char tmp[DATA_SIZE];

	err = seek_ip_addr(fam, cidr, tmp, get_low_addr(fam, cur), -1, SEEK_CUR);
	if (!err) {
		unsigned long args[11];

		if (param && !copy_from_user((void *) args, (void *) param,
		                             size * sizeof(long))) {
			struct audit_ip_acl *new, *new2;
			__u8 cidr2;

			cidr2 = (__u8) args[2];

			new = create_new_ip_entry(cur, fam, low, tmp, cidr, direc);
			if (IS_ERR(new))
				return PTR_ERR(new);

			err = seek_ip_addr(fam, cidr2, tmp, get_low_addr(fam, cur), -1,
			                   SEEK_CUR);
			if (err)
				return err;

			new2 = alloc_ip_acl((char *) low, (char *) tmp, fam, cidr2, 0);
			if (IS_ERR(new2))
				return PTR_ERR(new2);

			new2->net = new;
			new->subnet = new2;
			err = ip_acl_update_msk(fam, new2, param, size);
			if (err)
				return err;

			return 0;
		}

		return -EFAULT;
	}

	return err;
}


static int distribute_ip_insert_before(int fam, __u8 cidr,
                                       struct audit_ip_acl *net,
                                       struct audit_ip_acl *entry)
{
	struct audit_ip_acl *tmp;
	void *low, *high;
	int pos;

	low = get_low_addr(fam, entry);
	high = get_high_addr(fam, entry);

	for (tmp = net->prev; tmp; tmp = tmp->prev) {
		pos = check_rng_ip_addr(fam, tmp, low, high);
		switch (pos) {
			struct audit_ip_acl *tmp2;
	
			case ACL_IPADDR_OVERLAY_BEGIN:
				break;
			case ACL_IPADDR_MATCH_LAST: {
				struct audit_ip_acl *prev;
				int pos2;

				if (!tmp->subnet) {
					tmp->subnet = entry;
					entry->prev = NULL;
					entry->net = tmp;

					goto out;
				}

				if (tmp->subnet->cidr != entry->cidr) {
#ifdef CONFIG_AUDIT_DEBUG
					printk(KERN_DEBUG "distribute_ip_insert_before: "
				                      "unsupported if-condition (1).\n");
#endif
					goto out;
				}

				for (prev = NULL, tmp2 = tmp->subnet; tmp2;
				     tmp2 = tmp2->next) {

					prev = tmp2;

					pos2 = check_rng_ip_addr(fam, tmp2, low, high);
					switch (pos2) {
						case ACL_IPADDR_AFTER:
							continue;
						case ACL_IPADDR_BEFORE:
							if (prev)
								goto insert_before;

							entry->next = tmp->subnet;
							entry->net = tmp;
							entry->prev = NULL;

							if (entry->next)
								entry->next->prev = entry;

							goto out;
						default:
#ifdef CONFIG_AUDIT_DEBUG
							printk(KERN_DEBUG "distribute_ip_insert_before: "
			                       "untouched switch-branch (2) %d\n", pos2);
#endif
					}
				}

				if (!prev) {
#ifdef CONFIG_AUDIT_DEBUG
					printk(KERN_DEBUG "distribute_ip_insert_before: "
					                  "unsupported if-condition (2)\n");
#endif
					goto out;
				}

insert_before:
				entry->next = prev->next;
				entry->net = prev->net;
				entry->prev = prev;

				prev->next = entry;
				if (entry->next)
					entry->next->prev = entry;

				goto out;
			}
			case ACL_IPADDR_MATCH:
				if (tmp->subnet) {
					for (tmp2 = tmp->subnet; tmp2->next; tmp2 = tmp2->next) {
					}
	
					tmp2->next = entry;
					entry->prev = tmp2;
				} else
					tmp->subnet = entry;
	
				entry->net = tmp;
				goto out;
			default:
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "distribute_ip_insert_before: "
				                  "untouched switch-branch (1) %d\n", pos);
#endif
				return -EINVAL;
		}
	}

	tmp = alloc_ip_acl(low, high, fam, cidr, entry->direc);
	if (!IS_ERR(tmp)) {
		tmp->net = net->net;

		tmp->subnet = entry;
		entry->net = tmp;

		tmp->next = net;
		if (net->prev) {
			tmp->prev = net->prev;
			tmp->prev->next = tmp;
		} else
			net->net->subnet = tmp;
			
		net->prev = tmp;
	} else
		return PTR_ERR(tmp);

out:
	return 0;
}


static int distribute_ip_insert_behind(int fam, __u8 cidr,
                                       struct audit_ip_acl *net,
                                       struct audit_ip_acl *entry)
{
	struct audit_ip_acl *tmp;
	void *low_addr, *high_addr;
	int pos;

	low_addr = get_low_addr(fam, entry);
	high_addr = get_high_addr(fam, entry);

	for (tmp = net->next; tmp; tmp = tmp->next) {
		pos = check_rng_ip_addr(fam, tmp, low_addr, high_addr);
		switch (pos) {
			struct audit_ip_acl *prev, *tmp2;

			case ACL_IPADDR_AFTER:
				break;
			case ACL_IPADDR_MATCH_FIRST:
			case ACL_IPADDR_OVERLAY_BEGIN: {
				char foo1[DATA_SIZE], foo2[DATA_SIZE];
				int err;

				err = seek_ip_addr(fam, tmp->cidr, foo1, 
				                   get_high_addr(fam, tmp), 1, SEEK_CUR);
				if (err)
					return err;

				err = seek_ip_addr(fam, entry->cidr, foo2, foo1, -1, SEEK_CUR);
				if (err)
					return err;

				if (!entry->direc) {
					tmp2 = NULL;
					err = ip_split_entry(fam, &entry, &tmp2, foo1);
					if (err)
						return err;

					umount_entry(tmp2);
					umount_entry(entry);

					tmp2->net = tmp;
					if (!tmp->subnet) {
						tmp->subnet = tmp2;
					} else {
						for (prev = tmp->subnet; prev->next;
						     prev = prev->next) {
						}

						prev->next = tmp2;
						tmp2->prev = prev;
					}
				}

				err = seek_ip_addr(fam, tmp->cidr, foo2,
				                   get_high_addr(fam, entry), 0, SEEK_CUR);
				if (err)
					return err;

				tmp2 = alloc_ip_acl(foo1, foo2, fam, tmp->cidr, 0);
				if (!IS_ERR(tmp2)) {
					tmp2->net = tmp->net;

					tmp2->next = tmp->next;
					if (tmp->next)
						tmp->next->prev = tmp2;
					tmp->next = tmp2;
					tmp2->prev = tmp;

					tmp2->subnet = entry;
					entry->net = tmp2;
				} else
					return PTR_ERR(tmp2);

				goto out;
			}
			case ACL_IPADDR_MATCH_LAST:
			case ACL_IPADDR_MATCH:
				entry->net = tmp;

				if (!tmp->subnet) {
					tmp->subnet = entry;
					goto out;
				}

				for (tmp2 = tmp->subnet, prev = NULL, pos = ACL_IPADDR_BEFORE;
				     tmp2 && pos == ACL_IPADDR_BEFORE;
					 prev = tmp2, tmp2 = tmp2->next) {

					pos = check_rng_ip_addr(fam, entry, get_low_addr(fam, tmp2),
					                        get_high_addr(fam, tmp2));
				}

				if (!prev) {
					entry->next = tmp->subnet;
					entry->prev = NULL;
					if (entry->next)
						entry->next->prev = entry;
	
					tmp->subnet = entry;
				} else {
					entry->next = prev->next;
					if (entry->next) {
						entry->next->prev = entry;
					}

					prev->next = entry;
					entry->prev = prev;
				}

				goto out;
			case ACL_IPADDR_BEFORE:
				tmp2 = alloc_ip_acl((char *) get_low_addr(fam, entry),
				                    (char *) get_high_addr(fam, entry), fam,
				                    cidr, entry->direc);
				if (!IS_ERR(tmp2)) {
					if (!tmp->prev)
						return -EFAULT;

					tmp2->next = tmp;
					tmp2->prev = tmp->prev;
					
					tmp->prev->next = tmp2;
					tmp->prev = tmp2;
					
					tmp2->net = tmp->net;
					tmp2->subnet = entry;
					entry->net = tmp2;
				} else
					return PTR_ERR(tmp);

				goto out;
			default:
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "distribute_ip_insert_behind: "
				                  "untouched switch branch (1) %d \n", pos);
#endif
				break;
		}
	}

	tmp = alloc_ip_acl((char *) get_low_addr(fam, entry),
	                   (char *) get_high_addr(fam, entry), fam, cidr,
	                   entry->direc);
	if (!IS_ERR(tmp)) {
		struct audit_ip_acl *tmp2;

		tmp->net = net->net;
		tmp->subnet = entry;
		entry->net = tmp;

		for (tmp2 = net; tmp2->next; tmp2 = tmp2->next) {
		}

		tmp2->next = tmp;
		tmp->prev = tmp2;
	} else
		return PTR_ERR(tmp);

out:
	return 0;
}


static int umount_entry(struct audit_ip_acl *entry)
{
	if (!entry)
		return -EFAULT;
	
	if (entry->next) {
		struct audit_ip_acl *succ;

		succ = entry->next;
 		if (entry->prev) {
			succ->prev = entry->prev;
			entry->prev->next = entry;
		} else {
			succ->prev = NULL;
			if (succ->net && succ->net->subnet == entry)
				succ->net->subnet = succ;
		}
	} else {
		if (entry->prev) {
			struct audit_ip_acl *prev;

			prev = entry->prev;
			if (entry->next) {
				prev->next = entry->next;
				prev->next->prev = prev;
			} else
				prev->next = NULL;
		}
	}

	entry->net = NULL;
	entry->next = NULL;
	entry->prev = NULL;

	return 0;
}


static int distribute_insert_cur(int fam, struct audit_ip_acl *net,
                                 struct audit_ip_acl *cur)
{
	struct audit_ip_acl *tmp;
	
	cur->net = net;
	if (!net->subnet) {
		net->subnet = cur;
		return 0;
	}

	for (tmp = net->subnet; tmp; tmp = tmp->next) {
		if (cmp_ip_addr(fam, get_low_addr(fam, tmp), get_high_addr(fam, cur),
		                cur->cidr) < 0) {

			if (!tmp->next) {
				tmp->next = cur;
				cur->prev = tmp;

				break;
			}

			continue;
		}

		if (tmp->prev) {
			tmp->prev->next = cur;
			cur->prev = tmp->prev;
		} else {
			tmp->net->subnet = cur;
			cur->prev = NULL;
		}
		
		tmp->prev = cur;
		cur->next = tmp;

		break;
	}
	
	return 0;
}


static int distribute_match_first(int fam, __u8 cidr, struct audit_ip_acl *net,
                                  struct audit_ip_acl *cur)
{
	struct audit_ip_acl *tmp;
	char low[DATA_SIZE];
	int err;

	err = seek_ip_addr(fam, cidr, (void *) low, get_high_addr(fam, net), 1,
	                   SEEK_CUR);
	if (!err) {
		err = ip_split_entry(fam, &cur, &tmp, low);
		if (!err) {
 			umount_entry(cur);

			err = distribute_ip_insert_behind(fam, cidr, net, cur);
			if (!err) {
 				umount_entry(tmp);
				distribute_insert_cur(fam, net, tmp);
			}
		}
	}

	return err;
}


static int distribute_match_last(int fam, __u8 cidr, struct audit_ip_acl *net,
                                 struct audit_ip_acl *cur)
{
	struct audit_ip_acl *tmp;
	int err;

	err = ip_split_entry(fam, &cur, &tmp, get_low_addr(fam, net));
	if (!err) {
 		umount_entry(tmp);
		err = distribute_ip_insert_before(fam, cidr, net, tmp);
		if (!err) {
 			umount_entry(cur);
			err = distribute_insert_cur(fam, net, cur);
		}
	}

	return err;
}


static int distribute_ip_subnet(int fam, struct audit_ip_acl *net,
                                struct audit_ip_acl *subnet)
{
	struct audit_ip_acl *cur, *tmp, *next;
	int err, pos;
	__u8 cidr;

	cidr = net->cidr;
	for (err = 0, cur = subnet; cur && !err; cur = next) {
		next = cur->next;

		pos = check_rng_ip_addr(fam, net, get_low_addr(fam, cur),
		                        get_high_addr(fam, cur));
		switch (pos) {
			case ACL_IPADDR_MATCH:
 				umount_entry(cur);
				err = distribute_insert_cur(fam, net, cur);
				break;
			case ACL_IPADDR_OVERLAY_BEGIN:
				err = ip_split_entry(fam, &net, &tmp, get_low_addr(fam, cur));
				if (!err)
					err = distribute_match_first(fam, cidr, net, cur);

				break;
			case ACL_IPADDR_MATCH_FIRST:
				if (cmp_ip_addr(fam, get_high_addr(fam, net),
				                get_high_addr(fam, cur), net->cidr) != -1) {
					char low[DATA_SIZE];

					err = seek_ip_addr(fam, cidr, (void *) low,
				                   get_high_addr(fam, cur), 1, SEEK_CUR);
					if (!err) {
						err = ip_split_entry(fam, &net, &tmp, low);
						if (!err) {
 							umount_entry(cur);
							err = distribute_insert_cur(fam, tmp, cur);
						}
					}

					break;
				}

				err = distribute_match_first(fam, cidr, net, cur);
				break;
			case ACL_IPADDR_INBOUND: {
				char low[DATA_SIZE];

				if (cmp_ip_addr(fam, get_low_addr(fam, net),
				                get_low_addr(fam, cur), net->cidr) == -1) {
					err = ip_split_entry(fam, &net, &tmp,
					                     get_low_addr(fam, cur));
					if (!err) {
						err = seek_ip_addr(fam, cidr, (void *) low,
						                   get_high_addr(fam, cur), 1,
						                   SEEK_CUR);
						if (!err) {
							err = ip_split_entry(fam, &net, &tmp, low);
							if (!err) {
 								umount_entry(cur);
								err = distribute_insert_cur(fam, tmp, cur);
							}
						}
					}

					break;
				}
								
				err = ip_split_entry(fam, &cur, &tmp, get_low_addr(fam, net));
				if (!err) {
 					umount_entry(tmp);
					err = distribute_ip_insert_before(fam, cidr, net, tmp);
					if (!err) {	
						err = seek_ip_addr(fam, cidr, (void *) low,
						                   get_high_addr(fam, net), 1,
						                   SEEK_CUR);
						if (!err) {
							err = ip_split_entry(fam, &cur, &tmp, low);
							if (!err) {
 								umount_entry(cur);
								err = distribute_ip_insert_behind(fam, cidr,
								                                  net, cur);
								if (!err) {
 									umount_entry(tmp);
									err = distribute_insert_cur(fam, net, tmp);
								}
							}
						}
					}
				}

				break;
			}
			case ACL_IPADDR_OVERLAY_END: {
				char low[DATA_SIZE];

				err = seek_ip_addr(fam, cidr, (void *) low,
				                   get_high_addr(fam, cur), 1, SEEK_CUR);
				if (!err) {
					err = ip_split_entry(fam, &net, &tmp, low);
					if (!err) {
						net = tmp;
						err = distribute_match_last(fam, cidr, net, cur);
					}
				}

				break;
			}
			case ACL_IPADDR_MATCH_LAST:
				if (cmp_ip_addr(fam, get_low_addr(fam, net),
				                get_low_addr(fam, cur), net->cidr) == 1) {
					err = distribute_match_last(fam, cidr, net, cur);
					break;
				}

				err = ip_split_entry(fam, &net, &tmp, get_low_addr(fam, cur));
				if (!err) {
 					umount_entry(cur);
					err = distribute_insert_cur(fam, net, cur);
				}

				break;
			case ACL_IPADDR_BEFORE: {
				int pos2;

 				umount_entry(cur);
				err = distribute_ip_insert_before(fam, cidr, net, cur);
				for ( ; net; net = net->prev) {
					pos2 = check_rng_ip_addr(fam, net, get_low_addr(fam, cur),
					                         get_high_addr(fam, cur));

					if (pos2 != ACL_IPADDR_BEFORE)
						break;
				}

				break;
			}
			case ACL_IPADDR_AFTER: {
				int pos2;

 				umount_entry(cur);

				err = distribute_ip_insert_behind(fam, cidr, net, cur);
				for ( ; net; net = net->next) {
					pos2 = check_rng_ip_addr(fam, net, get_low_addr(fam, cur),
					                         get_high_addr(fam, cur));

					if (pos2 != ACL_IPADDR_AFTER)
						break;
				}

				break;
			}
			default:
				if (pos < 0)
					err = pos;

#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "distribute_ip_subnet: "
				                  "unsupported switch-branch %d\n", pos);
#endif
		}
	}

	return err;
}


static struct audit_ip_acl *get_ip_layer(struct audit_ip_acl *anchor, int fam, 
                                         void *low, void *high, __u8 cidr, 
                                         __u8 direc, unsigned long param,
                                         unsigned int size)
{
	struct audit_ip_acl *net, *cur, *new, *prev;
	int pos, err;

	if (!anchor || !low)
		return ERR_PTR(-EFAULT);

	if (!cidr)
		return anchor;

	if (!high) {
		high = low;
		direc = 1;
	} else {
		if (!cmp_ip_addr(fam, low, high, cidr))
			direc = 1;
	}

	net = anchor;
	pos = ACL_IPADDR_MATCH;
	for (cur = net->subnet, err = 0, prev = NULL; !err && cur;
	     net = cur->net, prev = NULL) {

		if (cur->cidr > cidr) {
			cur = NULL;
			break;
		}

		if (cur->cidr == cidr)
			break;

		for ( ; cur && !err && pos != ACL_IPADDR_BEFORE; cur = cur->next) {
			prev = cur;

			pos = check_rng_ip_addr(fam, cur, low, high);
			switch (pos) {
				case ACL_IPADDR_MATCH:
					net = cur;
					cur = cur->subnet;
					goto break_loop;
				case ACL_IPADDR_OVERLAY_BEGIN:
					new = NULL;

					err = ip_split_entry(fam, &cur, &new, low);
					if (err)
						return ERR_PTR(err);

				case ACL_IPADDR_MATCH_FIRST:
					if (cmp_ip_addr(fam, get_high_addr(fam, cur), high,
					                cur->cidr) != -1) {
						if (cur->subnet && cur->subnet->cidr <= cidr) {
							net = cur;
							cur = cur->subnet;

							pos = ACL_IPADDR_MATCH;
							goto break_loop;
						}
					}

					cur = layer_match_first(cur, fam, low, high, direc,
											param, size);
					if (IS_ERR(cur))
						return cur;

					if (!cmp_ip_addr(fam, low, get_low_addr(fam, cur),
					                 cur->cidr)) {
						net = cur;
						cur = cur->subnet;

						pos = ACL_IPADDR_MATCH;
						goto break_loop;
					}

					break;
				case ACL_IPADDR_INBOUND:
					new = NULL;

					if (cmp_ip_addr(fam, get_high_addr(fam, cur), high,
					                cur->cidr) == 1) {

						err = ip_split_entry(fam, &cur, &new, low);
						if (err)
							return ERR_PTR(err);
					} else {
						err = create_new_layer_before(cur, fam, low, cur->cidr,
						                              direc, param, size);
						if (err)
							return ERR_PTR(err);

						err = seek_ip_addr(fam, cidr, low,
						                   get_low_addr(fam, cur), 0, SEEK_CUR);
						if (err)
							return ERR_PTR(err);
					}

					cur = layer_match_first(cur, fam, low, high, direc, param,
					                        size);
					if (IS_ERR(cur))
						return cur;
						
					if (!cmp_ip_addr(fam, low, get_low_addr(fam, cur),
					                 cur->cidr)) {
						net = cur;
						cur = cur->subnet;

						pos = ACL_IPADDR_MATCH;
						goto break_loop;
					}

					break;
				case ACL_IPADDR_MATCH_LAST:
					if (cmp_ip_addr(fam, get_low_addr(fam, cur), low,
					                cur->cidr) == 1) {
						err = create_new_layer_before(cur, fam, low, cur->cidr,
						                              direc, param, size);

						if (err)
							return ERR_PTR(err);

						err = seek_ip_addr(fam, cidr, low,
						                   get_low_addr(fam, cur), 0, SEEK_CUR);
					} else {
						if (cur->subnet && cur->subnet->cidr <= cidr) {
							net = cur;
							cur = cur->subnet;

							pos = ACL_IPADDR_MATCH;
							goto break_loop;
						}
						err = ip_split_entry(fam, &cur, &new, low);
					}

					if (err)
						return ERR_PTR(err);

					net = cur;
					cur = cur->subnet;

					pos = ACL_IPADDR_MATCH;
					goto break_loop;
				case ACL_IPADDR_OVERLAY_END: {
					char tmp[DATA_SIZE];
					new = NULL;
				
					err = create_new_layer_before(cur, fam, low, cur->cidr,
					                              direc, param, size);
					if (err)
						return ERR_PTR(err);

					err = seek_ip_addr(fam, cur->cidr, tmp, high, 1, SEEK_CUR);
					if (err)
						return ERR_PTR(err);

					err = ip_split_entry(fam, &cur, &new, tmp);
					if (err)
						return ERR_PTR(err);

					err = seek_ip_addr(fam, cidr, low, get_low_addr(fam, new),
					                   0, SEEK_CUR);
					if (err)
						return ERR_PTR(err);

					net = new;
					cur = new->subnet;

					pos = ACL_IPADDR_MATCH;
					goto break_loop;
				}
				case ACL_IPADDR_AFTER:
					break;
				case ACL_IPADDR_BEFORE:
					new = create_new_ip_entry(cur, fam, low, high, cur->cidr,
					                          direc);

					if (IS_ERR(new))
						return new;

					net = new;
					cur = new->subnet;

					pos = ACL_IPADDR_MATCH;
					goto break_loop;
				default:
					err = pos;
			}

			if (!cur)
				break;
		}	
break_loop:
		if (!cur)
			break;
	}

	if (err)
		return ERR_PTR(err);

	if (!cur) {
		if (!net->subnet) {
			cur = alloc_ip_acl((char *) low, (char *) high, fam, cidr, direc);

			if (IS_ERR(cur))
				return cur;

			cur->net = net;
			net->subnet = cur;

			return cur;
		}

		if (pos == ACL_IPADDR_MATCH) {
			struct audit_ip_acl *subnet;

			subnet = net->subnet;

			cur = alloc_ip_acl((char *) low, (char *) high, fam, cidr, direc);

			if (IS_ERR(cur))
				return cur;

			cur->net = net;
			net->subnet = cur;

			err = distribute_ip_subnet(fam, cur, subnet);
			if (err) {
				net->subnet = subnet;
				free_ip_acl(cur);

				return ERR_PTR(err);
			}

			return cur;
		}

		if (cmp_ip_addr(fam, low, high, cidr)) {
		}

		cur = alloc_ip_acl((char *) low, (char *) high, fam, net->subnet->cidr,
		                   direc);

		if (IS_ERR(cur))
			return cur;

		if (!prev) {
			cur->next = net->subnet;
			net->subnet = cur;
		} else {
			cur->next = prev->next;
			prev->next = cur;
			cur->prev = prev;
		}

		if (cur->next)
			cur->next->prev = cur;
		cur->net = net;

		prev = cur;
		cur = alloc_ip_acl((char *) low, (char *) high, fam, cidr, direc);
		if (!IS_ERR(cur)) {
			prev->subnet = cur;
			cur->net = prev;
		}
	}

	return cur;
}


static struct audit_ip_acl *acl_get_ip_entry(struct audit_ip_acl *anchor,
                                             int fam, void *low, void *high,
                                             __u8 cidr, __u8 direc,
                                             unsigned long args,
                                             unsigned int size)
{
	struct audit_ip_acl *cur;
	
	cur = get_ip_entry(anchor, fam, low, high, cidr, direc);
	if (!cur) {
		cur = get_ip_layer(anchor, fam, low, high, cidr, direc, args, size);
		if (!IS_ERR(cur))
			cur = create_ip_entry(cur, fam, low, high, cidr, direc, args, size);
	}

	return cur;
}


static struct audit_ip_acl *get_ip_entry(struct audit_ip_acl *anchor, int fam,
                                         void *low, void *high, __u8 cidr,
                                         __u8 direc)
{
	if (anchor && low) {
		struct audit_ip_acl *cur;
		int pos;
		if (!high)
			high = low;
	
		for (cur = anchor; cur; ) {
			pos = check_rng_ip_addr(fam, cur, low, high);
			if (pos == ACL_IPADDR_MATCH) {
				if (cur->cidr > cidr)
					return NULL;

				if (cur->cidr == cidr)
					return cur;
	
				cur = cur->subnet;
				continue;
			}
			
			cur = cur->next;
		}

		return NULL;
	}

	return ERR_PTR(-EFAULT);
}


static void *acl_get_entry_ipv4(unsigned long *param, unsigned int size)
{
	struct audit_ip_acl *retval;
	unsigned long args[11];
	__u8 cidr;

	if (size != 9 && size != 11) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "acl_get_entry_ipv4: wrong size given %d\n", size);
#endif
		return ERR_PTR(-EINVAL);
	}

	if (!param || copy_from_user((void *) args, (void *) param,
	                             size * sizeof(long)))
		return ERR_PTR(-EFAULT);

	if (size == 9 && verify_area(VERIFY_READ, (void *) args[8], AU_MSK_SIZE8))
		return ERR_PTR(-EFAULT);

	retval = &acl_ipv4_anchor;

	cidr = (__u8) args[2];
	if (cidr) {
		struct in_addr low;

		retval = ERR_PTR(-EFAULT);
		if (!copy_from_user((char *) &low, (char *) args[0],
			               sizeof(struct in_addr)))
		{
			struct in_addr high;
			void *high_p = NULL;
			__u8 direc = 1;

			if (args[1]) {
				if (copy_from_user((char *) &high, (char *) args[1],
				                   sizeof(struct in_addr)))
					return ERR_PTR(-EFAULT);	
	
				high_p = (void *) &high;
				direc = 0;
			}

			retval = acl_get_ip_entry(&acl_ipv4_anchor, AF_INET, (void *) &low,
			                          high_p, cidr, direc, (unsigned long) args,
			                          size);
		}
	}

	return retval;
}


static void *acl_get_entry_ipv6(unsigned long *param, unsigned int size)
{
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	struct audit_ip_acl *retval;
	unsigned long args[11];
	__u8 cidr;

	if (size != 9 && size != 11) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "acl_get_entry_ipv6: wrong size given %d\n", size);
#endif
		return ERR_PTR(-EINVAL);
	}

	if (!param || copy_from_user((void *) args, (void *) param,
	                             size * sizeof(long)))
		return ERR_PTR(-EFAULT);

	if (size == 9 && verify_area(VERIFY_READ, (void *) args[8], AU_MSK_SIZE8))
		return ERR_PTR(-EFAULT);

	retval = &acl_ipv6_anchor;

	cidr = (__u8) args[2];
	if (cidr) {
		struct in6_addr low;

		retval = ERR_PTR(-EFAULT);
		if (!copy_from_user((char *) &low, (char *) args[0],
			               sizeof(struct in6_addr)))
		{
			struct in6_addr high;
			void *high_p = NULL;
			__u8 direc = 1;

			if (args[1]) {
				if (copy_from_user((char *) &high, (char *) args[1],
				                   sizeof(struct in6_addr)))
					return ERR_PTR(-EFAULT);	
	
				high_p = (void *) &high;
				direc = 0;
			}

			retval = acl_get_ip_entry(&acl_ipv6_anchor, AF_INET6, (void *) &low,
			                          high_p, cidr, direc, (unsigned long) args,
			                          size);
		}
	}

	return retval;
#else
	return ERR_PTR(-ENOSYS);
#endif
}


static int acl_set_ipv4(int cmd, void *entry, unsigned long *param,
                        unsigned int size)
{
	return ip_acl_update_msk(AF_INET, (struct audit_ip_acl *) entry, 
	                         (unsigned long) param, size);
}


static int acl_set_ipv6(int cmd, void *entry, unsigned long *param,
                        unsigned int size)
{
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	return -ENOSYS;
#else
	return ip_acl_update_msk(AF_INET6, (struct audit_ip_acl *) entry,
	                         (unsigned long) param, size);
#endif 
}


static int acl_host_lock(void)
{
	spin_lock(&acl_host_spinlock);
	return 0;
}


static int acl_ipv4_lock(void)
{
	spin_lock(&acl_ipv4_spinlock);
	return 0;
}


static int acl_ipv6_lock(void)
{
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	spin_lock(&acl_ipv6_spinlock);
#endif
	return 0;
}


static int acl_host_unlock(void)
{
	spin_unlock(&acl_host_spinlock);
	return 0;
}


static int acl_ipv4_unlock(void)
{
	spin_unlock(&acl_ipv4_spinlock);
	return 0;
}


static int acl_ipv6_unlock(void)
{
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	spin_unlock(&acl_ipv6_spinlock);
#endif
	return 0;
}


int (*acl_check_perm[]) (unsigned long *, unsigned int) = {
	acl_check_perm_file,
	acl_check_perm_losa,
	acl_check_perm_losa,
	acl_check_perm_losa
};

void * (*acl_get_entry[]) (unsigned long *, unsigned int) = {
	acl_get_entry_file,
	acl_get_entry_file,
	acl_get_entry_ipv4,
	acl_get_entry_ipv6
};

int (*acl_set[]) (int, void *, unsigned long *, unsigned int) = {
	acl_set_exec,
	acl_set_losa,
	acl_set_ipv4,
	acl_set_ipv6
};

int (*acl_lock[]) (void) = {
	acl_host_lock,
	acl_host_lock,
	acl_ipv4_lock,
	acl_ipv6_lock
};

int (*acl_unlock[]) (void) = {
	acl_host_unlock,
	acl_host_unlock,
	acl_ipv4_unlock,
	acl_ipv6_unlock
};


#endif


static int acl_check_cmd(int cmd)
{
#ifdef CONFIG_AUDIT
	switch (cmd) {
		case ACL_SET_EXEC:
			return 0;
		case ACL_SET_AUEVT:
		case ACL_SET_AUMSK:
			return 1;
		case ACL_SET_IPV4_AUEVT:
		case ACL_SET_IPV4_AUMSK:
			return 2;
		case ACL_SET_IPV6_AUEVT:
		case ACL_SET_IPV6_AUMSK:
			return 3;
	}
#else
	if (cmd == ACL_SET_EXEC)
		return 0;
#endif

#ifdef CONFIG_AUDIT_DEBUG
	printk(KERN_DEBUG "acl_check_cmd: unsupported cmd %d\n", cmd);
#endif
	return -EINVAL;
}


asmlinkage int sys_acl_ctl(int cmd, unsigned long args, unsigned int size)
{
	int err;
	unsigned long *param;

	err = -ENOMEM;
	param = (unsigned long *) kmalloc(size * sizeof(long), GFP_KERNEL);
	if (param) {
		err = -EFAULT;
		if (!copy_from_user(param, (void *) args, size * sizeof(long))) {
			err = acl_check_cmd(cmd);
			if (err >= 0) {
				int acl_type = err;

				acl_lock[acl_type]();

				err = acl_check_perm[acl_type](param, size);
				if (!err) {
					void *entry;

					entry = acl_get_entry[acl_type](param, size);
					if (!IS_ERR(entry)) {
						err = acl_set[acl_type](cmd, entry, param, size);
					} else
						err = PTR_ERR(entry);
				}

				acl_unlock[acl_type]();
			}
		}

		kfree(param);
	}
		
	return err;
}


#define ACL_CHECK_OP	300
#define ACL_CHECK_AUMSK	301

int acl_check_rel_entry(struct acl_rel_entry *entry, int type, int evt)
{
	int retval = 0;

#ifdef CONFIG_AUDIT
	if (type == ACL_CHECK_AUMSK) {
		retval = -1;

		if (entry->v.msk) {
			__u32 tmp;

			tmp = 1;
			tmp &= entry->v.msk->perm_msk32[evt / 32] >> (evt % 32);
			if (tmp) {
				tmp = 3;
				tmp &= entry->v.msk->val_msk32[evt / 16] >> (2 * (evt % 16));

				retval = (int) tmp;
			}
		}

		return retval;
	}
#endif

	if ((entry->op_used >> evt) & 1) {
		if ((entry->op_mask >> evt) & 1)
			retval = 1;
	}
	
	return retval;
}


static int acl_check_rel(struct acl_entry *entry, int type, int event)
{
	int ret;
	struct acl_rel_entry *q;

	if (!entry) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "acl_check_rel: no entry given\n");
#endif
		return 0;
	}

	for (q = entry->rel, ret = -1; q; q = q->next) {
		switch (q->type) {
			case ACL_TYPE_PID:
				if (q->id == current->pid) {
					ret = acl_check_rel_entry(q, type, event);
					q = NULL;
				}

				break;
			case ACL_TYPE_UID:
#ifdef CONFIG_AUDIT
				if ((type == ACL_CHECK_AUMSK && q->id == current->audit->aid)
					|| (type == ACL_CHECK_OP && q->id == current->uid))
#else
				if (q->id == current->uid)
#endif
				{
					ret = acl_check_rel_entry(q, type, event);
					q = NULL;
				}

				break;
			case ACL_TYPE_GID:
				if (q->id == current->audit->agid) {
					ret = acl_check_rel_entry(q, type, event);
					q = NULL;
				}

				break;
			default:
				ret = acl_check_rel_entry(q, type, event);
				q = NULL;
		}

		if (!q)
			break;
	}
	
	return ret;
}

int acl_check(char *name, int type, int event)
{
	struct acl_entry *p = &acl_host_anchor;
	char cur_dir[PATH_MAX];
	unsigned long hash;
	int ret, i;

	if (!name || *(name++) != '/')
		return 0;

	if (!p)
		return 0;

	if (!spin_trylock(&acl_host_spinlock))
		return -EBUSY;

	ret = acl_check_rel(p, type, event);
	if (ret < 0)
		ret = 0;

	p = p->cld;

	while (p) {
		for (i = 0;
		     name && (cur_dir[i] = *(name++)) != '/' && cur_dir[i++] != '\0';
		      ) {
		}

		cur_dir[i] = '\0';

		if (cur_dir[0] == '\0')
			break;

		hash = acl_get_hash(cur_dir);

		while (p) {
			if (p->hash == hash && !strcmp(p->name, cur_dir)) {
				int new_ret = 0;

				new_ret = acl_check_rel(p, type, event);
				if (new_ret >= 0)
					ret = new_ret;

				break;
			}

			p = p->sib_r;
		}

		if (p)
			p = p->cld;
	}

	spin_unlock(&acl_host_spinlock);

	return ret;
}


int acl_check_op(char *name, int evt)
{
	int ret;

	ret = acl_check(name, ACL_CHECK_OP, evt);
	if (ret < 0)
		return 0;
	
	return ret;
}


int acl_check_host(char *name, int evt, int status)
{
	int ret;

	ret = acl_check(name, ACL_CHECK_AUMSK, evt);
	switch (ret) {
		case 0:
			break;
		case 1:
			ret = 0;

			if (!status || status > 0)
				ret = 1;
			break;
		case 2:
			ret = 0;

			if (status < 0)
				ret = 1;
			break;
		default:
			ret = 1;
	}

#ifdef CONFIG_AUDIT_PROC
	if (ret)
		audit_evt_filtered(AU_HOST_REC);
#endif
	
	return ret;
}


static struct audit_port_acl *get_port_acl(struct audit_port_acl *anchor,
                                           int major_evt, int event,
                                           struct sk_buff *skb)
{
	struct audit_port_acl *retval, *port;
	__u16 s_port, d_port;
	__u64 size, new_size;

	if (!anchor || !skb)
		return NULL;
	
	s_port = 0;
	d_port = 0;

	retval = anchor;
	switch (major_evt) {
		case AUE_TCP_RCV:
			if (skb && skb->h.th) {
				s_port = skb->h.th->source;
				d_port = skb->h.th->dest;
			} else
				retval = NULL;

			break;
		case AUE_UDP_RCV:
			if (skb && skb->h.uh) {
				s_port = skb->h.uh->source;
				d_port = skb->h.uh->dest;
			} else
				retval = NULL;

			break;
		default:
			retval = NULL;
	}

	size = 0xffffffffffffffff;
	for (port = retval, retval = NULL; port; port = port->next) {
		new_size = 1;

		if (s_port < port->low_sport || s_port > port->high_sport ||
		    d_port < port->low_dport || d_port > port->high_dport) {

			if (!port->direc ||
			    d_port < port->low_sport || d_port > port->high_sport ||
			    s_port < port->low_dport || s_port > port->high_dport) {

				continue;
			}

			new_size = 2;
		}

		new_size = new_size * (port->high_sport - port->low_sport);
		new_size = new_size * (port->high_dport - port->low_dport);

		if (new_size <= size) {
			if (new_size == size) {
				if (audit_evt_val(port->msk, event, 1) < 0)
					continue;
			}

			size = new_size;
			retval = port;
		}
	}

	return retval;
}


static int acl_check_ip_entry(struct audit_ip_acl *entry, int major_evt,
                              int evt, struct sk_buff *skb)
{
	struct audit_port_acl *port;
	int retval, tmp;

	tmp = audit_evt_val(entry->msk, evt, 1);
	if (tmp < 0)
		return tmp;

	retval = tmp;
	if (res_msk[major_evt]) {
		port = get_port_acl(entry->ports, major_evt, evt, skb);
		if (port) {
			tmp = audit_evt_val(port->msk, evt, 1);
			if (tmp < 0)
				return tmp;

			retval = tmp;
		}
	}

	return retval;
}


int acl_check_net_ipv4(unsigned int evt, unsigned int subevt, void *skb_p)
{
	int event, retval;
	struct sk_buff *skb = skb_p;

	if (!skb)
		return -EFAULT;

	event = ((evt - AU_MIN_NET_EVT) << 4) + subevt;
	if (event & 0xffffff00) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "acl_check_net_ipv4: bad event given %d\n", event);
#endif
		return -EINVAL;
	}

	retval = -EBUSY;
	if (spin_trylock(&acl_ipv4_spinlock)) {
		retval = acl_check_ip_entry(&acl_ipv4_anchor, evt, event, skb);
		if (retval < 0)
			goto unlock_out;
	
		if (skb->nh.iph && acl_ipv4_anchor.subnet) {
			struct in_addr s_addr;
			struct audit_ip_acl *cur;

			s_addr.s_addr = skb->nh.iph->saddr;
			for (cur = acl_ipv4_anchor.subnet; cur; cur = cur->subnet) {
				for ( ; cur; cur = cur->next) {
					if (cur->direc &&
				    	cur->tip.ipv4_addr.s_addr != s_addr.s_addr)
						continue;

					if (cmp_ip_addr(AF_INET, get_low_addr(AF_INET, cur),
						            (void *) &s_addr, cur->cidr) == 1 ||
						cmp_ip_addr(AF_INET, get_high_addr(AF_INET, cur),
						            (void *) &s_addr, cur->cidr) == 1) {

						continue;
					}

					retval = acl_check_ip_entry(cur, evt, event, skb);
					if (retval < 0)
						goto unlock_out;

					break;
				}

				if (!cur)
					break;
			}
		}

unlock_out:
		spin_unlock(&acl_ipv4_spinlock);
	}

	return retval;
}


int acl_check_net_ipv6(unsigned int evt, unsigned int subevt, void *skb_p)
{
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	int event, retval;
	struct sk_buff *skb = skb_p;

	if (skb)
		return -EFAULT;

	event = ((evt - AU_MIN_NET_EVT) << 4) + subevt;
	if (event & 0xffffff00) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "acl_check_net_ipv6: bad event given\n");
#endif
		return -EINVAL;
	}

	retval = -EBUSY;
	if (spin_trylock(&acl_ipv6_spinlock)) {
		retval = acl_check_ip_entry(&acl_ipv6_anchor, evt, event, skb);
		if (retval < 0)
			goto unlock_out;

		if (skb->nh.ipv6h && acl_ipv6_anchor.subnet) {
			struct in6_addr s_addr;
			struct audit_ip_acl *cur;

			memcpy((char *) &s_addr, (char *) &skb->nh.ipv6h->saddr,
			       sizeof(struct in6_addr));

			for (cur = acl_ipv6_anchor.subnet; cur; cur = cur->subnet) {
				for ( ; cur; cur = cur->next) {
					if (cur->direc &&
					    !cmp_ip_addr(AF_INET6, get_low_addr(AF_INET6, cur),
						             (void *) &s_addr, cur->cidr)) {

						continue;
					}

					if (cmp_ip_addr(AF_INET6, get_low_addr(AF_INET6, cur),
						            (void *) &s_addr, cur->cidr) == 1 ||
						cmp_ip_addr(AF_INET6, get_high_addr(AF_INET6, cur),
						            (void *) &s_addr, cur->cidr) == 1) {
	
						continue;
					}

					retval = acl_check_ip_entry(cur, evt, event, skb);
					if (retval < 0)
						goto unlock_out;

					break;
				}
	
				if (!cur)
					break;
			}
		}
	
unlock_out:
		spin_unlock(&acl_ipv6_spinlock);
	}

	return retval;
#else
	return -ENOSYS;
#endif
}
