#include <asm/param.h>
#include <asm/uaccess.h>

#include <linux/string.h>
#include <linux/sched.h>
#include <linux/utsname.h>
#include <linux/limits.h>
#include <linux/mm.h>

#include <linux/acl.h>
#include <linux/audit.h>
#include <linux/audit_event.h>

#include "buffer.h"

extern struct audit_dyn_buf *au_env_buf;

int audit_sysboot(void)
{
	struct audit_record *rec;
	struct audit_host_record *data;

	if (!AU_SVC_INITIALIZED(&au_svc)) {
		return 0;
	}

	if (!current->audit || !audit_doit(AUE_SYSTEMBOOT, 0)) {
#ifdef CONFIG_AUDIT_DEBUG
		if (current->audit)
			printk("audit_sysboot: event filtered for user %d.\n",
			       current->audit->aid);
#endif
		return 0;
	}

	if ((rec = audit_get_free_record(AU_HOST_REC, 0, AUE_SYSTEMBOOT)) == NULL) {
		return 0;
	}

	data = (struct audit_host_record *) rec->data;
	if (!data) {
		rec->state = AU_UNUSED;
		return 0;
	}
	
	data->ruid = 0;
	data->rgid = 0;
	data->euid = 0;
	data->egid = 0;
	data->aid = 0;
	data->session = 0;
	data->from = 0;
	data->tty = 0;
	data->owner = 0;
	data->gowner = 0;
	data->dev = 0;
	data->perm = 0;
	data->inode = 0;

	rec->event = AUE_SYSTEMBOOT;

	data->status = AU_SYSCALL_SUCCESSFULL;

	data->arg1 = CURRENT_TIME - (jiffies / HZ);
	data->arg2 = 0;
	data->env = NULL;

	data->file_len = 0;
	data->file2_len = 0;

	rec->state = AU_FULL;

	return 0;
}

int audit_setaid(int status, int aid, int old_aid)
{
	struct audit_record *rec;
	struct audit_host_record *data;

	if ((rec = audit_get_free_record(AU_HOST_REC, 0, AUE_SETAUID)) == NULL)
		return 0;

	data = (struct audit_host_record *) rec->data;
	if (!data) {
		rec->state = AU_UNUSED;
		return 0;
	}

	audit_fill_static(data);
	audit_fill_filedata(data, NULL);

	rec->event = AUE_SETAUID;

	if (!status) {
		data->aid = old_aid;
		data->status = AU_SYSCALL_SUCCESSFULL;
	} else
		data->status = AU_SYSCALL_FAILED;

	data->arg1 = aid;
	data->arg2 = 0;

	data->env = NULL;

	data->file_len = 0;
	data->file2_len = 0;

	rec->state = AU_FULL;

	return 0;
}

static int audit_host_event(unsigned int event_id,
                            struct audit_app_event *app_event)
{
	int len;
	struct audit_record *rec;
	struct audit_host_record *data;

	if ((rec = audit_get_free_record(AU_HOST_REC, 1, event_id)) == NULL)
		return 0;
	
	data = (struct audit_host_record *) rec->data;
	if (!data) {
		rec->state = AU_UNUSED;
		return 0;
	}
	
	data->status = 0;
	if (app_event)
		data->status = app_event->status;

	if (!audit_doit(event_id, data->status)) {
		rec->state = AU_UNUSED;
		return 0;
	}

	audit_fill_static(data);

	rec->event = event_id;

	data->arg1 = 0;
	data->arg2 = 0;
	data->env = NULL;

	data->file_len = 0;
	data->file2_len = 0;

	audit_fill_cfiledata(data, NULL);

	if (!app_event) {
		rec->state = AU_FULL;
		return 0;
	}

	data->arg1 = app_event->arg1;
	data->arg2 = app_event->arg2;

	len = STRLEN_USER(app_event->file);
	if (len < PATH_MAX && (len - 1) > 0) {
		audit_fill_cfiledata(data, app_event->file);
		audit_resolve_filename(data->file, &data->file_len, app_event->file,
		                       len);
	}

	len = STRLEN_USER(app_event->file2);
	if (len < PATH_MAX && (len - 1) > 0)
		audit_resolve_filename(data->file2, &data->file2_len, app_event->file2,
		                       len);
	
	if (app_event->env) {
		void *src[2];

		src[0] = (void *) app_event->env;
		src[1] = NULL;

		audit_fill_dyn_buf2(au_env_buf, &data->env, src, 1, AU_U_SPACE);
	}

	rec->state = AU_FULL;

	return 0;
}

int sys_audit_event(int type, int id, void *data_p)
{
	int retval;

	if (id < 100 ||id > AU_MAX_EID)
		return -EINVAL;

	if (!audit_check_capability(AU_CAP_MSG))
		return -EPERM;

	retval =  -EFAULT;
	if (data_p) {
		switch (type) {
			case AU_HOST_REC: {
				struct audit_app_event *data;

				data = (struct audit_app_event *) data_p;
				if (audit_doit(id, data->status))
					retval = audit_host_event(id, data);

				break;
			}
			case AU_NET_REC:
				retval = -ENOSYS;
				break;
			default:
				retval = -EINVAL;
		}
	}

	return retval;
}
