#include <asm/uaccess.h>
#include <asm/ptrace.h>
#include <asm/unistd.h>

#include <linux/mm.h>
#include <linux/limits.h>
#include <linux/file.h>

#include <linux/audit.h>
#include <linux/audit_event.h>

#include "buffer.h"

#ifdef CONFIG_ACL
#include <linux/acl.h>
#endif

extern struct audit_dyn_buf *au_env_buf;

static int __audit_fork(int status, unsigned int evt)
{
	struct audit_record *rec;

	rec = audit_get_free_record(AU_HOST_REC, 1, evt);
	if (rec) {
		struct audit_host_record *data;

		data = (struct audit_host_record *) rec->data;
		if (data) {
			audit_fill_static(data);
			audit_fill_filedata(data, NULL);

			if (status < 0) {
				data->status = AU_SYSCALL_FAILED;
				data->arg1 = 0;
			} else {
				data->status = AU_SYSCALL_SUCCESSFULL;
				data->arg1 = status;
			}

			data->arg2 = 0;
			data->env = NULL;
			data->file_len = 0;
			data->file2_len = 0;
		
			rec->state = AU_FULL;
		} else
			rec->state = AU_UNUSED;
	}

	return 0;
}


asmlinkage int audit_fork(struct pt_regs regs)
{
	int error;

	error = ((int (*) (struct pt_regs)) orig_sys_call_table[__NR_fork])(regs);

	if (audit_doit(AUE_FORK, error))
		__audit_fork(error, AUE_FORK);

	return error;
}


asmlinkage int audit_clone(struct pt_regs regs)
{
	int error;

	error = ((int (*) (struct pt_regs)) orig_sys_call_table[__NR_clone])(regs);

	if (audit_doit(AUE_CLONE, error))
		__audit_fork(error, AUE_CLONE);

	return error;
}


int audit_vfork(int retval)
{
	struct audit_record *rec;

	rec = audit_get_free_record(AU_HOST_REC, 1, AUE_VFORK);
	if (rec) {
		struct audit_host_record *data;

		data = (struct audit_host_record *) rec->data;
		if (data) {
			audit_fill_static(data);
			audit_fill_filedata(data, NULL);

			rec->event = AUE_VFORK;

			if (retval < 0) {
				data->status = AU_SYSCALL_FAILED;
				data->arg1 = 0;
			} else {
				data->status = AU_SYSCALL_SUCCESSFULL;
				data->arg1 = retval;
			}

			data->arg2 = 0;
			data->env = NULL;
			data->file_len = 0;
			data->file2_len = 0;
		
			rec->state = AU_FULL;
		} else
			rec->state = AU_UNUSED;
	}

	return 0;
}

static int audit_do_execve(char *filename, char **argv, char **envp,
                           struct pt_regs *regs)
{
	struct audit_dyn_buf_entry *env;
	char magic_code[128];
	int error;
	void *src[3];

	memset(magic_code, 0, 128);

	src[0] = (void *) argv;
	src[1] = (void *) envp;
	src[2] = NULL;

	env = NULL;
	if (audit_fill_dyn_buf(au_env_buf, &env, src, 2) && env) {
		env->state = AU_FREE;
		env = NULL;
	}

	if (!IS_ERR(filename)) {
		struct file *file;

		file = open_exec(filename);
		if (!IS_ERR(file)) {
			kernel_read(file, 0, magic_code, 128);

			allow_write_access(file);
			fput(file);

			error = do_execve(filename, argv, envp, regs);
		} else
			error = PTR_ERR(file);
	} else
		error = PTR_ERR(filename);

	if (audit_doit(AUE_EXECVE, error)) {
		struct audit_record *rec;

#ifdef CONFIG_ACL
		if (!IS_ERR(filename) && acl_check_host(filename, AUE_EXECVE, error))
			goto error_out;	
#endif

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_EXECVE);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (!data) {
				rec->state = AU_UNUSED;
				goto error_out;
			}

			audit_fill_static(data);

			data->env = env;
			env = NULL;

			data->arg1 = 0;
			data->arg2 = 0;

			data->file_len = 0;
			if (!IS_ERR(filename)) {
				audit_fill_cfiledata(data, filename);
				audit_resolve_filename(data->file, &data->file_len, filename,
				                       STRLEN(filename));
			}

			data->file2_len = 0;
			if (magic_code[0] == '#' && magic_code[1] == '!') {
				int i;

				for (i = 0; i < 128 && magic_code[i] != '\0'; i++)
					data->file2[i] = magic_code[i];

				data->file2_len = i;
			}

			data->status = !error ? AU_SYSCALL_SUCCESSFULL : AU_SYSCALL_FAILED;

			rec->state = AU_FULL;
		}
	}

out:
	return error;

error_out:
	if (env)
		env->state = AU_FREE;

	goto out;
}


asmlinkage int audit_init_execve(struct pt_regs regs)
{
	int error;
	char *filename;

	filename = getname((char *) regs.ebx);
	if (AU_SVC_INITIALIZED(&au_svc)) {
		error = audit_do_execve(filename, (char **) regs.ecx,
		                        (char **) regs.edx, &regs);
	} else {
		error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx,
		                  &regs);
	}

	if (error == 0)
		current->ptrace &= ~PT_DTRACE;

	if (!IS_ERR(filename))
		putname(filename);

	return error;
}

