#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/string.h>
#include <linux/wait.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/utsname.h>
#include <linux/module.h>

#include <asm/atomic.h>
#include <asm/uaccess.h>

#include <linux/audit.h>

#include "buffer.h"
#include "proc.h"

static kmem_cache_t *au_buf_cache = NULL;

struct audit_buf au_buf = AUDIT_BUFFER_INIT;
struct audit_svc_info au_svc = AUDIT_SERVICE_INFO_INIT;

extern int audit_getmodus(void);

static struct audit_buffer_info au_dummy_buf_info = AU_BUFFER_INFO;

#ifdef CONFIG_AUDIT_PROC

unsigned long au_rec_demand = 0;
unsigned long au_rec_add = 0;

unsigned long au_rec_unused = 0;
unsigned long au_rec_saved = 0;

unsigned long au_max_add = 0;

unsigned long au_suspend_max = 0;
unsigned long au_suspend = 0;

#ifdef CONFIG_AUDIT_DEBUG

static inline void audit_service_proc(int ret, unsigned long limit, int size)
{
	static unsigned long last_sav = 0;
	unsigned long cur_saved = 0 ;

	cur_saved = (au_rec_unused + au_rec_saved) - last_sav;
	last_sav += cur_saved;

	au_svc_proc.saved = cur_saved;
	au_svc_proc.retval = ret;
	au_svc_proc.size = size;
	au_svc_proc.limit = limit;

	audit_proc_buf_ptr(&au_svc_proc.buf_end, &au_buf);
}	

#endif 

#endif

extern int audit_abort(void);

void audit_evt_filtered(unsigned int type)
{
	if (AU_SVC_INITIALIZED(&au_svc) && type < AU_REC_TYPES)
		au_svc.buffers[type]->filtered++;
}

static inline void buf_init(void *p, kmem_cache_t *cache, unsigned long flags)
{
	struct audit_record *new_p = p;

	memset(new_p, 0, sizeof(struct audit_record));
	new_p->state = AU_FREE;
}


int init_buf_cache(void)
{
	if (au_buf_cache)
		return 0;
	
	KLADT_PRINT("audit: initializing buffer cache ...\n");
	
	au_buf_cache = kmem_cache_create("au_buf_cache",
	                                 sizeof(struct audit_record), 0,
	                                 SLAB_HWCACHE_ALIGN, buf_init, NULL);
	
	if (!au_buf_cache) {
		printk(KERN_NOTICE "audit: buffer cache initialisation failed !!!\n");
		return -ENOMEM;
	}

	return 0;
}

int audit_buffer_rollback(void)
{
	if (atomic_read(&au_svc.state) != AU_DOIT)
		return 0;
	
	if (au_buf.first_full) {
		struct audit_record *p;
	
		p = au_buf.first_free;

		do {
			if (p->pid == current->pid && p->state == AU_USED)
				p->state = AU_UNUSED;

			p = p->prev;
		} while (p->next != au_buf.first_full);
	}

	return 0;
}

static int audit_clear_header_buffer(struct audit_record *anchor)
{
	struct audit_record *p;

	if (!anchor)
		return -EINVAL;

	if (anchor->prev)
		anchor->prev->next = NULL;

	while (anchor) {
		p = anchor->next;
		kmem_cache_free(au_buf_cache, anchor);
	
		anchor = p;
	}

	return 0;
}

struct audit_record *audit_init_header_buffer(unsigned int size, int flag)
{
	struct audit_record *cur_p = NULL, *retval = NULL;
	long i;

	retval = kmem_cache_alloc(au_buf_cache, flag);
	if (!retval)
		goto out;

	memset(retval, 0, sizeof(struct audit_record));
	retval->state = AU_FREE;
	retval->index = 0;

	for (cur_p = retval, i = 1; i < size; i++) {
		cur_p->next = kmem_cache_alloc(au_buf_cache, flag);
		if (!cur_p->next)
			goto free_out;

		memset(cur_p->next, 0, sizeof(struct audit_record));
		cur_p->next->state = AU_FREE;

		cur_p->next->index = i;

		cur_p->next->prev = cur_p;
		cur_p = cur_p->next;
	}

	retval->prev = cur_p;
	cur_p->next = retval;

out:
	return retval;

free_out:
	audit_clear_header_buffer(retval);
	retval = NULL;

	goto out;
}

static int audit_init_buffer(unsigned int cnt, int **b_init)
{
	int retval, cur_type;
	long i;

	if (au_buf.addr) {
		printk(KERN_NOTICE "Buffer already initializied \n");
		return -EBUSY;
	}
	
	if (!b_init || !*b_init)
		return -EFAULT;

	retval = -EINVAL;
	if (cnt > 0) {
		unsigned int buf_cnt, recs[AU_REC_TYPES];
		int i, tmp[2];
		
		audit_buffer_init(au_svc.buffers);
		for (i = 0; i < AU_REC_TYPES; i++) {
			if (!au_svc.buffers[i])
				au_svc.buffers[i] = &au_dummy_buf_info;
		}

		buf_cnt = 0;

		for (i = 0; i < AU_REC_TYPES; i++)
			recs[i] = 0;

		retval = 0;
		for (i = 0; i < cnt && !retval; i++) {
			memcpy(&tmp, b_init[i], 2 * sizeof(int));

			if (tmp[0] < 0 || AU_REC_TYPES <= tmp[0]) {
				retval = -EINVAL;
				break;
			}

			if (au_svc.buffers[tmp[0]]->type != AU_UNDEF) {
				recs[tmp[0]] += tmp[1];
				buf_cnt += tmp[1];
			} else
				printk(KERN_WARNING "audit_init_buffers: "
				                    "type %d is not supported by this kernel",
				                    tmp[0]);
		}

		if (buf_cnt && !retval) {
			if (!au_buf_cache)
				return -ENOMEM;

			for (cur_type = 0; cur_type < AU_REC_TYPES; cur_type++) {
				if (!recs[cur_type]) {
					au_svc.buffers[cur_type]->state = AU_UNDEF;
					continue;
				}

				retval = au_svc.buffers[cur_type]->init(au_svc.buffers[cur_type],
				                                      recs[cur_type]);
				if (retval)
					goto free_out;
			}

			au_buf.addr = audit_init_header_buffer(buf_cnt, GFP_KERNEL);
			if (!au_buf.addr)
				goto free_out;

			au_buf.first_free = au_buf.addr;
			au_buf.to_save = au_buf.addr;
		}
	}

out:
	return retval;

free_out:
	for (i = 0; i < cur_type; i++) {
		if (au_svc.buffers[i]->destroy)
			au_svc.buffers[i]->destroy(au_svc.buffers[i]);
	}

	au_buf.addr = NULL;
	au_buf.first_full = NULL;
	au_buf.first_free = NULL;
	au_buf.to_save = NULL;

	goto out;
}



static int audit_clear_buffer(void)
{
	int i;

	if (!au_buf.addr)
		return -ESRCH;

	if (au_buf.additional) {
		struct audit_record *p, *q;

		p = au_buf.additional;
		while (p) {
			q = p->next;

			if (au_svc.buffers[p->type]->free_rec) {
				au_svc.buffers[p->type]->free_rec(au_svc.buffers[p->type],
				                                  p->data);
			}

			kmem_cache_free(au_buf_cache, p);

#ifdef CONFIG_AUDIT_PROC
			if (au_rec_add > au_max_add)
				au_max_add = au_rec_add;

			au_rec_add--;
#endif

			p = q;
		}

		au_buf.additional = NULL;
		au_buf.last_add = NULL;
	}
		
	audit_clear_header_buffer(au_buf.addr);
	au_buf.first_full = NULL;
	au_buf.first_free = NULL;
	au_buf.to_save = NULL;
	au_buf.addr = NULL;

	for (i = 0; i < AU_REC_TYPES; i++) {
		if (au_svc.buffers[i]->destroy)
			au_svc.buffers[i]->destroy(au_svc.buffers[i]);
	}

	atomic_set(&au_buf.state, AU_FREE);

	for (i = 0; i < AU_REC_TYPES; i++) {
		if (au_svc.buffers[i]->buf_wake_up)
			au_svc.buffers[i]->buf_wake_up();
	}

#ifdef CONFIG_AUDIT_PROC
	au_suspend = 0;
#endif

	return 0;
}


static void convert2additional(struct audit_record *rec, int type)
{
	rec->prev = NULL;

	if (rec->state == AU_FREE) {
		rec->state = AU_UNUSED;
		rec->data = NULL;

		return;
	}

	if (rec->type == type && rec->data)
		switch (type) {
			case AU_HOST_REC:
				((struct audit_host_record *) rec->data)->next = NULL;
				break;
#ifdef CONFIG_NET_AUDIT
			case AU_NET_REC:
				((struct audit_net_record *) rec->data)->next = NULL;
				break;
#endif
			default:
				BUG();
				break;
		}

	return;
}


static void make_header2additional(struct audit_buf *buf, int type)
{
	struct audit_record *p, *first_full, *last_full;

	first_full = NULL;
	for (p = buf->first_free; p->next != buf->first_free; p = p->next) {
		convert2additional(p, type);

		if (!first_full) {
			p->prev = p;

			first_full = p;
		}
	}

	convert2additional(p, type);
	last_full = p;
	last_full->next = NULL;


	if (buf->additional) {
		struct audit_record *last_add;

		if (buf->first_free != buf->first_full) {
			buf->last_add->next = first_full;
			goto out;
		}

		last_add = NULL;

		for (p = buf->additional; p; p = p->next) {
			if (!p->next) {
				last_add->next->prev = NULL;

				if (last_add) {
					last_full->next = last_add->next;
					last_add->next = first_full;

					goto out;
				}

				last_full->next = buf->additional;
				buf->additional = first_full;

				goto out;
			}

			if (p->next->prev == p->next) {
				last_add = p;
			}
		}
	}

	buf->additional = first_full;

out:
	return;
}


static int audit_reset_buffer(int type, int count)
{
	int retval, new_size, i;
	unsigned long flags;
	struct audit_record *new_buf;
	struct audit_buffer_info new_data;

	if (!au_buf.addr)
		return -ESRCH;

	if (au_svc.buffers[type]->type == AU_UNDEF)
		return -EFAULT;

	if (count < 0) 
		return -EINVAL;

	if (count < AU_BUFMIN) {
		printk(KERN_NOTICE "audit_reset_buffer: "
		                   "given value (%d) is to small, setting to %d.\n",
		       count, AU_BUFMIN);

		count = AU_BUFMIN;
	}

	retval = au_svc.buffers[type]->init(&new_data, count);
	if (retval)
		return retval;

	new_size = 0;
	for (i = 0; i < AU_REC_TYPES; i++) {
		if (i == type && count > AU_BUFMIN) {
			new_size += count;
			continue;
		}

		new_size += au_svc.buffers[i]->buf_size;
	}

	new_buf = audit_init_header_buffer(new_size, GFP_ATOMIC);
	if (!new_buf) {
		au_svc.buffers[type]->destroy(&new_data);
		return -ENOMEM;
	}

	spin_lock_irqsave(&au_buf.lock, flags);

	if (atomic_read(&au_svc.state) != AU_DOIT) {

		au_svc.buffers[type]->destroy(&new_data);

		audit_clear_header_buffer(new_buf);
		
		retval = -ESRCH;
		goto out;
	}

	au_svc.buffers[type]->reduse(au_svc.buffers[type]);

	if (au_buf.first_full)
		make_header2additional(&au_buf, type);


	au_buf.first_free = new_buf;

	if (!au_buf.first_full) {
		au_buf.to_save = au_buf.first_free;

		audit_clear_header_buffer(au_buf.addr);
	}

	au_buf.addr = new_buf;
	au_buf.session_closed = 1;
	au_buf.first_full = NULL;

	if (au_buf.additional) {
		if (au_buf.to_save->prev && au_buf.to_save->prev != au_buf.to_save) {
			au_buf.to_save = au_buf.additional;
		}
	}
	
	au_svc.buffers[type]->state = AU_FREE;
	au_svc.buffers[type]->rec_size = new_data.rec_size;
	au_svc.buffers[type]->buf_size = new_data.buf_size;
	au_svc.buffers[type]->buf_wm = new_data.buf_wm;
	au_svc.buffers[type]->free_recs = new_data.buf_size;
	au_svc.buffers[type]->buf_addr = new_data.buf_addr;
	au_svc.buffers[type]->cur_rec = new_data.cur_rec;

	atomic_set(&au_buf.state, AU_FREE);
	wake_up_auditd();

out:
	spin_unlock_irqrestore(&au_buf.lock, flags);

	return retval;
}

int audit_buffer_ctl(int cmd, unsigned long *arg)
{
	int retval = 0;

	switch (cmd) {
		case AU_BUF_INIT: {
			int i, **size;
			unsigned int buf_no;
			unsigned long src, args[2];

			retval = -EBUSY;
			if (atomic_read(&au_svc.state) == AU_DOIT)
				goto out;
	
			retval = -EINVAL;
			if (!arg)
				goto out;

			retval = -EFAULT;
			if (copy_from_user((void *) args, (void *) arg, 2 * sizeof(long)))
				goto out;

			buf_no = (unsigned int) args[0];

			retval = -ENOMEM;
			size = (int **) kmalloc((buf_no + 1) * sizeof(int *), GFP_KERNEL);
			if (!size)
				goto out;

			retval = 0;
			for (i = 0; ; i++) {
				size[i] = NULL;

				if (i == buf_no)
					break;

				if (get_user(src, ((int *) args[1]) + i)) {
					retval = -EFAULT;
					break;
				}

				size[i] = (int *) kmalloc(2 * sizeof(int), GFP_KERNEL);
				if (size[i]) {
					if (copy_from_user((void *) size[i], (void *) src,
					                   2 * sizeof(int))) {
						retval = -EFAULT;
						break;
					}
				} else
					retval = -ENOMEM;
			}

			if (!retval)
				retval = audit_init_buffer(buf_no, size);

			for ( ; i >= 0; i--) {
				if (size[i])
					kfree(size[i]);
			}

			kfree(size);

			break;
		}
		case AU_BUF_DESTROY: {
			retval = -EBUSY;
			if (atomic_read(&au_svc.state) == AU_DOIT)
				goto out;

			retval = audit_clear_buffer();

			break;
		}
		case AU_BUF_CONFIG: {
			unsigned long args[3];

			retval = -ESRCH;
			if (!au_buf.addr)
				goto out;

			retval = -EINVAL;
			if (!arg)
				goto out;

			retval = -EFAULT;
			if (copy_from_user((void *) args, (void *) arg, 3 * sizeof(long))) {
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "unable to copy args from address %lu.\n",
				       (unsigned long) arg);
#endif
				goto out;
			}
			
			retval = -EINVAL;
			if ((AU_REC_TYPES <= args[0])
			    || (au_svc.buffers[args[0]]->type == AU_UNDEF))
				goto out;

			if (args[1] == AU_BUF_SET_SIZE) {
				retval = audit_reset_buffer((int ) args[0], (int) args[2]);
				goto out;
			}

			retval = au_svc.buffers[args[0]]->ctl(au_svc.buffers[args[0]],
			                                      args[1], args[2]);
			break;
		}
		default:
			retval = -EINVAL;
	}
out:

	return retval;
}

static inline struct audit_record *audit_new_header(struct audit_buf *buf)
{
	struct audit_record *new_p;

	new_p = kmem_cache_alloc(au_buf_cache, GFP_ATOMIC);
	if (new_p) {
		memset(new_p, 0, sizeof(struct audit_record));

		new_p->state = AU_USED;
		if (!buf->additional) {
			new_p->prev = new_p;
			buf->additional = new_p;
		} else {
			if (buf->session_closed) {
				buf->session_closed = 0;
				new_p->prev = new_p;
			}

			buf->last_add->next = new_p;
		}

		buf->last_add = new_p;

#ifdef CONFIG_AUDIT_PROC
		au_rec_add++;
#endif
	}

	return new_p;
}

static struct audit_record *take_first_free(struct audit_buf *buf)
{
	struct audit_record *p;

	p = buf->first_free;

	if (!p) {
#ifdef CONFIG_AUDIT_DEBUG
		panic("AUDIT: take_first_free: no first free record in this buffer.\n");
#endif
		return NULL;
	}

	if ((p->state == AU_FREE) && (p != buf->first_full)) {
		p->state = AU_USED;

		if (!buf->first_full)
			buf->first_full = p;

		buf->first_free = p->next;

		return p;
	}
	
	return NULL;
}

struct audit_record *audit_get_free_record(unsigned int type,
                                           int interruptable, unsigned int evt)
{
	struct audit_record *p;
	unsigned long flags;
	int len;

	p = NULL;
	if (atomic_read(&au_svc.state) != AU_DOIT)
		goto out;
	
#ifdef CONFIG_AUDIT_PROC
	au_rec_demand++;
#endif

	if (AU_REC_TYPES <= type || au_svc.buffers[type]->state == AU_UNDEF)
		goto drop_out;

	spin_lock_irqsave(&au_buf.lock, flags);

	if (atomic_read(&au_svc.state) != AU_DOIT)
		goto unlock_out;

	p = take_first_free(&au_buf);
	if (!p) {
		wake_up_auditd();

		if (audit_getmodus() != AU_SECURE)
			goto unlock_out;

		if (interruptable && current->pid != au_svc.auditd_pid) {
			do {

#ifdef CONFIG_AUDIT_PROC
				au_suspend++;
				if (au_suspend > au_suspend_max)
					au_suspend_max = au_suspend;
#endif
	
				spin_unlock_irqrestore(&au_buf.lock, flags);
				if (au_svc.buffers[type]->buf_suspend())
					goto drop_out;

				spin_lock_irqsave(&au_buf.lock, flags);
				if (atomic_read(&au_svc.state) != AU_DOIT)
					goto unlock_out;
	
				p = take_first_free(&au_buf);
			} while (!p);
		} else {
			p = audit_new_header(&au_buf);
			if (!p)
				goto unlock_out;
		}
	}
		
	p->pid = current->pid;
	p->type = type;
	p->event = evt;

	p->data = au_svc.buffers[type]->get_free(au_svc.buffers[type]);
	do {
		if (!p->data) {

#ifdef CONFIG_AUDIT_DEBUG
			panic("audit_get_free_record: get_free return a NULL-pointer.\n");
#endif

			p->state = AU_UNUSED;
			goto drop_out;
		}

		spin_unlock_irqrestore(&au_buf.lock, flags);

		if (IS_ERR(p->data)) {
			if (PTR_ERR(p->data) != -ENOSPC) {
				KLADT_PRINT("audit_get_free_record: error %d in get_free.\n",
				            (int) PTR_ERR(p->data));
	
				p->data = NULL;
				p->state = AU_UNUSED;
				p = NULL;
	
				goto out;
			}

			if (!spin_is_locked(&au_svc.lock)) {
				atomic_set(&au_buf.state, AU_FULL);
				wake_up_auditd();
			}
	
			p->data = au_svc.buffers[type]->get_new();
			if (!p->data) {
				if (!interruptable || current->pid == au_svc.auditd_pid) {
					p->state = AU_UNUSED;
					goto drop_out;
				}
	
#ifdef CONFIG_AUDIT_PROC
				au_suspend++;
				if (au_suspend > au_suspend_max)
					au_suspend_max = au_suspend;
#endif

				if (au_svc.buffers[type]->buf_suspend()) {
					p->state = AU_UNUSED;
					goto drop_out;
				}
	
				spin_lock_irqsave(&au_buf.lock, flags);
				if (atomic_read(&au_svc.state) != AU_DOIT)
					goto unlock_out;
			}
		}
	} while (!p->data);
			
	p->date = CURRENT_TIME;

	len = strlen(system_utsname.nodename);

	if (strncpy(p->hostname, system_utsname.nodename, len) < 0) {
		KLADT_PRINT("audit_get_free_record: copy hostname failed %s.\n",
		            system_utsname.nodename);
		len = 0;
	}

	p->hostname[len] = '\0';
	p->hostname_len = len;

out:
	return p;

unlock_out:
	spin_unlock_irqrestore(&au_buf.lock, flags);
drop_out:
	au_svc.buffers[type]->dropped++;
	if (current->pid == au_svc.auditd_pid) {
		KLADT_PRINT("audit_get_free_record: "
		            "dropped record request from the audit daemon.\n");
	}
	p = NULL;

	goto out;
}


EXPORT_SYMBOL(audit_get_free_record);

static inline int rec2char(char *buf, struct audit_record *p)
{
	int size, err;
			
	memcpy(buf, (void *) p, AU_BUF_FIX);

	size = AU_BUF_FIX;
	buf += size;

	memcpy(buf, p->hostname, p->hostname_len);
	buf += p->hostname_len;
	size += p->hostname_len;

	*buf = '\0';
	buf++;
	size++;

	err = au_svc.buffers[p->type]->rec2chr(au_svc.buffers[p->type], buf,
	                                       p->data);
	if (err > 0) {
		size += err;
		buf += err;

		*buf = '\0';
		size++;

		err = 0;
	}

	if (err) {
		size = err;
	}

	return size;
}


static inline struct audit_record *nxt2save(struct audit_buf *buf,
                                            struct audit_record *cur2save)
{
	AUDIT_DEBUG_PROC_VAR_INC(nxt2save);

	if (!cur2save->prev || cur2save->prev == cur2save) {
		buf->additional = cur2save->next;

		if (!cur2save->next || cur2save->next->prev == cur2save->next) {
#ifdef CONFIG_AUDIT_DEBUG
			static unsigned int full = 0, free = 0;
#endif

			if (!buf->first_full) {
				KLADT_PRINT("nxt2save: "
						"end of additional return first_free(%d, %d) - %d.\n",
						buf->first_free->index, buf->first_free->state,
						++free);

				return buf->first_free;
			}
	
			KLADT_PRINT("nxt2save: "
						"end of additional return first_full(%d, %d) - %d.\n",
						buf->first_full->index, buf->first_full->state,
						++full);

			return buf->first_full;
		}

		return cur2save->next;
	}

	if (cur2save->next == buf->first_full) {
		buf->first_full = NULL;
		buf->session_closed = 1;
		atomic_set(&buf->state, AU_FREE);
	
		if (buf->additional) {
			return buf->additional;
		}
	}

	return cur2save->next;
}

static int audit_record2char(char *buf)
{
	struct audit_record *p;
	unsigned long flags;
	int size;

	if (!au_buf.to_save) {
		KLADT_PRINT("audit_record2char: memory leak in the audit-buffer.\n");
		return -EFAULT;
	}

	AUDIT_DEBUG_PROC_VAR_INC(rec2chr);
		
	spin_lock_irqsave(&au_buf.lock, flags);

	for (size = 0; !size; ) {
		p = au_buf.to_save;

#ifdef CONFIG_AUDIT_DEBUG
		audit_proc_save_last_rec(&au_svc_proc.last_rec, p);
#endif

		switch (p->state) {
		  case AU_FULL:
			AUDIT_DEBUG_PROC_VAR_INC(rec2chr_full);

			size = rec2char(buf, p);
#ifdef CONFIG_AUDIT_PROC
			if (size == -ENODATA)
				au_rec_unused++;
			else
				au_rec_saved++;
#endif
			break;
		  case AU_UNUSED:
			AUDIT_DEBUG_PROC_VAR_INC(rec2chr_unused);

#ifdef CONFIG_AUDIT_PROC
			au_rec_unused++;
#endif
			break;
		  case AU_USED:
			AUDIT_DEBUG_PROC_VAR_INC(rec2chr_used);

			goto unlock_out;
		  case AU_FREE:
			AUDIT_DEBUG_PROC_VAR_INC(rec2chr_free);

			if (au_buf.first_full && !au_buf.additional) {
				au_buf.first_full = NULL;
				au_buf.session_closed = 1;
				atomic_set(&au_buf.state, AU_FREE);
			}

			goto unlock_out;
		  default:
			AUDIT_DEBUG_PROC_VAR_INC(rec2chr_err);

			KLADT_PRINT("audit_record2char: "
					"found record with unknown state %d.\n",
					p->state);
		}

	au_svc.buffers[p->type]->free_rec(au_svc.buffers[p->type], p->data);
	
	au_buf.to_save = nxt2save(&au_buf, p);

	p->state = AU_FREE;
	p->pid = 0;
	p->data = NULL;

	if (!p->prev || p == p->prev) {
		kmem_cache_free(au_buf_cache, p);
#ifdef CONFIG_AUDIT_PROC
		if (au_rec_add > au_max_add)
			au_max_add = au_rec_add;

			au_rec_add--;
#endif
		}
	}

unlock_out:
	spin_unlock_irqrestore(&au_buf.lock, flags);

	AUDIT_DEBUG_PROC_ERR(rec2chr, size);

	return size;
}


int wake_up_auditd(void)
{
	int ret;

	ret = -EBUSY;
	if (!spin_is_locked(&au_svc.lock) && !au_svc.wakeup_cnt) {
		ret = send_sig(SIGALRM, au_svc.auditd_task, 1);
		if (!ret)
		au_svc.wakeup_cnt++;
	}

	return ret;
}


static inline int au_svc_buf2file(struct file *file,
						  struct audit_svc_info *svc, int size)
{
	ssize_t (*write)(struct file *, const char *, size_t, loff_t *);
	int retval;

	write = file->f_op->write;
	
	AUDIT_DEBUG_PROC_VAR_INC(buf2file);

	retval = locks_verify_area(FLOCK_VERIFY_WRITE,
					   file->f_dentry->d_inode, file,
					   file->f_pos, size); 
	if (retval) {
		KLADT_PRINT(
	   "sys_audit_service: locking area failed %d \n",
	   retval);

		goto out;
	}

	retval = -EFAULT;
	if (!copy_to_user(svc->user_buf, svc->svc_buf, size)) {	
		int written, towrite, again;
		char *buf;

		again = 5;
	
		written = 0;
		towrite = size;
		buf = svc->user_buf;
		do {
			if (!again) {
				KLADT_PRINT("au_svc_buf2file: unable to write buffer in 5 "
					"trails.\n");
				retval = -EIO;
				break;
			}

			again--;

			retval = write(file, buf, towrite, &file->f_pos);
			if (retval < 0) {
				KLADT_PRINT("au_svc_buf2file: writing record failed %d.\n",
							retval);
				break;
			}

			written += retval;
			towrite -= retval;
			buf = (char *) ((unsigned long) buf + (unsigned long) retval);
		} while ((retval + written) < size);
	} else
		KLADT_PRINT("au_svc_buf2file: copy_to_user err.\n");

out:
	AUDIT_DEBUG_PROC_ERR(buf2file, retval);

	return retval;
}


static inline int get_data_from_buf(struct audit_svc_info *svc)
{
	int size;

	AUDIT_DEBUG_PROC_VAR_INC(getdata);

	for (size = 0;;) {
		size = audit_record2char(au_svc.svc_buf);
		if (size > 0)
			break;

		if (!size) {
			break;
		}

		switch (size) {
		  case -ENODATA:
			AUDIT_DEBUG_PROC_VAR_INC(enodata);

			continue;
		  default:
			KLADT_PRINT("sys_audit_service: "
						"audit_record2_char returned %d.\n", size);
		}
	}

	AUDIT_DEBUG_PROC_ERR(getdata, size);

	return size;
}

int sys_audit_service(int fd, unsigned long limit)
{
	int i, retval, ret;
	struct file *file;
	static int size = 0;

	ret = -EFAULT;
	if (!au_svc.svc_buf || !au_svc.user_buf)
		goto out;
	
	ret = -EPERM;
	if (!audit_check_capability(AU_CAP_ADM)) {
		KLADT_PRINT("sys_audit_service: permissions denied\n");
		goto out;
	}

	ret = -EBUSY;
	if (!spin_trylock(&au_svc.lock))
		goto out;

#ifdef CONFIG_AUDIT_DEBUG
	spin_lock(&au_svc_proc.lock);

	memset((char *) &au_svc_proc, 0,
	       sizeof(struct au_svc_proc_s) - sizeof(spinlock_t));

	audit_proc_buf_ptr(&au_svc_proc.buf_start, &au_buf);
#endif

	ret = 0;
	if (limit > AU_MAX_TRAILSIZE)
		limit = AU_MAX_TRAILSIZE;

	file = fget(fd);
	if (!file) {
		KLADT_PRINT("sys_audit_service: fget() failed \n");
		retval = -EBADF;
		goto unlock_out;
	}
	
	retval = -EINVAL;
	if (file->f_op && file->f_op->write && (file->f_mode & FMODE_WRITE)) {
		ret = 0;

		do {
			retval = 0;

			if (!size || (*(((unsigned long *) au_svc.svc_buf) + 1) == 0)) {
				size = get_data_from_buf(&au_svc);
				if (!size) {
					break;
				}

				if (size < 0) {
					retval = size;
					break;
				}
			}

			retval = -EFBIG;
			if (size < limit) {
				retval = au_svc_buf2file(file, &au_svc, size);
				if (retval > 0) {

					limit -= retval;
					ret += retval;
	
					size = 0;
				} else
					KLADT_PRINT("sys_audit_service: "
					            "au_svc_buf2file() failed with %d\n",
					            retval);
			} else
				KLADT_PRINT("sys_audit_service: "
				            "no more left space (%d) on device (%lu) \n",
				            size, limit);
		} while (retval > 0);
	}

	fput(file);

unlock_out:
	spin_unlock(&au_svc.lock);

	for (i = 0; i < AU_REC_TYPES; i++) {
		if (au_svc.buffers[i]->buf_wake_up)
			au_svc.buffers[i]->buf_wake_up();
	}

	if (retval < 0)
		ret = retval;

#ifdef CONFIG_AUDIT_PROC
	au_suspend = 0;

#ifdef CONFIG_AUDIT_DEBUG
	audit_service_proc(ret, limit, size);
	spin_unlock(&au_svc_proc.lock);
#endif

#endif

out:
	au_svc.wakeup_cnt = 0;

	return ret;
}
