#include <linux/sched.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/module.h>

#include <asm/uaccess.h>

#include <linux/acl.h>
#include <linux/audit.h>
#include <linux/audit_event.h>

#include "buffer.h"
#include "timer.h"

extern int audit_getmodus(void);

pid_t audit_get_pid(void)
{
	if (AU_SVC_INITIALIZED(&au_svc))
		return au_svc.auditd_pid;

	return 0;
}


struct task_struct *audit_get_task(void)
{
	if (AU_SVC_INITIALIZED(&au_svc))
		return au_svc.auditd_task;

	return NULL;
}

#define AUDIT_NOP() do { } while(0)

int audit_abort(void)
{
	if (AU_SVC_INITIALIZED(&au_svc)) {
#ifdef CONFIG_AUDIT_TIMER
		disable_audit_timer();
#endif
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "audit: aborted \n");
#endif
		AUDIT_NOP();
	}

	atomic_set(&au_svc.state, AU_STOPPED);

	au_svc.auditd_pid = AU_UNDEF;
	au_svc.auditd_task = NULL;

	au_svc.user_buf = NULL;
	if (au_svc.svc_buf) {
		vfree(au_svc.svc_buf);
		au_svc.svc_buf = NULL;
	}

	return 0;
}

static int audit_start(char *addr)
{
	int ret;
	unsigned int type;

	ret = -EBUSY;
	if (AU_SVC_INITIALIZED(&au_svc))
		goto out;

	ret = -EFAULT;
	if (!addr || verify_area(VERIFY_WRITE, addr, AU_WRITE_BUFFER_SIZE))
		goto out;
	
	ret = 0;
	for (type = 0; type < AU_REC_TYPES; type++) {
		if (!au_svc.buffers[type])
			continue;

		ret += au_svc.buffers[type]->buf_size;
	}
	
	if (!ret) {
		return -ENOSPC;
	}

	ret = -ENOMEM;
	au_svc.svc_buf = vmalloc(AU_WRITE_BUFFER_SIZE);
	if (au_svc.svc_buf) {
		memset(au_svc.svc_buf, 0, AU_WRITE_BUFFER_SIZE);

		au_svc.user_buf = addr;

		au_svc.auditd_pid = current->pid;
		au_svc.auditd_task = find_task_by_pid(au_svc.auditd_pid);

		if (au_svc.auditd_task) {
			int old_state = atomic_read(&au_svc.state);

			atomic_set(&au_svc.state, AU_DOIT);
			ret = 0;

			if (old_state == AU_UNDEF)
				audit_sysboot();

#ifdef CONFIG_AUDIT_TIMER
			enable_audit_timer();
#endif
		}
	}

out:
	return ret;
}


int audit_init_mask(int sel, int id, unsigned long *mask)
{
	int ret;
	struct task_struct *p = NULL;
	__u32 msk[AU_MSK_SIZE32];

	if (!mask || copy_from_user((char *) msk, (char *) mask, AU_MSK_SIZE8))
		return -EFAULT;

	ret = -EINVAL;
	switch (sel) {
		case AU_SETMASK_ALL : {
			ret = 0;

			lock_kernel();

			for_each_task(p) {
				if (!p->audit)
					continue;

				ret = audit_set_evt_msk(&p->audit->msk, msk, GFP_ATOMIC);
				if (ret)
					break;
			}

			unlock_kernel();

			break;
		}
		case AU_SETMASK_PID : {
			ret = 0;

			lock_kernel();

			for_each_task(p) {
				if (!p->audit || p->pid != id)
					continue;

				ret = audit_set_evt_msk(&p->audit->msk, msk, GFP_ATOMIC);
				if (ret)
					break;
			}

			unlock_kernel();

			break;
		}
		case AU_SETMASK_UID : {
			ret = 0;

			lock_kernel();

			for_each_task(p) {
				if (!p->audit || p->audit->aid != id)
					continue;

				ret = audit_set_evt_msk(&p->audit->msk, msk, GFP_ATOMIC);
				if (ret)
					break;
			}

			unlock_kernel();

			break;
		}
		case AU_SETMASK_GID : {
			ret = 0;

			lock_kernel();

			for_each_task(p) {
				if (!p->audit || p->audit->agid != id)
					continue;

				ret = audit_set_evt_msk(&p->audit->msk, msk, GFP_ATOMIC);
				if (ret)
					break;
			}

			unlock_kernel();

			break;
		}
	}

	return ret;
}


int audit_setmask(int sel, int id, int event, int value)
{
	struct task_struct *p = NULL;
	int ret;
	
	if (event < 0 || event > AU_MAX_EID)
		return -EINVAL;

	switch (sel) {
		case AU_SETMASK_ALL:
			ret = 0;

			lock_kernel();

			for_each_task(p) {
				if (!p->audit)
					continue;

				ret = audit_update_evt_msk(&p->audit->msk, event, value, 2,
				                           GFP_ATOMIC);
			}

			unlock_kernel();

			break;
		case AU_SETMASK_PID:
			ret = 0;

			lock_kernel();

			for_each_task(p) {
				if (!p->audit || p->pid != id)
					continue;

				ret = audit_update_evt_msk(&p->audit->msk, event, value, 2,
				                           GFP_ATOMIC);
			}

			unlock_kernel();

			break;
		case AU_SETMASK_UID:
			ret = 0;

			lock_kernel();

			for_each_task(p) {
				if (!p->audit || p->audit->aid != id)
					continue;

				ret = audit_update_evt_msk(&p->audit->msk, event, value, 2,
				                           GFP_ATOMIC);
			}

			unlock_kernel();

			break;
		case AU_SETMASK_GID:
			ret = 0;

			lock_kernel();

			for_each_task(p) {
				if (!p->audit || p->audit->agid != id)
					continue;

				ret = audit_update_evt_msk(&p->audit->msk, event, value, 2,
				                           GFP_ATOMIC);
			}

			unlock_kernel();

			break;
		default:
			ret = -EINVAL;
	}

	return ret;
}

int sys_audit_setaid(int aid)
{
	int old_aid = 0, ret;

	ret = -EPERM;
	if (!current->audit->set || audit_check_capability(AU_CAP_SETID)) {
		old_aid = current->audit->aid;

		current->audit->aid = aid;
		current->audit->set = 1;

		ret = 0;
	}
#ifdef CONFIG_AUDIT_DEBUG
	else
		printk(KERN_DEBUG "sys_audit_setaid: access denied for user %d.\n",
		       current->audit->aid);
#endif
	if (audit_doit(AUE_SETAUID, ret))
		audit_setaid(ret, aid, old_aid);

	return ret;
}

int audit_login(int aid, gid_t agid, pid_t sid, unsigned long from)
{
	int ret;

	ret = sys_audit_setaid(aid);
	if (!ret) {
		current->audit->agid = agid;
		current->audit->sid = sid;
		current->audit->from = from;
	}

	return ret;
}

int audit_getmask(int sel, int id, void *buf)
{
	int ret;

	lock_kernel();

	if (!buf || verify_area(VERIFY_WRITE, buf, AU_MSK_SIZE8)) {
		unlock_kernel();
		return -EFAULT;
	}

	ret = -EINVAL;	
	switch (sel) {
		struct task_struct *p;

		case AU_GETMASK_PID:
			for_each_task(p) {
				if (p->audit && p->pid == id) {
					__u32 msk[AU_MSK_SIZE32];

					ret = audit_get_acl_msk(p->audit->msk, (__u32 *) msk);
					if (!ret)
						memcpy(buf, (char *) msk, AU_MSK_SIZE8);

					break;
				}
			}

			break;
		case AU_GETMASK_UID:
			for_each_task(p) {
				if (p->audit && p->audit->aid == id) {
					__u32 msk[AU_MSK_SIZE32];

					ret = audit_get_acl_msk(p->audit->msk, (__u32 *) msk);
					if (!ret)
						memcpy(buf, (char *) msk, AU_MSK_SIZE8);

					break;
				}
			}

			break;
		case AU_GETMASK_GID:
			for_each_task(p) {
				if (p->audit && p->audit->agid == id) {
					__u32 msk[AU_MSK_SIZE32];

					ret = audit_get_acl_msk(p->audit->msk, (__u32 *) msk);
					if (!ret)
						memcpy(buf, (char *) msk, AU_MSK_SIZE8);

					break;
				}
			}

			break;
	}

	unlock_kernel();

	return ret;
}

int sys_audit_ctl(int cmd, unsigned long args)
{
	int err = 0;

	switch (cmd) {
		case AU_ON : {
			if (audit_getmodus() == AU_DISABLED)
				return -ESRCH;

			if (!audit_check_capability(AU_CAP_ADM))
				return -EPERM;

			err = audit_start((char *) args);

			break;
		}
		case AU_OFF : {
			if (audit_getmodus() == AU_DISABLED)
				return -ESRCH;

			if (!audit_check_capability(AU_CAP_ADM))
				return -EPERM;

			audit_abort();

			break;
		}
		case AU_SETAID : {
			if (audit_getmodus() == AU_DISABLED)
				return -ESRCH;

			err = sys_audit_setaid((int) args);

			break;
		}
		case AU_LOGIN : {
			unsigned long a[4];
			
			if (audit_getmodus() == AU_DISABLED)
				return -ESRCH;

			if (copy_from_user(a, (char *) args, 4 * sizeof(unsigned long)))
				return -EFAULT;

			err = audit_login((int) a[0], (gid_t) a[1], (pid_t) a[2], a[3]);

			break;
		}
		case AU_INITMASK : {
			unsigned long a[3];

			if (audit_getmodus() == AU_DISABLED)
				return -ESRCH;

			if (!audit_check_capability(AU_CAP_ADM) &&
			    !audit_check_capability(AU_CAP_SETID))
				return -EPERM;

			if (copy_from_user(&a, (char *) args, 3 * sizeof(unsigned long)))
				return -EFAULT;

			err = audit_init_mask((int) a[0], (int) a[1], (unsigned long *) a[2]);

			break;
		}
		case AU_SETMASK : {
			unsigned long a[4];

			if (audit_getmodus() == AU_DISABLED)
				return -ESRCH;

			if (!audit_check_capability(AU_CAP_ADM))
				return -EPERM;

			if (copy_from_user(&a, (char *) args, 4 * sizeof(unsigned long)))
				return -EFAULT;

			err = audit_setmask((int) a[0], (int) a[1], (int) a[2], (int) a[3]);

			break;
		}
		case AU_BUFFER_CTL : {
			unsigned long a[2];

			if (audit_getmodus() == AU_DISABLED)
				return -ESRCH;

			if (!audit_check_capability(AU_CAP_ADM))
				return -EPERM;

			if (copy_from_user(&a, (char *) args, 2 * sizeof(unsigned long)))
				return -EFAULT;

			err = audit_buffer_ctl((int) a[0], (unsigned long *) a[1]);

			break;
		}
		case AU_CAP_CTL : {
			unsigned long a[2];

			if (copy_from_user(&a, (char *) args, 2 * sizeof(unsigned long)))
				return -EFAULT;

			err = sys_audit_cap_ctl((int) a[0], (unsigned long *) a[1]);

			break;
		}
		case AU_GETMASK : {
			unsigned long a[3];

			if (audit_getmodus() == AU_DISABLED)
				return -ESRCH;

			if (!audit_check_capability(AU_CAP_ADM))
				return -EPERM;

			if (copy_from_user(&a, (char *) args, 3 * sizeof(unsigned long)))
				return -EFAULT;

			err = audit_getmask((int) a[0], (int) a[1],  (void *) a[2]);

			break;
		}
		default :
			err = -EINVAL;
	}

	return err;
}


static int audit_ok(struct audit_svc_info *svc)
{
	if (!AU_SVC_INITIALIZED(svc))
		return 0;

	if (!svc->auditd_task || svc->auditd_task->pid != svc->auditd_pid) {
		struct task_struct *au_task;

		au_task = find_task_by_pid(svc->auditd_pid);

		if (!au_task) {
			audit_abort();
			return 0;
		}
	}

	return 1;
}

int audit_doit(unsigned int event, int status)
{
	int ret;

	ret = 0;
	if (audit_ok(&au_svc)) {
		ret = 3;

		if (((AU_MIN_HOST_EVT <= event) && (event <= AU_MAX_HOST_EVT)) ||
		    ((AU_MIN_USER_EVT <= event) && (event <= AU_MAX_USER_EVT))) {
			if (current->audit) {
				ret = audit_evt_val(current->audit->msk, event, 2);
	
				if (ret < 0) {
#ifdef CONFIG_AUDIT_DEBUG
					if (ret == -EFAULT) {
						printk(KERN_DEBUG
						       "audit_doit: lost event mask for user %d \n",
							   current->audit->aid);
					}
#endif
					ret = 3;
				}
	
			}
		}
	
		if (ret && (ret < 3)) {
			if ((ret == 1 && (!status || status > 0)) || (ret == 2 && status < 0))
				ret = 0;
		}
	}
	
#ifdef CONFIG_AUDIT_PROC
	if (!ret)
		audit_evt_filtered(AU_HOST_REC);
#endif

	return ret;
}


#ifdef CONFIG_NET_AUDIT

int audit_net_doit(unsigned int evt, unsigned int subevt, struct sk_buff *skb,
                   int fam)
{
	int retval;

	retval = 0;
	if (audit_ok(&au_svc)) {
		retval = 1;

		switch (fam) {
#ifdef CONFIG_ACL
			case AF_INET:
				retval = acl_check_net_ipv4(evt, subevt, skb);
				break;
			case AF_INET6:
				retval = acl_check_net_ipv6(evt, subevt, skb);
				break;
#endif
		}

		if (retval < 0) {
			retval = 1;
		}

#ifdef CONFIG_AUDIT_PROC
		if (!retval)
			audit_evt_filtered(AU_NET_REC);
#endif
	}

	return retval;
}

#endif
