#ifdef __KERNEL__

#include <linux/sched.h>
#include <linux/list.h>
#include <linux/string.h>
#include <linux/wait.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>

#include <asm/atomic.h>
#include <asm/uaccess.h>

#else

#include "structs.h"

#endif

#include <linux/acl.h>

struct audit_acl_msk acl_msk_list;

static spinlock_t acl_msk_lock = SPIN_LOCK_UNLOCKED;
static kmem_cache_t *au_msk_cachep = NULL;

static inline void msk_cache_init(void *p, kmem_cache_t *cache,
                                  unsigned long flags)
{
	struct audit_acl_msk *new_p = p;

	memset(new_p, 0, sizeof(struct audit_acl_msk));
}


int init_msk_cachep(void)
{
	if (au_msk_cachep)
		return 0;
	
	acl_msk_list.next = &acl_msk_list;
	acl_msk_list.prev = &acl_msk_list;

	atomic_set(&acl_msk_list.count, 0);
	memset((char *) acl_msk_list.perm_msk32, 0xff, AU_MSK_SIZE8);
#ifdef CONFIG_AUDIT_DEBUG
	memset((char *) acl_msk_list.val_msk32, 0xff, AU_MSK_SIZE8);
#else
	memset((char *) acl_msk_list.val_msk32, 0, AU_MSK_SIZE8);
#endif

	au_msk_cachep = kmem_cache_create("audit_msk_cachep",
	                                  sizeof(struct audit_acl_msk), 0,
	                                  SLAB_HWCACHE_ALIGN, msk_cache_init, NULL);
	
	if (!au_msk_cachep)
		return -ENOMEM;

	return 0;
}


int kfree_audit_msk(struct audit_acl_msk *msk)
{
	if (!au_msk_cachep)
		return -EFAULT;
	
	if (msk) {
		if (msk == &acl_msk_list)
			return -EINVAL;

		msk->prev->next = msk->next;
		msk->next->prev = msk->prev;

		kmem_cache_free(au_msk_cachep, msk);
	}

	return 0;
}


static struct audit_acl_msk *acl_get_msk_entry(__u32 *perm_msk, __u32 *evt_msk,
                                               int flag)
{
	struct audit_acl_msk *msk;
	int i, match;

	msk = &acl_msk_list;
	do {
		match = 1;

		for (i = 0; match && i < AU_MSK_SIZE32; i++) {
			if (msk->perm_msk32[i] != perm_msk[i])
				match = 0;
			if (msk->val_msk32[i] != evt_msk[i])
				match = 0;
		}

		if (match) {
			atomic_inc(&msk->count);
			return msk;
		}

		msk = msk->next;
	} while (msk != &acl_msk_list);

	msk = (struct audit_acl_msk *) kmem_cache_alloc(au_msk_cachep, flag);
	if (!msk)
		return ERR_PTR(-ENOMEM);

	memcpy((char *) msk->perm_msk8, (char *) perm_msk, AU_MSK_SIZE8);
	memcpy((char *) msk->val_msk8, (char *) evt_msk, AU_MSK_SIZE8);

	atomic_set(&msk->count, 1);
		
	msk->next = acl_msk_list.next;
	msk->prev = &acl_msk_list;
	acl_msk_list.next->prev = msk;
	acl_msk_list.next = msk;

	return msk;
}
	

static int update_evt_msk(__u32 *perm_msk, __u32 *val_msk, int evt, int val,
                          unsigned int evt_width)
{
	int max_val;

	if (!evt_width || (evt_width > 31)) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "update_evt_msk: bad event width given %d\n",
		       evt_width);
#endif
		return -EINVAL;
	}

	max_val = (1 << evt_width) - 1;

	if (val < 0 || val > max_val) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "update_evt_msk: bad value given %d\n", val);
#endif
		return -EINVAL;
	}

	if ((0 <= evt) && (evt < ((AU_MSK_SIZE8 << 3) >> (evt_width - 1)))) {
		__u32 invert, mask;
		int row;

		invert = max_val << (evt_width * (evt % (32 >> (evt_width - 1))));
		invert = ~invert;

		mask = val << (evt_width * (evt % (32 >> (evt_width - 1))));

		row = evt >> (5 - (evt_width - 1));
		val_msk[row] &= invert;
		val_msk[row] |= mask;

		mask = 1 << (evt % 32);

		row = evt >> 5;
		invert = ~mask;

		perm_msk[row] &= invert;
		perm_msk[row] |= mask;

		return 0;
	}
	
	return -EINVAL;
}


int audit_evt_val(struct audit_acl_msk *msk, int evt, unsigned int evt_width)
{
	int ret;

	if (!msk)
		return -EFAULT;

#ifdef CONFIG_AUDIT_DEBUG
	if (evt_width > 16) {
		printk(KERN_DEBUG "audit_evt_val: to large event width.\n");
		BUG();
	}
#endif

	if (0 > evt || evt > ((AU_MSK_SIZE8 << 3) >> (evt_width - 1))) {
#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "audit_evt_val: bad event given %d\n", evt);
#endif
		return -EINVAL;
	}
	
	ret = -EBUSY;
	if (spin_trylock(&acl_msk_lock)) {
		unsigned long perm_msk, val_msk;

		perm_msk = msk->perm_msk32[evt >> 5];
		val_msk = msk->val_msk32[(evt * evt_width) >> 5];

		spin_unlock(&acl_msk_lock);

		ret = 0;
		if (val_msk) {
			int pos;
			
			pos = evt & ((1 << 5) - 1);
			if ((perm_msk >> pos) & 1) {
				pos = evt % (32 / evt_width);
				ret = (val_msk >> (pos * evt_width)) & ((1 << evt_width) - 1);
			}
		} 
	}

	return ret;
}


int audit_get_acl_msk(struct audit_acl_msk *entry, __u32 *msk)
{
	spin_lock(&acl_msk_lock);

	if (entry) {
		memcpy((char *) msk, (char *) entry->val_msk8, AU_MSK_SIZE8);
	} else
		memset((char *) msk, 0, AU_MSK_SIZE8);
	
	spin_unlock(&acl_msk_lock);

	return 0;
}


int audit_update_evt_msk(struct audit_acl_msk **entry_p, int evt, int val,
                         int evt_width, int alloc_flag)
{
	struct audit_acl_msk *entry;
	__u32 perm_msk[AU_MSK_SIZE32];
	__u32 val_msk[AU_MSK_SIZE32];
	int err;

	spin_lock(&acl_msk_lock);
	
	if (!entry_p) {
		spin_unlock(&acl_msk_lock);
		return -EFAULT;
	}

	entry = *entry_p;

	if (entry) {
		if (atomic_read(&entry->count) == 1) {
			err = update_evt_msk(entry->perm_msk32, entry->val_msk32,
			                     evt, val, evt_width);

			if (err) {
				kfree_audit_msk(entry);
				entry = NULL;
			}

			goto out;
		}

		memcpy((char *) perm_msk, (char *) entry->perm_msk8, AU_MSK_SIZE8);
		memcpy((char *) val_msk, (char *) entry->val_msk8, AU_MSK_SIZE8);

		if (atomic_dec_and_test(&entry->count)) {
			kfree_audit_msk(entry);
			entry = NULL;
		}
	} else {
		memset((char *) perm_msk, 0, AU_MSK_SIZE8);
		memset((char *) val_msk, 0, AU_MSK_SIZE8);
	}

	err = update_evt_msk(perm_msk, val_msk, evt, val, evt_width);
	if (!err) {
		entry = acl_get_msk_entry(perm_msk, val_msk, alloc_flag);
		if (IS_ERR(entry)) {
			err = PTR_ERR(entry);
			entry = NULL;
		}
	}

out:
	spin_unlock(&acl_msk_lock);

	*entry_p = entry;
	return err;
}


int audit_set_evt_msk(struct audit_acl_msk **entry_p, __u32 *evt_msk,
                      int alloc_flag)
{
	struct audit_acl_msk *entry;
	__u32 perm_msk[AU_MSK_SIZE32];
	__u32 val_msk[AU_MSK_SIZE32];
	int err;

	spin_lock(&acl_msk_lock);

	if (!entry_p) {
		spin_unlock(&acl_msk_lock);
		return -EFAULT;
	}

	entry = *entry_p;

	if (entry) {
		if (atomic_read(&entry->count) == 1) {
			memset((char *) entry->perm_msk8, 0xff, AU_MSK_SIZE8);
			memcpy((char *) entry->val_msk8, (char *) evt_msk,
			       AU_MSK_SIZE8);

			err = 0;
			goto out;
		}

		if (atomic_dec_and_test(&entry->count)) {
			kfree_audit_msk(entry);
			entry = NULL;
		}
	}
	
	memset((char *) perm_msk, 0xff, AU_MSK_SIZE8);
	memcpy((char *) val_msk, (char *) evt_msk, AU_MSK_SIZE8);

	err = 0;

	entry = acl_get_msk_entry(perm_msk, evt_msk, alloc_flag);
	if (IS_ERR(entry)) {
		err = PTR_ERR(entry);
		entry = NULL;
	}

out:
	*entry_p = entry;
	spin_unlock(&acl_msk_lock);

	return err;

}

int audit_copy_evt_msk(struct audit_struct *dest, struct audit_struct *src)
{
	spin_lock(&acl_msk_lock);

	dest->msk = src->msk;
	if (!dest->msk)
		dest->msk = &acl_msk_list;

	atomic_inc(&dest->msk->count);

	spin_unlock(&acl_msk_lock);

	return 0;
}
