#include <asm/uaccess.h>

#include <linux/slab.h>
#include <linux/audit.h>

#include "buffer.h"

static kmem_cache_t *au_host_buf_cache = NULL;
DECLARE_WAIT_QUEUE_HEAD(au_hostbuf_wait);

#ifdef CONFIG_AUDIT_PROC
extern unsigned int au_dummy[];
#endif

struct audit_dyn_buf *au_env_buf = NULL;

static inline void host_buf_init(void *p, kmem_cache_t *cache,
                                 unsigned long flags)
{
	struct audit_host_record *new_p = p;

	memset(new_p, 0, sizeof(struct audit_host_record));
	new_p->state = AU_FREE;
	new_p->next = NULL;
}


int init_host_buf_cache(void)
{
	if (au_host_buf_cache)
		return 0;
	
#ifdef CONFIG_AUDIT_DEBUG
	printk(KERN_DEBUG "init_host_buf_cache: initializing host buffer cache.\n");
#endif
	
	au_host_buf_cache = kmem_cache_create("au_host_buf_cache",
	                                      sizeof(struct audit_host_record), 0,
	                                      SLAB_HWCACHE_ALIGN,
	                                      host_buf_init, NULL);
	
	if (!au_host_buf_cache) {
		printk(KERN_NOTICE "audit: init host buffer cache failed !!!\n");
		return -ENOMEM;
	}

	return 0;
}

static int audit_close_buf(struct audit_buffer_info *info)
{
	if (!info)
		return -EFAULT;

	if (info->state != AU_SEND_ALRM || info->free_recs == 0) {
		if (!wake_up_auditd())
			info->state = AU_SEND_ALRM;
	}

	return 0;
}

static int audit_destroy_host_buffer(struct audit_buffer_info *info)
{
	if (!info || info->type != AU_HOST_REC)
		return -EINVAL;

	if (info->buf_addr) {
		struct audit_host_record *p, *q;

		p = (struct audit_host_record *) info->buf_addr;

		p = p->next;
		while (p && p->next != info->buf_addr) {
			q = p->next;
			kmem_cache_free(au_host_buf_cache, p);

			p = q;
		}

		kmem_cache_free(au_host_buf_cache, info->buf_addr);
	}		
	
	info->buf_size = 0;
	info->rec_size = 0;
	info->free_recs = 0; 
	info->buf_wm = 0;

	info->dropped = 0;

	info->buf_addr = NULL;
	info->cur_rec = NULL;

	info->state = AU_FULL;
	info->type = AU_REC_TYPES;

	audit_delete_dyn_buf(au_env_buf);
	au_env_buf = NULL;

	return 0;
}

static int audit_init_host_buffer(struct audit_buffer_info *p,
                                  unsigned int cnt)
{
	const int rec_size = sizeof(struct audit_host_record);
	int retval = 0;

	if (!p || cnt < 0)
		return -EINVAL;

	p->type = AU_HOST_REC;

	if (cnt) {
		struct audit_host_record *new_p, *last;
		unsigned int i;

		if (!au_host_buf_cache)
			return -ENOMEM;

		last = NULL;
		p->buf_addr = NULL;

		retval = -ENOMEM;
		for (i = 0; i < cnt; i++) {
			new_p = kmem_cache_alloc(au_host_buf_cache, GFP_KERNEL);
			if (!new_p)
				goto err_out;

			memset(new_p, 0, sizeof(struct audit_host_record));
			new_p->state = AU_FREE;
			new_p->next = NULL;

			if (!p->buf_addr)
				p->buf_addr = (void *) new_p;
			else
				last->next = new_p;

			last = new_p;
		}

		last->next = (struct audit_host_record *) p->buf_addr;

		retval = 0;
	} else
		p->buf_addr = NULL;

	p->buf_size = cnt;
	p->rec_size = rec_size;
	p->free_recs = cnt;

	if ((p->buf_wm = (unsigned int) (cnt / 10)) < AU_MIN_WATERMARK)
		p->buf_wm = AU_MIN_WATERMARK;

	p->cur_rec = p->buf_addr;

	p->state = AU_FREE;

out:
	return retval;

err_out:
	audit_destroy_host_buffer(p);
	goto out;
}


static int audit_reduse_host_buffer(struct audit_buffer_info *info)
{
	struct audit_host_record *cur;

	if (!info)
		return -EFAULT;

	cur = (struct audit_host_record *) info->buf_addr;

	if (cur) {
		float wm_value;

		struct audit_host_record *tmp;
		struct audit_host_record *prev;

		wm_value = info->buf_wm / info->buf_size;

		prev = cur;
		cur = cur->next;

		while (cur && cur != info->buf_addr) {
			if (cur->state == AU_FREE) {
				tmp = cur;

				cur = cur->next;
				prev->next = cur;

				if (tmp->env && tmp->env->state != AU_FREE)
					tmp->env->state = AU_FREE;

				kmem_cache_free(au_host_buf_cache, tmp);

				info->buf_size--;
				info->free_recs--;

				continue;
			}

			prev = cur;
			cur = cur->next;
		}

		tmp = (struct audit_host_record *) info->buf_addr;
		if (tmp->state == AU_FREE) {
			if (tmp == tmp->next) {

				info->buf_size = 0;
				info->free_recs = 0;

				info->buf_addr = NULL;
				info->cur_rec = NULL;
			} else {
				prev = tmp->next;

				info->buf_addr = tmp->next;

				info->buf_size--;
				info->free_recs--;
			}

			if (tmp->env && tmp->env->state == AU_FULL)
				tmp->env->state = AU_FREE;

			kmem_cache_free(au_host_buf_cache, tmp);
		}

		info->buf_wm = (unsigned int) (info->buf_size * wm_value);
	}
			
	return 0;
}

static int audit_ctl_host_buffer(struct audit_buffer_info *info,
                                 unsigned long cmd, unsigned long arg)
{
	int ret = 0;

	if (!info)
		return -EFAULT;

	switch (cmd) {
		case AU_BUF_SET_SIZE: {
			ret = -ENOSYS;
			break;
		}
		case AU_BUF_SET_WM: {
			int new_wm, old_wm;

			new_wm = (int) arg;

			old_wm = info->buf_wm;

			info->buf_wm = (unsigned int) ((info->buf_size * 100) / new_wm);
			if (info->buf_wm < AU_MIN_WATERMARK) {
				info->buf_wm = AU_MIN_WATERMARK;

				ret = 0;
			} else {
				if (info->buf_wm > info->buf_size) {
					info->buf_wm = old_wm;

					ret = -EINVAL;
				} else {
					ret = 0;
				}
			}

			if (info->buf_wm > info->free_recs)
				audit_close_buf(info);

			break;
		}
		case AU_DYN_BUF_SET_WM: {
			if (!au_env_buf) {
				ret = -ENOMEM;
				au_env_buf = audit_dyn_buf_init("host");
			}

			if (au_env_buf) {
				ret = audit_dyn_buf_update_wm(au_env_buf, (int) arg);
			}
					
			break;
		}
		case AU_DYN_BUF_SET_ORDER: {
			if (!au_env_buf) {
				ret = -ENOMEM;
				au_env_buf = audit_dyn_buf_init("host");
			}

			if (au_env_buf) {
				ret = audit_dyn_update_order(&au_env_buf, (int) arg);
			}

			break;
		}
		default: {
			ret = -EINVAL;
		}
	}

	return ret;
}

static void *audit_get_free_host_rec(struct audit_buffer_info *info)
{
	struct audit_host_record *p;

	if (info->type != AU_HOST_REC)
		return ERR_PTR(-EINVAL);

	if (info->buf_wm > info->free_recs)
		audit_close_buf(info);

	p = ERR_PTR(-ENOSPC);
	if (info->free_recs > 0) {
		struct audit_host_record *nfree;
		
		nfree = (struct audit_host_record *) info->cur_rec;
		
		p = ERR_PTR(-EFAULT);
		if (nfree->next) {
			p = ERR_PTR(-ENOSPC);

			if (nfree->state == AU_FREE) {
				info->cur_rec = nfree->next;
				info->free_recs--;

				nfree->state = AU_FULL;

				p = nfree;
			}
		}
	}

	return p;
}

static int audit_host_rec2chr(struct audit_buffer_info *info, char *buf,
                              struct audit_host_record *rec)
{
	int size;

	if (info->type != AU_HOST_REC) {
		printk(KERN_INFO "audit_host_rec2chr: invalid record type %d.\n",
		       info->type);

#ifdef CONFIG_AUDIT_DEBUG
		BUG();
#else
		return -EINVAL;
#endif
	}

	if (!rec)
		return -EFAULT;

	if (rec->state != AU_FULL) {
		if (rec->env)
			rec->env->state = AU_FREE;

		KLADT_PRINT("audit_host_rec2chr: "
		            "expect full record, but state was %d (add?%d)\n",
		             rec->state, (!(rec->next) ? 1 : 0));

		size = -ENODATA;
	} else {
		int err;

		memcpy(buf, rec, AU_HOST_REC_FIX);
		size = AU_HOST_REC_FIX;
		buf += AU_HOST_REC_FIX;

		err = audit_dyn_rec2chr(buf, rec->env);
		if (err > 0) {
			size += err;
			buf += err;

			memcpy(buf, rec->file + (unsigned short)(rec->file_len >> 16),
		          (unsigned short) rec->file_len);
			buf += (unsigned short) rec->file_len;
			size += (unsigned short) rec->file_len;

			*buf = '\0';
			buf++;
			size++;

			memcpy(buf, rec->file2 + (unsigned short)(rec->file2_len >> 16),
			       (unsigned short) rec->file2_len);
			buf += (unsigned short) rec->file2_len;
			size += (unsigned short) rec->file2_len;

			*buf = '\0';
			buf++;
			size++;
		} else
			size = err;
	}

	rec->env = NULL;

	return size;
}

static void *audit_get_new_host_rec(void)
{
	struct audit_host_record *new_p;

	new_p = kmem_cache_alloc(au_host_buf_cache, GFP_ATOMIC);
	if (new_p) {
		memset(new_p, 0, sizeof(struct audit_host_record));
		new_p->state = AU_FULL;
		new_p->next = NULL;
	}

	return new_p;
}

static int audit_free_host_rec(struct audit_buffer_info *info,
                               struct audit_host_record *rec)
{

	if (!rec)
		return -EFAULT;

	if (rec->env) {
		rec->env->state = AU_FREE;
		rec->env = NULL;
	}

		
	if (rec->next) {
		if (rec->state != AU_FREE) {
			rec->state = AU_FREE;

			info->free_recs++;
			if (info->buf_wm < info->free_recs)
				info->state = AU_FREE;
		}
	} else {
		kmem_cache_free(au_host_buf_cache, rec);
	}

	return 0;
}

static int audit_host_suspend(void)
{
	sleep_on(&au_hostbuf_wait);
	return 0;
}

static void audit_host_wake_up(void)
{
	wake_up(&au_hostbuf_wait);
}

struct audit_buffer_info au_host_buf_info;

int audit_init_host(struct audit_buffer_info **info)
{
	if (!info)
		return -EFAULT;

	if (info[AU_HOST_REC] == &au_host_buf_info)
		return 0;
	
	au_host_buf_info.state = AU_UNDEF;
	au_host_buf_info.type = AU_HOST_REC;
	au_host_buf_info.rec_size = 0;
	au_host_buf_info.buf_size = 0;
	au_host_buf_info.buf_wm = 0;
	au_host_buf_info.free_recs = 0;
	au_host_buf_info.buf_addr = NULL;
	au_host_buf_info.cur_rec = NULL;

	au_host_buf_info.init = audit_init_host_buffer;
	au_host_buf_info.destroy = audit_destroy_host_buffer;
	au_host_buf_info.reduse = audit_reduse_host_buffer;
	au_host_buf_info.ctl = audit_ctl_host_buffer;
	au_host_buf_info.get_new = audit_get_new_host_rec;
	au_host_buf_info.get_free = audit_get_free_host_rec;
	au_host_buf_info.buf_suspend = audit_host_suspend;
	au_host_buf_info.buf_wake_up = audit_host_wake_up;

	au_host_buf_info.rec2chr = (int (*) (struct audit_buffer_info *, char *, void *)) audit_host_rec2chr;
	au_host_buf_info.free_rec = (int (*) (struct audit_buffer_info *, void *)) audit_free_host_rec;

	info[AU_HOST_REC] = (struct audit_buffer_info *) &au_host_buf_info;

	return 0;
}
