#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/module.h>

#include <linux/audit.h>

#include "buffer.h"

#ifdef CONFIG_AUDIT_PROC
extern unsigned int au_dummy[];
#endif

extern int audit_getmodus(void);

static kmem_cache_t *au_net_buf_cache = NULL;

DECLARE_WAIT_QUEUE_HEAD(au_netbuf_wait);

extern int init_dyn_buf_cache(void);
struct audit_dyn_buf *au_net_env_buf = NULL;

static inline void net_buf_init(void *p, kmem_cache_t *cache,
                                unsigned long flags)
{
	struct audit_net_record *new_p = p;

	memset(new_p, 0, sizeof(struct audit_net_record));
	new_p->state = AU_FREE;
	new_p->next = NULL;
}


int init_net_buf_cache(void)
{
	if (au_net_buf_cache)
		return 0;
	
	KLADT_PRINT("init_net_buf_cache: initializing net buffer cache.\n");
	
	au_net_buf_cache = kmem_cache_create("au_net_buf_cache",
	                                     sizeof(struct audit_net_record), 0,
	                                     SLAB_HWCACHE_ALIGN,
	                                     net_buf_init, NULL);
	
	if (!au_net_buf_cache) {
		printk(KERN_NOTICE "audit: init net buffer cache failed !!!\n");
		return -ENOMEM;
	}

	return 0;
}

static int audit_close_buf(struct audit_buffer_info *info)
{
	if (!info)
		return -EFAULT;

	if (info->state != AU_SEND_ALRM || info->free_recs == 0) {
		if (!wake_up_auditd())
			info->state = AU_SEND_ALRM;
	}

	return 0;
}


int audit_check_net_buffer(void)
{
	if (AU_SVC_INITIALIZED(&au_svc)) {
		if ((au_buf.first_free->state == AU_FULL) &&
		    (audit_getmodus() == AU_SECURE)) {

#ifdef CONFIG_AUDIT_DEBUG
			if (au_buf.nfs_root) {
				au_buf.netbuf_closed = 1;
				return AU_FREE;
			}
#else
			return AU_FULL;
#endif
		}
	}

	return AU_FREE;
}

static int audit_destroy_net_buffer(struct audit_buffer_info *info)
{

	if (!info || info->type != AU_NET_REC)
		return -EINVAL;

	if (info->buf_addr) {
		struct audit_net_record *p, *q;

		p = (struct audit_net_record *) info->buf_addr;

		p = p->next;
		while (p && p->next != info->buf_addr) {
			q = p->next;
			kmem_cache_free(au_net_buf_cache, p);

			p = q;
		}

		kmem_cache_free(au_net_buf_cache, info->buf_addr);
	}		
	
	info->buf_size = 0;
	info->rec_size = 0;
	info->free_recs = 0; 
	info->buf_wm = 0;

	info->dropped = 0;

	info->buf_addr = NULL;
	info->cur_rec = NULL;

	info->state = AU_UNDEF;
	info->type = AU_REC_TYPES;

	audit_delete_dyn_buf(au_net_env_buf);
	au_net_env_buf = NULL;

	return 0;
}

static int audit_init_net_buffer(struct audit_buffer_info *p, unsigned int cnt)
{
	const int rec_size = sizeof(struct audit_net_record);
	int retval = 0;

	if (!p || cnt < 0)
		return -EINVAL;

	p->type = AU_NET_REC;

	if (cnt) {
		struct audit_net_record *new_p, *last = NULL;
		unsigned int i;

		if (!au_net_buf_cache)
			return -ENOMEM;

		p->buf_addr = NULL;

		retval = -ENOMEM;
		for (i = 0; i < cnt; i++) {
			new_p = kmem_cache_alloc(au_net_buf_cache, GFP_KERNEL);
			if (!new_p)
				goto err_out;

			memset(new_p, 0, sizeof(struct audit_net_record));
			new_p->state = AU_FREE;
			new_p->next = NULL;

			if (!p->buf_addr)
				p->buf_addr = (void *) new_p;
			else
				last->next = new_p;

			last = new_p;
		}

		last->next = (struct audit_net_record *) p->buf_addr;

		retval = 0;
	} else
		p->buf_addr = NULL;

	p->buf_size = cnt;
	p->rec_size = rec_size;
	p->free_recs = cnt;

	if ((p->buf_wm = (unsigned int) (cnt / 10)) < AU_MIN_WATERMARK)
		p->buf_wm = AU_MIN_WATERMARK;

	p->cur_rec = p->buf_addr;

	p->state = AU_FREE;

out:
	return retval;

err_out:
	audit_destroy_net_buffer(p);
	goto out;
}

static int audit_reduse_net_buffer(struct audit_buffer_info *info)
{
	struct audit_net_record *cur;

	if (!info)
		return -EFAULT;

	cur = (struct audit_net_record *) info->buf_addr;

	if (cur) {
		float wm_value;

		struct audit_net_record *tmp;
		struct audit_net_record *prev;

		wm_value = info->buf_wm / info->buf_size;

		prev = cur;
		cur = cur->next;

		while (cur && cur != info->buf_addr) {
			if (cur->state == AU_FREE) {
				tmp = cur;

				cur = cur->next;
				prev->next = cur;

				kmem_cache_free(au_net_buf_cache, tmp);

				info->buf_size--;
				info->free_recs--;

				continue;
			}

			prev = cur;
			cur = cur->next;
		}

		tmp = (struct audit_net_record *) info->buf_addr;
		if (tmp->state == AU_FREE) {
			if (tmp == tmp->next) {
				info->buf_size = 0;
				info->free_recs = 0;

				info->buf_addr = NULL;
				info->cur_rec = NULL;
			} else {
				prev = tmp->next;

				info->buf_addr = tmp->next;

				info->buf_size--;
				info->free_recs--;
			}

			kmem_cache_free(au_net_buf_cache, tmp);
		}

		info->buf_wm = (unsigned int) (info->buf_size * wm_value);
	}
			
	return 0;
}

static int audit_ctl_net_buffer(struct audit_buffer_info *info,
                                unsigned long cmd, unsigned long arg)
{
	int retval;

	if (!info)
		return -EFAULT;
	
	switch (cmd) {
		case AU_BUF_SET_SIZE: {
			retval = -ENOSYS;
			break;
		}
		case AU_BUF_SET_WM: {
			int new_wm, old_wm;

			new_wm = (int) arg;
			old_wm = info->buf_wm;

			info->buf_wm = (unsigned int) ((info->buf_size * 100) / new_wm);
			if (info->buf_wm < AU_MIN_WATERMARK) {
#ifdef CONFIG_AUDIT_DEBUG
				printk(KERN_DEBUG "audit_ctl_net_buffer: "
				                  "to small watermark given for the net record "
				                  "buffer, set it to %d.\n",
				       info->buf_wm);
#endif
				info->buf_wm = AU_MIN_WATERMARK;

				retval = 0;
			} else {
				if (info->buf_wm > info->buf_size) {
#ifdef CONFIG_AUDIT_DEBUG
					printk(KERN_DEBUG "audit_ctl_net_buffer: "
					                  "the given watermark for the net record "
					                  "buffer is to large.\n");
#endif
					info->buf_wm = old_wm;

					retval = -EINVAL;
				} else {
					retval = 0;
				}
			}

			if (info->buf_wm > info->free_recs) {
				audit_close_buf(info);
			}

			break;
		}
		case AU_DYN_BUF_SET_WM: {
			if (!au_net_env_buf) {
				retval = -ENOMEM;
				au_net_env_buf = audit_dyn_buf_init("net");
			}

			if (au_net_env_buf) {
				retval = audit_dyn_buf_update_wm(au_net_env_buf, (int) arg);
			}
					
			break;
		}
		case AU_DYN_BUF_SET_ORDER: {
			if (!au_net_env_buf) {
				retval = -ENOMEM;
				au_net_env_buf = audit_dyn_buf_init("net");
			}

			if (au_net_env_buf) {
				retval = audit_dyn_update_order(&au_net_env_buf, (int) arg);
			}

			break;
		}
		default: {
			retval = -EINVAL;
		}
	}

	return retval;
}

static void *audit_get_free_net_rec(struct audit_buffer_info *info)
{
	struct audit_net_record *p;

	if (info->type != AU_NET_REC)
		return ERR_PTR(-EINVAL);

	if (info->buf_wm > info->free_recs) {
		audit_close_buf(info);
	}

	p = ERR_PTR(-ENOSPC);
	if (info->free_recs > 0) {
		struct audit_net_record *nfree;

		nfree = (struct audit_net_record *) info->cur_rec;

		p = ERR_PTR(-EFAULT);
		if (nfree->next) {
			p = ERR_PTR(-ENOSPC);

			if (nfree->state == AU_FREE) {
				info->cur_rec = nfree->next;
				info->free_recs--;

				nfree->state = AU_FULL;

				p = nfree;
			}
		}
	}

	return p;
}

static void *audit_get_new_net_rec(void)
{
	struct audit_net_record *new_p;

	new_p = kmem_cache_alloc(au_net_buf_cache, GFP_ATOMIC);
	if (new_p) {
		memset(new_p, 0, sizeof(struct audit_net_record));
		new_p->state = AU_FULL;
	}

	return new_p;
}

static int audit_net_rec2chr(struct audit_buffer_info *info, char *buf,
                             struct audit_net_record *rec)
{
	int size;

	if (!rec)
		return -EFAULT;

	if (rec->state != AU_FULL) {
		if (rec->env)
			rec->env->state = AU_FREE;

		size = -ENODATA;
	} else {
		int err;

		memcpy(buf, (char *) rec, AU_NET_REC_FIX);

		size = AU_NET_REC_FIX;
		buf += AU_NET_REC_FIX;

		if (rec->mac_prot_size) {
			memcpy(buf, rec->mac.raw, rec->mac_prot_size);
			size += rec->mac_prot_size;
			buf += rec->mac_prot_size;
		}
		if (rec->nh_prot_size) {
			memcpy(buf, rec->nh.raw, rec->nh_prot_size);
			size += rec->nh_prot_size;
			buf += rec->nh_prot_size;
		}
		if (rec->h_prot_size) {
			memcpy(buf, rec->h.raw, rec->h_prot_size);
			size += rec->h_prot_size;
			buf += rec->h_prot_size;
		}

		err = audit_dyn_rec2chr(buf, rec->env);
		if (err < 0) {
			size = err;
		} else {
			size += err;
			buf += err;
	
			if (rec->namelen) {
				memcpy(buf, rec->name, rec->namelen);
				size += rec->namelen;
				buf += rec->namelen;
			}
		}
	}

	rec->env = NULL;

	return size;
}

static int audit_free_net_rec(struct audit_buffer_info *info,
                              struct audit_net_record *rec)
{

	if (!rec)
		return -EFAULT;

	if (rec->env) {
		rec->env->state = AU_FREE;
		rec->env = NULL;
	}

	if (rec->next) {
		if (rec->state != AU_FREE) {
			rec->state = AU_FREE;

			info->free_recs++;
			if (info->buf_wm < info->free_recs) {
				info->state = AU_FREE;
#ifdef CONFIG_AUDIT_DEBUG
				au_buf.netbuf_closed = 0;
#endif
			}
		}
	} else {
		kmem_cache_free(au_net_buf_cache, rec);
	}

	return 0;
}

static int audit_net_suspend(void)
{
	sleep_on(&au_netbuf_wait);
	return 0;
}

static void audit_net_wake_up(void)
{
	wake_up(&au_netbuf_wait);
}


struct audit_buffer_info au_net_buf_info;

int audit_init_net(struct audit_buffer_info **info)
{
	if (!info)
		return -EFAULT;

	if (info[AU_NET_REC] != &au_net_buf_info) {
		au_net_buf_info.state = AU_UNDEF;
		au_net_buf_info.type = AU_NET_REC;
		au_net_buf_info.rec_size = 0;
		au_net_buf_info.buf_size = 0;
		au_net_buf_info.buf_wm = 0;
		au_net_buf_info.free_recs = 0;
		au_net_buf_info.buf_addr = NULL;
		au_net_buf_info.cur_rec = NULL;
		au_net_buf_info.init = audit_init_net_buffer;
		au_net_buf_info.destroy = audit_destroy_net_buffer;
		au_net_buf_info.reduse = audit_reduse_net_buffer;
		au_net_buf_info.ctl = audit_ctl_net_buffer;
		au_net_buf_info.get_free = audit_get_free_net_rec;
		au_net_buf_info.get_new = audit_get_new_net_rec;
		au_net_buf_info.buf_suspend = audit_net_suspend;
		au_net_buf_info.buf_wake_up = audit_net_wake_up;

		au_net_buf_info.rec2chr = (int (*) (struct audit_buffer_info *, char *, void *)) audit_net_rec2chr;
		au_net_buf_info.free_rec = (int (*) (struct audit_buffer_info *, void *)) audit_free_net_rec;

		info[AU_NET_REC] = (struct audit_buffer_info *) &au_net_buf_info;
	}

	return 0;
}
