#include <linux/audit.h>
#include <linux/string.h>
#include <linux/proc_fs.h>

#include "buffer.h"
#include "proc.h"
#include "version.h"

extern int audit_getmodus(void);

extern unsigned long au_rec_demand;
extern unsigned long au_rec_saved;
extern unsigned long au_rec_unused;
extern unsigned long au_max_add;
extern unsigned long au_suspend_max;


#define AUDIT_PRINT_PROC(page, left, err_label, format, args...) \
{ \
	char buffer[128]; \
	int str_len; \
 \
	str_len = snprintf(buffer, 128, format, ## args); \
	if ((left - str_len) < 0) \
		goto err_label; \
	memcpy(page + (PAGE_SIZE - left), buffer, str_len); \
	left -= str_len; \
}
	

#ifdef CONFIG_AUDIT_DEBUG

struct au_svc_proc_s au_svc_proc = AU_SVC_PROC_INIT;

#endif

extern int audit_get_dyn_buf_info(char *buf, unsigned int size);

int audit_get_proc(char *page)
{
	int tmp, left = PAGE_SIZE;

	if (!audit_check_capability(AU_CAP_ADM))
		return 0;

	AUDIT_PRINT_PROC(page, left, out, "version: %s\n", AUDIT_VERSION);

	switch (atomic_read(&au_svc.state)) {
	  case AU_DOIT:
		AUDIT_PRINT_PROC(page, left, out, "state: running\n");
		break;
	  case AU_STOPPED:
		AUDIT_PRINT_PROC(page, left, out, "state: stopped\n");
		break;
	  case AU_UNDEF:
		AUDIT_PRINT_PROC(page, left, out, "state: disabled\n");
		break;
	  default:
		AUDIT_PRINT_PROC(page, left, out, "state: unknown (error)\n");
	}

	switch (audit_getmodus()) {
	  case AU_DISABLED:
		AUDIT_PRINT_PROC(page, left, out, "mode: disabled\n");
		break;
	  case AU_NORMAL:
		AUDIT_PRINT_PROC(page, left, out, "mode: normal\n");
		break;
	  case AU_SECURE:
		AUDIT_PRINT_PROC(page, left, out, "mode: secure\n");
		break;
	  default:
		AUDIT_PRINT_PROC(page, left, out, "mode: unknown (error)\n");
	}
		
	if (au_svc.auditd_task) {
		AUDIT_PRINT_PROC(page, left, out, "daemon pid: %d \n\n",
		                 au_svc.auditd_pid);
	}

	AUDIT_PRINT_PROC(page, left, out,
	                 "\nrecords informations: \n          demand: %lu \n",
	                 au_rec_demand);
	AUDIT_PRINT_PROC(page, left, out, "           saved: %lu\n", au_rec_saved);
	AUDIT_PRINT_PROC(page, left, out, "         ignored: %lu\n", au_rec_unused);
	AUDIT_PRINT_PROC(page, left, out, "  max additional: %lu\n", au_max_add);
	AUDIT_PRINT_PROC(page, left, out, "     max suspend: %lu\n",
	                 au_suspend_max);

	AUDIT_PRINT_PROC(page, left, out, "\nStatic buffers:\n");
	
	for (tmp = 0; tmp < AU_REC_TYPES; tmp++) {
		if (!au_svc.buffers[tmp])
			continue;

		if (au_svc.buffers[tmp]->state != AU_UNDEF) {
			AUDIT_PRINT_PROC(page, left, out, "  buffer (%d)\n", tmp);
			AUDIT_PRINT_PROC(page, left, out, "     state: %d type: %d\n",
			                 au_svc.buffers[tmp]->state,
			                 au_svc.buffers[tmp]->type);
			AUDIT_PRINT_PROC(page, left, out, "   dropped: %lu filtered: %lu\n",
			                 au_svc.buffers[tmp]->dropped,
			                 au_svc.buffers[tmp]->filtered);
			AUDIT_PRINT_PROC(page, left, out, 
			                 "      size: %d wm: %d free_recs: %d\n",
			                 au_svc.buffers[tmp]->buf_size,
			                 au_svc.buffers[tmp]->buf_wm,
			                 au_svc.buffers[tmp]->free_recs);
		}
	}	

	tmp = audit_get_dyn_buf_info(page + (PAGE_SIZE - left), left);
	if (tmp > 0)
		left -= tmp;

out:
	return (PAGE_SIZE - left);
}


static int audit_read_proc(char *page, char **start, off_t off, int count,
                           int *eof, void *data)
{
	int len;
	
	len = audit_get_proc(page);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}


#ifdef CONFIG_AUDIT_DEBUG

static int audit_get_buf(char *page)
{
	int left = PAGE_SIZE;
	char *buffer = NULL;
	struct audit_record *au_rec;
	int tmp, len;

	if (!audit_check_capability(AU_CAP_ADM))
		return 0;

	if (au_buf.addr) {
		for (len = 1, au_rec = au_buf.addr; au_rec->next != au_buf.addr;
		     au_rec = au_rec->next)
			len++;
	
		AUDIT_PRINT_PROC(page, left, out, "audit_buffer (size = %d)\n", len);

		if (len < left) { 
			int true_len;

			true_len = len + ((len / 50) * (1 + 5)) + (len / 10) + 2;
			buffer = kmalloc(true_len, GFP_KERNEL);
			if (buffer) {
				int j, i = 0;

				memset(buffer, 0, true_len);

				au_rec = au_buf.addr;

				len = 0;

				for (j = 1; j < 5; j++, len++)
					buffer[len] = ' ';

				do {
					if (!au_rec)
						break;
	
					switch (au_rec->state) {
					  case AU_FULL:
						buffer[len] = '#';
						break;
					  case AU_FREE:
						buffer[len] = '-';
						break;
					  case AU_USED:
						buffer[len] = 'o';
						break;
					  case AU_UNUSED:
						buffer[len] = 'u';
						break;
					  default:
						buffer[len] = '?';
					}

				    au_rec = au_rec->next;

					len++;
					i++;

					if (i && (i % 50) == 0) {
						buffer[len++] = '\n';
						for (j = 1; j < 5; j++, len++)
							buffer[len] = ' ';
					} else {
						if (i && (i % 10) == 0)
							buffer[len++] = ' ';
					}
				} while (au_rec != au_buf.addr);

				buffer[len++] = '\n';
				memcpy(page + (PAGE_SIZE - left), buffer, len);
				left -= len;
	
				kfree(buffer);
				buffer = NULL;
			}
		}
	}

	for (tmp = 0; tmp < AU_REC_TYPES; tmp++) {
		if (!au_svc.buffers[tmp])
			continue;

		if (au_svc.buffers[tmp]->state == AU_UNDEF)
			continue;

		AUDIT_PRINT_PROC(page, left, out, "buffer (%d - %d)\n", tmp,
		                 au_svc.buffers[tmp]->state);
		AUDIT_PRINT_PROC(page, left, out,
		                 "   size: %d wm: %d dropped: %lu filtered: %lu free_recs: %d\n",
		                 au_svc.buffers[tmp]->buf_size,
		                 au_svc.buffers[tmp]->buf_wm,
		                 au_svc.buffers[tmp]->dropped,
		                 au_svc.buffers[tmp]->filtered,
		                 au_svc.buffers[tmp]->free_recs);

		len = au_svc.buffers[tmp]->buf_size;
		if (len < left) {
			struct audit_host_record *hp = NULL;
			struct audit_net_record *np = NULL;

			switch (tmp) {
			  case AU_HOST_REC:
				hp = (struct audit_host_record *) au_svc.buffers[tmp]->buf_addr;
				break;
			  case AU_NET_REC:
				np = (struct audit_net_record *) au_svc.buffers[tmp]->buf_addr;
				break;
			}

			if (hp || np)
				buffer = kmalloc(len + 2, GFP_KERNEL);

			if (buffer) {
				char *p = buffer;
				unsigned int state = 42, i;

				for (i = 0; i < len; i++) {
					switch (tmp) {
					  case AU_HOST_REC:
						if (!hp) {
							len = i;
							continue;
						}
						state = hp->state;
						hp = hp->next;
						break;
					  case AU_NET_REC:
						if (!np) {
							len = i;
							continue;
						}
						state = np->state;
						np = np->next;
						break;
					}

					switch (state) {
					  case AU_FREE:
						*p = '-';
						break;
					  case AU_FULL:
						*p = '#';
						break;
					  default:
						*p = '?';
					}

					p++;
				}

				buffer[len++] = '\n';
				buffer[len++] = 0;
				
				memcpy(page + (PAGE_SIZE - left), buffer, len);
				left -= len;

				kfree(buffer);
				buffer = NULL;
			}
		}
	}	

out:
	return (PAGE_SIZE - left);
}

static int audit_buf_proc(char *page, char **start, off_t off, int count,
                          int *eof, void *data)
{
	int len;
	
	len = audit_get_buf(page);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int audit_sys_svc_proc(char *page, char **start, off_t off, int count,
                              int *eof, void *data)
{
	int len = 0, left = PAGE_SIZE;
	
	if (spin_trylock(&au_svc_proc.lock)) {
		AUDIT_PRINT_PROC(page, left, out, "buffer status\n");

		AUDIT_PRINT_PROC(page, left, out, "  before:\n");
		AUDIT_PRINT_PROC(page, left, out,
		                 "   to_save -> addr 0x%08x index %d state %d\n",
		                 au_svc_proc.buf_start.to_save_addr,
		                 au_svc_proc.buf_start.to_save_index,
						 au_svc_proc.buf_start.to_save_state);
		AUDIT_PRINT_PROC(page, left, out,
		                 "      additional: %s\n",
		                 (au_svc_proc.buf_start.additional ? "yes" : "no"));
		AUDIT_PRINT_PROC(page, left, out,
		                 "   first_free -> addr 0x%08x index %d:\n",
		                 au_svc_proc.buf_start.first_free_addr,
		                 au_svc_proc.buf_start.first_free_index);
		AUDIT_PRINT_PROC(page, left, out,
		                 "   first_full -> addr 0x%08x index %d:\n",
		                 au_svc_proc.buf_start.first_full_addr,
		                 au_svc_proc.buf_start.first_full_index);

		AUDIT_PRINT_PROC(page, left, out, "  after:\n");
		AUDIT_PRINT_PROC(page, left, out,
		                 "   to_save -> addr 0x%08x index %d state %d\n",
		                 au_svc_proc.buf_end.to_save_addr,
		                 au_svc_proc.buf_end.to_save_index,
						 au_svc_proc.buf_start.to_save_state);
		AUDIT_PRINT_PROC(page, left, out,
		                 "      additional: %s\n",
		                 (au_svc_proc.buf_end.additional ? "yes" : "no"));
		AUDIT_PRINT_PROC(page, left, out,
		                 "   first_free -> addr 0x%08x index %d:\n",
		                 au_svc_proc.buf_end.first_free_addr,
		                 au_svc_proc.buf_end.first_free_index);
		AUDIT_PRINT_PROC(page, left, out,
		                 "   first_full -> addr 0x%08x index %d:\n",
		                 au_svc_proc.buf_end.first_full_addr,
		                 au_svc_proc.buf_end.first_full_index);

		AUDIT_PRINT_PROC(page, left, out, "\nlast record\n");
		AUDIT_PRINT_PROC(page, left, out,
		                 "   type: %lu state: %d\n",
		                 au_svc_proc.last_rec.type, au_svc_proc.last_rec.state);
		AUDIT_PRINT_PROC(page, left, out,
		                 "   pid: %lu event: %d\n",
		                 au_svc_proc.last_rec.pid, au_svc_proc.last_rec.event);

		AUDIT_PRINT_PROC(page, left, out, "\nerror report\n");
		AUDIT_PRINT_PROC(page, left, out, "   error: %d\n", au_svc_proc.error);
		AUDIT_PRINT_PROC(page, left, out,
		                 "   (%s) getdata, (%s) buf2file, (%s) rec2chr\n",
		                 (au_svc_proc.err_getdata ? "x" : "-"),
		                 (au_svc_proc.err_buf2file ? "x" : "-"),
		                 (au_svc_proc.err_rec2chr ? "x" : "-"));

		AUDIT_PRINT_PROC(page, left, out, "\nrecord2char parameter\n");
		AUDIT_PRINT_PROC(page, left, out, "   invoked: %d\n",
		                 au_svc_proc.rec2chr);
		AUDIT_PRINT_PROC(page, left, out, "   processed records\n");
		AUDIT_PRINT_PROC(page, left, out,
		                 "    full: %d, free: %d, unused %d, used %d, err %d\n",
		                 au_svc_proc.rec2chr_full, au_svc_proc.rec2chr_free,
		                 au_svc_proc.rec2chr_unused, au_svc_proc.rec2chr_used,
		                 au_svc_proc.rec2chr_err);

		AUDIT_PRINT_PROC(page, left, out, "\nbuf2file parameter\n");
		AUDIT_PRINT_PROC(page, left, out, "   invoked: %d\n",
		                 au_svc_proc.buf2file);

		AUDIT_PRINT_PROC(page, left, out, "\nnxt2save parameter\n");
		AUDIT_PRINT_PROC(page, left, out, "   invoked: %d\n",
		                 au_svc_proc.nxt2save);

		AUDIT_PRINT_PROC(page, left, out, "\nnxt2save parameter\n");
		AUDIT_PRINT_PROC(page, left, out, "   invoked: %d\n",
		                 au_svc_proc.getdata);
		AUDIT_PRINT_PROC(page, left, out, "   records without data: %d\n",
		                 au_svc_proc.enodata);

		AUDIT_PRINT_PROC(page, left, out, "\nsys_audit_service parameter\n");
		AUDIT_PRINT_PROC(page, left, out, "   retval: %d, size %d\n",
		                 au_svc_proc.retval, au_svc_proc.size);
		AUDIT_PRINT_PROC(page, left, out, "   saved: %lu, limit %lu\n",
		                 au_svc_proc.saved, au_svc_proc.limit);

		spin_unlock(&au_svc_proc.lock);
	} else
		AUDIT_PRINT_PROC(page, left, out, "busy");

out:
	len = (PAGE_SIZE - left);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}
#endif

void audit_proc_init(void)
{
	struct proc_dir_entry *entry;

	entry = create_proc_read_entry("audit", S_IRUSR | S_IRGRP | S_IROTH,
	                                NULL, audit_read_proc, NULL);
#ifdef CONFIG_AUDIT_DEBUG
	entry = create_proc_read_entry("audit_buffer",
	                               S_IRUSR | S_IRGRP | S_IROTH,
	                               NULL, audit_buf_proc, NULL);
	entry = create_proc_read_entry("audit_service",
	                               S_IRUSR | S_IRGRP | S_IROTH,
	                               NULL, audit_sys_svc_proc, NULL);
#endif
	if (!entry)
		printk(KERN_NOTICE "audit_init: unable to create proc entry.\n");
}
