#include <linux/config.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/sys.h>

#include <asm/uaccess.h>
#include <asm/unistd.h>

#include <linux/audit.h>
#include <linux/acl.h>

#include "buffer.h"
#include "timer.h"

#ifdef CONFIG_AUDIT_PROC 
#include "proc.h"
#endif

extern int audit_getadmin(void);

static int __init audit_setup(char *line)
{
	int err, mod;

	if (!line)
		return -EINVAL;

#ifdef CONFIG_AUDIT
	mod = AU_DISABLED;

	if (!strncmp(line, "normal", 6))
		mod = AU_NORMAL;

	if (!strncmp(line, "secure", 6))
		mod = AU_SECURE;
	
	err = audit_setmod(mod);
#ifdef CONFIG_AUDIT_DEBUG
	if (err)
		printk(KERN_DEBUG "audit_setup: setmod failed\n");
#endif

#ifndef CONFIG_AU_STATIC_ADMIN
{
	uid_t admin;

	admin = 0;
	if (line[6] == ',') {
		if (line[7] == '-') {
			admin = (uid_t) simple_strtoul(line + 8, NULL, 10);
			admin = -1 * admin;
		} else
			admin = (uid_t) simple_strtoul(line + 7, NULL, 10);
	}

	err = audit_setadmin(admin);
#ifdef CONFIG_AUDIT_DEBUG
	if (err)
		printk(KERN_DEBUG "audit_setup: set admin failed\n");
#endif
}
#endif

#else
	printk("The audit subsystem is not compiled into the kernel.\n");
#endif

	return 0;
}

__setup("audit=", audit_setup);

#ifdef CONFIG_AUDIT

extern int init_buf_cache(void);
extern int init_host_buf_cache(void);
extern int init_dyn_buf_cache(void);
extern int init_msk_cachep(void);

#ifdef CONFIG_NET_AUDIT

extern int init_net_buf_cache(void);

#ifdef CONFIG_ACL
extern int init_netacl(void);
#endif
#endif

static kmem_cache_t *au_cachep = NULL;

static inline void audit_cache_init(void *p, kmem_cache_t *cache,
                                    unsigned long flags)
{
	struct audit_struct *new_p = p;

	memset(new_p, 0, sizeof(struct audit_struct));
}


static int init_audit_cachep(void)
{
	if (au_cachep)
		return 0;
	
#ifdef CONFIG_AUDIT_DEBUG
	printk(KERN_DEBUG "audit: initializing audit structure cache ...\n");
#endif
	
	au_cachep = kmem_cache_create("audit_cachep", sizeof(struct audit_struct),
	                              0, SLAB_HWCACHE_ALIGN, audit_cache_init,
	                              NULL);
	
	if (!au_cachep) {
		printk(KERN_NOTICE "audit: init audit structure cache failed !!!\n");
		return -ENOMEM;
	}

	return 0;
}


struct audit_struct *alloc_audit_struct(int flag)
{
	struct audit_struct *new_p;

	if (!au_cachep)
		return NULL;

	new_p = kmem_cache_alloc(au_cachep, flag);
	if (new_p)
		memset((char *) new_p, 0, sizeof(*new_p));
	
	return new_p;
}

	
int kfree_audit_struct(struct audit_struct *audit)
{
	if (!au_cachep)
		return -EFAULT;
	
	if (audit)
		kmem_cache_free(au_cachep, audit);

	return 0;
}

extern void *sys_call_table[];
void *orig_sys_call_table[NR_syscalls];

extern int audit_getmodus(void);

long audit_sys_kill(int pid, int sig)
{
	if (AU_SVC_INITIALIZED(&au_svc) && audit_getmodus() != AU_DISABLED) {
		if (pid == au_svc.auditd_pid &&
		    !audit_check_capability(AU_CAP_ADM))
			return -EPERM;
	}

	return ((long (*) (int, int)) orig_sys_call_table[__NR_kill])(pid, sig);
}

extern asmlinkage long sys_auditcall(int call, unsigned long *args);

extern asmlinkage long sys_ni_syscall(void);

extern asmlinkage long audit_setuid(uid_t uid);
extern asmlinkage long audit_setuid16(old_uid_t uid);
extern asmlinkage long audit_setgid(gid_t gid);
extern asmlinkage long audit_setgid16(old_gid_t gid);
extern asmlinkage long audit_setreuid(uid_t ruid, uid_t euid);
extern asmlinkage long audit_setreuid16(old_uid_t ruid, old_uid_t euid);
extern asmlinkage long audit_setregid(gid_t rgid, gid_t egid);
extern asmlinkage long audit_setregid16(old_gid_t rgid, old_gid_t egid);
extern asmlinkage long audit_setfsuid(uid_t uid);
extern asmlinkage long audit_setfsuid16(old_uid_t uid);
extern asmlinkage long audit_setfsgid(gid_t gid);
extern asmlinkage long audit_setfsgid16(old_gid_t gid);
extern asmlinkage long audit_setresuid(uid_t ruid, uid_t euid, uid_t suid);
extern asmlinkage long audit_setresuid16(old_uid_t ruid, old_uid_t euid,
                                         old_uid_t suid);
extern asmlinkage long audit_setresgid(gid_t rgid, gid_t egid, gid_t sgid);
extern asmlinkage long audit_setresgid16(old_gid_t rgid, old_gid_t egid,
                                         old_gid_t sgid);
extern asmlinkage long audit_setgroups(int gidsetsize, gid_t *grouplist);
extern asmlinkage long audit_setgroups16(int gidsetsize, old_gid_t *grouplist);
extern asmlinkage long audit_sethostname(char *name, int len);
extern asmlinkage long audit_setdomainname(char *name, int len);

extern asmlinkage int audit_ptrace(long request, long pid, long addr,
                                   long data);

extern asmlinkage long audit_chown(const char *filename, uid_t user,
                                   gid_t group);
extern asmlinkage long audit_chown16(const char *filename, old_uid_t user,
                                     old_gid_t group);
extern asmlinkage long audit_lchown(const char *filename, uid_t user,
                                    gid_t group);
extern asmlinkage long audit_lchown16(const char * filename, old_uid_t user,
                                      old_gid_t group);
extern asmlinkage long audit_fchown(unsigned int fd, uid_t user, gid_t group);
extern asmlinkage long audit_fchown16(unsigned int fd, old_uid_t user,
                                      old_gid_t group);

extern asmlinkage long audit_socketcall(int call, unsigned long *args);

extern asmlinkage int audit_fork(struct pt_regs regs);
extern asmlinkage int audit_clone(struct pt_regs regs);
extern asmlinkage int audit_init_execve(struct pt_regs regs);


int prepare_syscall_table(void)
{
	int ret;

	ret = -EBUSY;
	if (sys_call_table[__NR_security] == sys_ni_syscall) {
		sys_call_table[__NR_security] = sys_auditcall;

		memcpy(orig_sys_call_table, sys_call_table,
		       NR_syscalls * sizeof(unsigned long));

		sys_call_table[__NR_kill] = audit_sys_kill;

		sys_call_table[__NR_setuid] = audit_setuid16;
		sys_call_table[__NR_setuid32] = audit_setuid;
		sys_call_table[__NR_setgid] = audit_setgid16;
		sys_call_table[__NR_setgid32] = audit_setgid;
		sys_call_table[__NR_setreuid] = audit_setreuid16;
		sys_call_table[__NR_setreuid32] = audit_setreuid;
		sys_call_table[__NR_setregid] = audit_setregid16;
		sys_call_table[__NR_setregid32] = audit_setregid;
		sys_call_table[__NR_setfsuid] = audit_setfsuid16;
		sys_call_table[__NR_setfsuid32] = audit_setfsuid;
		sys_call_table[__NR_setfsgid] = audit_setfsgid16;
		sys_call_table[__NR_setfsgid32] = audit_setfsgid;
		sys_call_table[__NR_setresuid] = audit_setresuid16;
		sys_call_table[__NR_setresuid32] = audit_setresuid;
		sys_call_table[__NR_setresgid] = audit_setresgid16;
		sys_call_table[__NR_setresgid32] = audit_setresgid;
		sys_call_table[__NR_setgroups] = audit_setgroups16;
		sys_call_table[__NR_setgroups32] = audit_setgroups;
		sys_call_table[__NR_sethostname] = audit_sethostname;
		sys_call_table[__NR_setdomainname] = audit_setdomainname;

		sys_call_table[__NR_chown32] = audit_chown;
		sys_call_table[__NR_chown] = audit_chown16;
		sys_call_table[__NR_lchown32] = audit_lchown;
		sys_call_table[__NR_lchown] = audit_lchown16;
		sys_call_table[__NR_fchown32] = audit_fchown;
		sys_call_table[__NR_fchown] = audit_fchown16;

		sys_call_table[__NR_ptrace] = audit_ptrace;

		sys_call_table[__NR_socketcall] = audit_socketcall;

		sys_call_table[__NR_fork] = audit_fork;
		sys_call_table[__NR_clone] = audit_clone;
		sys_call_table[__NR_execve] = audit_init_execve;

		ret = 0;
	}

	return ret;
}


int __init audit_init(void)
{
	int err;

#ifdef CONFIG_AUDIT_DEBUG
	printk(KERN_DEBUG "audit_init: initialize audit subsystem.\n");
	if (ROOT_DEV == 0x00ff) {
		KLADT_PRINT("audit_init: "
		            "nfs-based root-fs => network-stack will not stopped.\n");
		au_buf.nfs_root = 1;
	}
#endif

	err = 0;

	err = prepare_syscall_table();
	if (err) {
		printk(KERN_NOTICE "audit_init: unable to prepare syscall table!\n");
		goto err_out;
	}

	init_audit_cachep();

	init_buf_cache();
	init_host_buf_cache();
	init_dyn_buf_cache();
	init_msk_cachep();

#ifdef CONFIG_NET_AUDIT

	init_net_buf_cache();

#ifdef CONFIG_ACL
	err = init_netacl();
	if (err) {
		printk(KERN_NOTICE "audit_init: unable to initialize acl subsystem!\n");
		goto err_out;
	}
#endif
#endif

#ifdef CONFIG_AUDIT_PROC 
	audit_proc_init();
#endif

#ifdef CONFIG_AUDIT_TIMER
	create_audit_timer(&au_svc);
#endif

#ifdef CONFIG_AU_STATIC_ADMIN
	audit_setadmin(CONFIG_AUDIT_ADMIN);
#endif

out:
	return err;

err_out:
	audit_setmod(AU_DISABLED);
	goto out;
}

__initcall(audit_init);
#endif



